unit FormatSQL;

interface

uses OWideSupp, DateUtils, FormatXML;

type
  Boolean3 = (bTrue, bFalse, bNone);

function StringToSql(aStr: OWideString): OWideString;
function DateTimeToSqlStr(aDT: TDateTime): OWideString;
function SqlStrToTime(aSqlStr: OWideString): TDateTime;
function SqlStrToDate(aSqlStr: OWideString): TDateTime;
function SqlStrToDateDef(aSqlStr: OWideString; DefDate: TDateTime): TDateTime;
function SqlStrToDateTime(aSqlStr: OWideString): TDateTime;
function SqlStrToDateTimeDef(aSqlStr: OWideString; DefDateTime: TDateTime): TDateTime;
function GetSqlNullDate: OWideString; overload;
function GetSqlNullDate(aCol: OWideString): OWideString; overload;
function GetSqlNullDateTime: OWideString; overload;
function GetSqlNullDateTime(aCol: OWideString): OWideString; overload;
function GetSqlNullTime: OWideString; overload;
function GetSqlNullTime(aCol: OWideString): OWideString; overload;
function GetSqlNullInteger(aCol: OWideString): OWideString;
function GetSqlFalse(aCol: OWideString): OWideString; overload;
function IsNullDate(aSqlStr: OWideString): Boolean; overload;
function IsNullDateTime(aSqlStr: OWideString): Boolean;
function IsNullTime(aSqlStr: OWideString): Boolean;
function DateToSqlStr(aDT: TDateTime): OWideString;
function TimeToSqlStr(aDT: TDateTime): OWideString;
function FloatToSqlStr(aF: Double): OWideString;
function SqlStrToFloat(aSqlStr: OWideString): Double;
function SqlStrToFloatDef(aSqlStr: OWideString; DefResult: Double): Double;
function SqlStrToInt(aSqlStr: OWideString): Integer;
function SqlStrToIntDef(aSqlStr: OWideString; DefResult: Integer): Integer;
function SqlStrToBoolean(aString: OWideString): Boolean;
function SqlStrToBooleanDef(aString: OWideString; DefResult: Boolean): Boolean;
function BooleanToSqlStr(aBoolean: Boolean): OWideString;
function SqlStrToBoolean3(aString: OWideString): Boolean3;
function SqlStrToBoolean3Def(aString: OWideString; DefResult: Boolean3): Boolean3;
function Boolean3ToSqlStr(aBoolean3: Boolean3): OWideString;
function BooleanToBoolean3(aBoolean: Boolean): Boolean3;

const
  AllowedSqlCompInteger: Array[0..2] of WideString = ('%0:s<>''0''', '%0:s<>''''', 'NOT ISNULL(%0:s)');
  AllowedSqlCompDate: Array[0..3] of WideString = ('%0:s<>''0000-00-00''', '%0:s<>''1899-12-30''', '%0:s<>''''', 'NOT ISNULL(%0:s)');
  AllowedSqlCompDateTime: Array[0..3] of WideString = ('%0:s<>''0000-00-00 00:00:00''', '%0:s<>''1899-12-30 00:00:00''', '%0:s<>''''', '%0:s IS NOT NULL');
  AllowedSqlCompBoolean: Array[0..1] of WideString = ('%0:s=''1''', 'NOT ISNULL(%0:s)');
  AllowedSqlCompString: Array[0..1] of WideString = ('%0:s<>''''', 'NOT ISNULL(%0:s)');

implementation

uses SysUtils{$IFDEF O_DELPHI_2007UP}, WideStrUtils{$ENDIF};

function StringToSql(aStr: OWideString): OWideString;
begin
  Result := aStr;
  Result := ChangeChar(Result, '\', '\\');
  Result := ChangeChar(Result, '''', '\''');
  Result := ChangeChar(Result, '"', '\"');
end;

function DateTimeToSqlStr(aDT: TDateTime): OWideString;
begin
  if aDT = 0 then
    Result := GetSqlNullDateTime
  else
    Result := FormatDateTime('yyyy-mm-dd hh:nn:ss', aDT)
end;

function TimeToSqlStr(aDT: TDateTime): OWideString;
begin
  if aDT = 0 then
    Result := GetSqlNullTime
  else
    Result := FormatDateTime('hh:nn:ss', aDT);
end;

function DateToSqlStr(aDT: TDateTime): OWideString;
begin
  if aDT = 0 then
    Result := GetSqlNullDate
  else
    Result := FormatDateTime('yyyy-mm-dd', aDT);
end;

function SqlStrToDateTimeDef(aSqlStr: OWideString; DefDateTime: TDateTime): TDateTime;
begin
  result := SqlStrToDateTime(aSqlStr);
  if Result = 0 then
    Result := DefDateTime;
end;

function GetSqlNullDate: OWideString;
begin
  Result := '0000-00-00';
end;

function GetSqlNullDate(aCol: OWideString): OWideString;
begin
  Result := '('+aCol+'="0000-00-00" OR '+aCol+'="1899-12-30" OR '+aCol+'="" OR ISNULL('+aCol+'))';
end;

function GetSqlFalse(aCol: OWideString): OWideString; overload;
begin
  Result := '('+aCol+'="0" OR '+aCol+'="" OR ISNULL('+aCol+'))';
end;

function GetSqlNullDateTime: OWideString;
begin
  Result := '0000-00-00 00:00:00';
end;

function GetSqlNullDateTime(aCol: OWideString): OWideString;
begin
  Result := '('+aCol+'="0000-00-00 00:00:00" OR '+aCol+'="1899-12-30 00:00:00" OR '+aCol+'="" OR ISNULL('+aCol+'))';
end;

function GetSqlNullTime: OWideString;
begin
  Result := '00:00:00';
end;

function GetSqlNullTime(aCol: OWideString): OWideString;
begin
  Result := '('+aCol+'="00:00:00" OR '+aCol+'="" OR ISNULL('+aCol+'))';
end;

function GetSqlNullInteger(aCol: OWideString): OWideString;
begin
  Result := '('+aCol+'="0" OR '+aCol+'="" OR ISNULL('+aCol+'))';
end;

function IsNullDate(aSqlStr: OWideString): Boolean;
begin
  result := (aSqlStr = '') or (aSqlStr = '0000-00-00') or (aSqlStr = '1899-12-30');
end;

function IsNullDateTime(aSqlStr: OWideString): Boolean;
begin
  result := (aSqlStr = '') or (aSqlStr = '0000-00-00 00:00:00') or (aSqlStr = '1899-12-30 00:00:00');
end;

function IsNullTime(aSqlStr: OWideString): Boolean;
begin
  result := (aSqlStr = '') or (aSqlStr = '00:00:00');
end;

function SqlStrToDateTime(aSqlStr: OWideString): TDateTime;
var year, month, day, hour, minute, second: Integer;
begin
  year := StrToIntDef(Copy(aSqlStr, 1, 4), 0);
  month := StrToIntDef(Copy(aSqlStr, 6, 2), 0);
  day := StrToIntDef(Copy(aSqlStr, 9, 2), 0);

  hour := StrToIntDef(Copy(aSqlStr, 12, 2), 0);
  minute := StrToIntDef(Copy(aSqlStr, 15, 2), 0);
  second := StrToIntDef(Copy(aSqlStr, 18, 2), 0);

  if not TryEncodeDateTime(Year, Month, Day, Hour, Minute, Second, 0, result) then
    result := 0;
end;

function SqlStrToDateDef(aSqlStr: OWideString; DefDate: TDateTime): TDateTime;
var year, month, day: Integer;
begin
  year := StrToIntDef(Copy(aSqlStr, 1, 4), 0);
  month := StrToIntDef(Copy(aSqlStr, 6, 2), 0);
  day := StrToIntDef(Copy(aSqlStr, 9, 2), 0);

  if not TryEncodeDate(Year, Month, Day, result) then
    result := DefDate;
end;

function SqlStrToDate(aSqlStr: OWideString): TDateTime;
begin
  Result := SqlStrToDateDef(aSqlStr, 0);
end;

function SqlStrToTime(aSqlStr: OWideString): TDateTime;
var hour, minute, second: Integer;
begin
  hour := StrToIntDef(Copy(aSqlStr, 1, 2), 0);
  minute := StrToIntDef(Copy(aSqlStr, 4, 2), 0);
  second := StrToIntDef(Copy(aSqlStr, 7, 2), 0);

  if not TryEncodeTime(Hour, Minute, Second, 0, result) then
    result := 0;
end;

function FloatToSqlStr(aF: Double): OWideString;
begin
  result := FloatToStr(aF);
  result := ChangeChar(result, ',', '.');
  result := ChangeChar(result, '''', '');
  result := ChangeChar(result, ' ', '');
end;

function SqlStrToFloat(aSqlStr: OWideString): Double;
begin
  Result := SqlStrToFloatDef(aSqlStr, 0);
end;

function SqlStrToFloatDef(aSqlStr: OWideString; DefResult: Double): Double;
var
  FS: TFormatSettings;
begin
  GetLocaleFormatSettings(0, FS);
  FS.DecimalSeparator := '.';
  aSqlStr := ChangeChar(aSqlStr, ',', '.');
  result := StrToFloatDef(aSqlStr, DefResult, FS);
end;

function SqlStrToInt(aSqlStr: OWideString): Integer;
begin
  Result := Trunc(SqlStrToFloat(aSqlStr));
end;

function SqlStrToIntDef(aSqlStr: OWideString; DefResult: Integer): Integer;
begin
  Result := Trunc(SqlStrToFloatDef(aSqlStr, DefResult));
end;

function SqlStrToBoolean(aString: OWideString): Boolean;
begin
  Result := aString = '1';
end;

function SqlStrToBooleanDef(aString: OWideString; DefResult: Boolean): Boolean;
begin
  if not DefResult then//DefResult = FALSE
    Result := aString = '1'
  else//DefResult = TRUE
    Result := not (aString = '0');
end;

function BooleanToSqlStr(aBoolean: Boolean): OWideString;
begin
  if aBoolean then
    Result := '1'
  else
    Result := '0';
end;

function SqlStrToBoolean3(aString: OWideString): Boolean3;
begin
  case SqlStrToInt(aString) of
    1: Result := bTrue;
    2: Result := bNone;
  else
    Result := bFalse;
  end;
end;

function SqlStrToBoolean3Def(aString: OWideString; DefResult: Boolean3): Boolean3;
begin
  if aString = '1' then
    Result := bTrue
  else if aString = '0' then
    Result := bFalse
  else if aString = '2' then
    Result := bNone
  else
    Result := DefResult;
end;

function Boolean3ToSqlStr(aBoolean3: Boolean3): OWideString;
begin
  case aBoolean3 of
    bTrue: Result := '1';
    bNone: Result := '2';
  else
    Result := '0';
  end;
end;

function BooleanToBoolean3(aBoolean: Boolean): Boolean3;
begin
  if aBoolean then
    Result := bTrue
  else
    Result := bFalse;
end;

end.
