unit FormatXML;

{
    Copyright (C) 2011 Ondrej Pokorny
      http://www.kluug.at
}

{$I OWideSupp.inc}

interface

uses Classes, Types,
  {$IFDEF O_DELPHI_XE2}
  Vcl.Graphics,
  {$ELSE}
  Graphics,
  {$ENDIF}
  OWideSupp;

{$IFDEF O_DELPHI_7}
type
  TVerticalAlignment = (taAlignTop, taAlignBottom, taVerticalCenter);
{$ENDIF}


  function StringToUTF8XML(aStr: OWideString; aSpaceToNBSP: Boolean = True; aLineBreakToCRLF: Boolean = True): UTF8String;
  function StringToWideXML(aStr: OWideString; aSpaceToNBSP: Boolean = True; aLineBreakToCRLF: Boolean = True): OWideString;
  function StringToWideHTML(aStr: OWideString): OWideString;
  function XMLToString(aXMLStr: OWideString): OWideString; overload;
  function XMLToString(aXMLStr: UTF8String): OWideString; overload;
  function AlignmentToStr(aAL: TAlignment): String;
  function AlignmentToHTML(aAL: TAlignment): String;
  function AlignmentToODS(aAL: TAlignment): String;
  function StrToAlignment(aStr: String): TAlignment;
  function VAlignmentToStr(aVAL: TVerticalAlignment): String;
  function VAlignmentToHTML(aVAL: TVerticalAlignment): String;
  function VAlignmentToODS(aVAL: TVerticalAlignment): String;
  function StrToVAlignment(aStr: String): TVerticalAlignment;
  function FontStyleToHTML(Text: OWideString; aFS: TFontStyles): OWideString;
  function FontStyleToStr(aFS: TFontStyles): String;
  function StrToFontStyle(aStr: String): TFontStyles;
  function PenStyleToStr(aPS: TPenStyle): String;
  function StrToPenStyle(aStr: String): TPenStyle;
  function BrushStyleToStr(aBS: TBrushStyle): String;
  function StrToBrushStyle(aStr: String): TBrushStyle;
  function SameFont(Font1, Font2: TFont): Boolean;
  function SameBrush(Brush1, Brush2: TBrush): Boolean;
  function SamePen(Pen1, Pen2: TPen): Boolean;
  function ColorToHTML(aC: TColor; aAddCross: Boolean = True): String;

  function ChangeChar(Str: OWideString; CharFrom, CharTo: OWideString): OWideString; {$IFDEF O_DELPHI_2007UP}inline;{$ENDIF}

  function DateToODS(aDate: TDateTime): String;
  function DateTimeToODS(aDateTime: TDateTime): String;
  function ODSStrToDateTime(aSqlStr: OWideString): TDateTime;
  function TimeToODS(aTime: TDateTime): String;
  function ODSStrToTime(aSqlStr: OWideString): TDateTime;

  function URLEncode(URL: UTF8String): AnsiString; overload;
  function URLEncode(URL: WideString): AnsiString; overload;
  function URLEncode(Stream: TStream): AnsiString; overload;

implementation

uses
  SysUtils, UtfConversion, {$IFNDEF UNICODE}TntSysUtils, {$IFDEF O_DELPHI_2007UP}WideStrUtils,{$ENDIF}{$ENDIF} DateUtils;

const
  AlignmentStrHTML: Array[taLeftJustify..taCenter] of String = ('left', 'right', 'center');
  AlignmentStr: Array[taLeftJustify..taCenter] of String = ('l', 'r', 'c');
  AlignmentStrODS: Array[taLeftJustify..taCenter] of String = ('start', 'end', 'center');
  VAlignmentStrHTML: Array[taAlignTop..taVerticalCenter] of String = ('top', 'bottom', 'center');
  VAlignmentStrODS: Array[taAlignTop..taVerticalCenter] of String = ('top', 'bottom', 'middle');
  VAlignmentStr: Array[taAlignTop..taVerticalCenter] of String = ('t', 'b', 'c');

function URLEncode(Stream: TStream): AnsiString;
var bt: Byte;
begin
  Result := '';
  Stream.Position := 0;
  while Stream.Position < Stream.Size do begin
    Stream.Read(bt, 1);
    if AnsiChar(bt) in ['0'..'9', 'a'..'z', 'A'..'Z', '.', '-', '_'] then
      Result := Result + AnsiChar(bt)
    else
      Result := Result + '%' + AnsiString(IntToHex(bt, 2));
  end;
end;

function URLEncode(URL: WideString): AnsiString;
begin
  Result := URLEncode(MyUTF8Encode(URL));
end;

function URLEncode(URL: UTF8String): AnsiString;
var
  I: integer;
begin
  Result := '';

  for I := 1 to Length(URL) do begin
    if URL[I] in ['0'..'9', 'a'..'z', 'A'..'Z', '.', '-', '_'] then
      Result := Result + URL[I]
    else
      Result := Result + '%' + AnsiString(IntToHex(ord(URL[I]), 2));
  end;
end;

function DateToODS(aDate: TDateTime): String;
begin
  Result := FormatDateTime('yyyy-mm-dd', aDate);
end;

function DateTimeToODS(aDateTime: TDateTime): String;
begin
  Result := FormatDateTime('yyyy-mm-dd', aDateTime)+'T'+FormatDateTime('hh:nn:ss', aDateTime);
end;

function ODSStrToDateTime(aSqlStr: OWideString): TDateTime;
var year, month, day, hour, minute, second: Integer;
begin
  year := StrToIntDef(Copy(aSqlStr, 1, 4), 0);
  month := StrToIntDef(Copy(aSqlStr, 6, 2), 0);
  day := StrToIntDef(Copy(aSqlStr, 9, 2), 0);

  hour := StrToIntDef(Copy(aSqlStr, 12, 2), 0);
  minute := StrToIntDef(Copy(aSqlStr, 15, 2), 0);
  second := StrToIntDef(Copy(aSqlStr, 18, 2), 0);

  if not TryEncodeDateTime(Year, Month, Day, Hour, Minute, Second, 0, result) then
    result := 0;
end;

function TimeToODS(aTime: TDateTime): String;
begin
  Result := 'PT'+FormatDateTime('hh', aTime)+'H'+FormatDateTime('nn', aTime)+'M'+FormatDateTime('ss', aTime)+'S';
end;

function ODSStrToTime(aSqlStr: OWideString): TDateTime;
var hour, minute, second: Integer;
begin
  hour := StrToIntDef(Copy(aSqlStr, 3, 2), 0);
  minute := StrToIntDef(Copy(aSqlStr, 6, 2), 0);
  second := StrToIntDef(Copy(aSqlStr, 9, 2), 0);

  if not TryEncodeTime(Hour, Minute, Second, 0, result) then
    result := 0;
end;

function ColorToHTML(aC: TColor; aAddCross: Boolean): String;
var xC: Integer;
begin
  xC := ColorToRGB(aC);
  if aAddCross then
    Result := '#'
  else
    Result := '';
  Result := Result +
    IntToHex(Byte(xC), 2) +
    IntToHex(Byte(xC shr 8), 2) +
    IntToHex(Byte(xC shr 16), 2);
end;

function VAlignmentToStr(aVAL: TVerticalAlignment): String;
begin
  result := VAlignmentStr[aVAl];
end;

function VAlignmentToHTML(aVAL: TVerticalAlignment): String;
begin
  result := VAlignmentStrHTML[aVAl];
end;

function VAlignmentToODS(aVAL: TVerticalAlignment): String;
begin
  result := VAlignmentStrODS[aVAl];
end;

function StrToVAlignment(aStr: String): TVerticalAlignment;
var I: TVerticalAlignment;
begin
  Result := taAlignTop;
  for I := low(I) to high(I) do
  if Pos(VAlignmentStr[I], aStr) = 1 then begin
    Result := I;
    exit;
  end;
end;

function AlignmentToStr(aAL: TAlignment): String;
begin
  result := AlignmentStr[aAl];
end;

function AlignmentToHTML(aAL: TAlignment): String;
begin
  result := AlignmentStrHTML[aAl];
end;

function AlignmentToODS(aAL: TAlignment): String;
begin
  result := AlignmentStrODS[aAl];
end;

function StrToAlignment(aStr: String): TAlignment;
var I: TAlignment;
begin
  Result := taLeftJustify;
  for I := low(I) to high(I) do
  //if aStr = AlignmentStr[I] then begin
  if Pos(AlignmentStr[I], aStr) = 1 then begin
    Result := I;
    exit;
  end;
end;


function FontStyleToHTML(Text: OWideString; aFS: TFontStyles): OWideString;
begin
  result := Text;
  if fsBold in aFS then
    result := '<b>'+result+'</b>';
  if fsItalic in aFS then
    result := '<i>'+result+'</i>';
  if fsUnderline in aFS then
    result := '<u>'+result+'</u>';
  if fsStrikeOut in aFS then
    result := '<s>'+result+'</s>';
end;

function FontStyleToStr(aFS: TFontStyles): String;
begin
  result := '';
  if fsBold in aFS then
    result := result + 'b';
  if fsItalic in aFS then
    result := result + 'i';
  if fsUnderline in aFS then
    result := result + 'u';
  if fsStrikeOut in aFS then
    result := result + 's';
end;

function StrToFontStyle(aStr: String): TFontStyles;
var OldFormat: Boolean;
begin
  aStr := Trim(LowerCase(aStr));
  result := [];
  OldFormat := (Pos(';', aStr) > 0) or ((Length(aStr) > 1) and not (AnsiChar(aStr[2]) in ['b', 'i', 'u', 's']));
  //old format
  if OldFormat then begin
    if Pos('bold', aStr) > 0 then
      result := result + [fsBold];
    if Pos('italic', aStr) > 0 then
      result := result + [fsItalic];
    if Pos('underline', aStr) > 0 then
      result := result + [fsUnderline];
    if Pos('strikeout', aStr) > 0 then
      result := result + [fsStrikeOut];
  end else begin
    if Pos('b', aStr) > 0 then
      result := result + [fsBold];
    if Pos('i', aStr) > 0 then
      result := result + [fsItalic];
    if Pos('u', aStr) > 0 then
      result := result + [fsUnderline];
    if Pos('s', aStr) > 0 then
      result := result + [fsStrikeOut];
  end;
end;

function SameFont(Font1, Font2: TFont): Boolean;
begin

  result := (Font1.Color = Font2.Color) and (Font1.Size = Font2.Size) and
    SameText(Font1.Name, Font2.Name) and (Font1.Style = Font2.Style);
end;

function BrushStyleToStr(aBS: TBrushStyle): String;
begin
  Result := IntToStr(Ord(aBS));
end;

function StrToBrushStyle(aStr: String): TBrushStyle;
begin
  Result := TBrushStyle(Ord(StrToIntDef(aStr, 0)));
end;

function PenStyleToStr(aPS: TPenStyle): String;
begin
  Result := IntToStr(Ord(aPS));
end;

function StrToPenStyle(aStr: String): TPenStyle;
begin
  Result := TPenStyle(Ord(StrToIntDef(aStr, 0)));
end;

function SameBrush(Brush1, Brush2: TBrush): Boolean;
begin
  result := (Brush1.Color = Brush2.Color) and (Brush1.Style = Brush2.Style);
end;

function SamePen(Pen1, Pen2: TPen): Boolean;
begin
  result := (Pen1.Color = Pen2.Color) and (Pen1.Style = Pen2.Style) and
    (Pen1.Width = Pen2.Width);
end;

function StringToUTF8XML(aStr: OWideString; aSpaceToNBSP, aLineBreakToCRLF: Boolean): UTF8String;
begin
  Result := MyUTF8Encode(StringToWideXML(aStr, aSpaceToNBSP, aLineBreakToCRLF));
end;

function StringToWideHTML(aStr: OWideString): OWideString;
begin
  result := aStr;
  //result := MyUTF8Encode(result); NESMI TU BYT - RESI SE AUTOMATICKY
  result := ChangeChar(result, '&', '&amp;');
  result := ChangeChar(result, '<', '&lt;');
  result := ChangeChar(result, '>', '&gt;');
  result := ChangeChar(result, '''', '&apos;');
  result := ChangeChar(result, '"', '&quot;');
  result := ChangeChar(result, #13, '<br>');
  result := ChangeChar(result, #10, '');
  if (Length(result) > 0) then begin//Spaces at the beginning and the end - they would be trimmed otherwise
    if (result[1] = ' ') then begin
      Delete(Result, 1, 1);
      result := '&nbsp;' + result;
    end;
    if (result[Length(result)] = ' ') then begin
      Delete(Result, Length(result), 1);
      result := result + '&nbsp;';
    end;
  end;
end;

function StringToWideXML(aStr: OWideString; aSpaceToNBSP, aLineBreakToCRLF: Boolean): OWideString;
begin
  result := aStr;
  //result := MyUTF8Encode(result); NESMI TU BYT - RESI SE AUTOMATICKY
  result := ChangeChar(result, '&', '&amp;');
  result := ChangeChar(result, '<', '&lt;');
  result := ChangeChar(result, '>', '&gt;');
  result := ChangeChar(result, '''', '&apos;');
  result := ChangeChar(result, '"', '&quot;');
  if aLineBreakToCRLF then begin
    result := ChangeChar(result, #13, '&#xD;');
    result := ChangeChar(result, #10, '&#xA;');
  end;
  if aSpaceToNBSP and (Result <> '') then begin//Spaces at the beginning and the end - they would be trimmed otherwise
    if (result[1] = ' ') then begin
      Delete(Result, 1, 1);
      result := '&nbsp;' + result;
    end;
    if (result[Length(result)] = ' ') then begin
      Delete(Result, Length(result), 1);
      result := result + '&nbsp;';
    end;
  end;
end;
function XMLToString(aXMLStr: OWideString): OWideString;
begin
  result := aXMLStr;
  result := ChangeChar(result, '&nbsp;', ' ');
  result := ChangeChar(result, '&sp;', ' ');//ow, my fault - its here just because of old formats! :( sorry
  result := ChangeChar(result, '&lt;', '<');
  result := ChangeChar(result, '&rt;', '>');//ow, my fault - its here just because of old formats! :( sorry
  result := ChangeChar(result, '&gt;', '>');
  result := ChangeChar(result, '&apos;', '''');
  result := ChangeChar(result, '&quot;', '"');
  result := ChangeChar(result, '&br;', #13#10);//ow, my fault - its here just because of old formats! :( sorry
  result := ChangeChar(result, '&#xD;', #13#10);
  result := ChangeChar(result, '&#xA;', '');
  result := ChangeChar(result, '&#13;', #13#10);
  result := ChangeChar(result, '&#10;', '');

  //must be last!!!
  result := ChangeChar(result, '&amp;', '&');
end;

function XMLToString(aXMLStr: UTF8String): OWideString;
begin
  result := XMLToString(MyUTF8Decode(aXMLStr));
end;

function ChangeChar(Str: OWideString; CharFrom, CharTo: OWideString): OWideString;
begin
  Result := OWideStringReplace(Str, CharFrom, CharTo, [rfReplaceAll, rfIgnoreCase]);
end;

end.
