unit OFiller;

{
    OFiller

    Copyright (C) 2012 Ondrej Pokorny
      http://www.kluug.at

}

{$I OWideSupp.inc}

interface

uses SysUtils, Classes, OWideSupp, OZip;

type
  TCustomOFiller = class;
  TCustomOFillerClass = class of TCustomOFiller;
  
  TCustomOFiller = class(TObject)
  private
    fZip: TOZipArchive;
    fFS: TOFileStream;

    fText: OWideString;

    procedure LoadText;
  protected
    class function TextFileName: String; virtual; abstract;
  public
    procedure LoadFromStream(aStream: TStream);
    procedure LoadFromFile(aFileName: OWideString);
    procedure Save;
    procedure Close;

    procedure ReplaceText(const OldPattern, NewPattern: OWideString);
    procedure ReplaceTextRAW(const OldPattern, NewPattern: OWideString);
  public
    constructor Create;
    destructor Destroy; override;
  public
    property Text: OWideString read fText write fText;
  end;

  TOFillerXLSX = class(TCustomOFiller)
  protected
    class function TextFileName: String; override;
  end;
  
  TOFillerDOCX = class(TCustomOFiller)
  protected
    class function TextFileName: String; override;
  end;
  
  TOFillerODS = class(TCustomOFiller)
  protected
    class function TextFileName: String; override;
  end;

  TOFillerODT = class(TCustomOFiller)
  protected
    class function TextFileName: String; override;
  end;
  
implementation

uses AbArcTyp;

{ TCustomOFiller }

procedure TCustomOFiller.Close;
begin
  if Assigned(fZip) then
    FreeAndNil(fZip);
  if Assigned(fFS) then
    FreeAndNil(fFS);

  fText := '';
end;

constructor TCustomOFiller.Create;
begin
  inherited;

  fZip := nil;
  fFS := nil;
end;

destructor TCustomOFiller.Destroy;
begin
  Close;

  inherited;
end;

procedure TCustomOFiller.LoadFromFile(aFileName: OWideString);
begin
  Close;
  
  fFS := TOFileStream.Create(aFileName, fmOpenReadWrite);
  //fZip := TOZipArchive.Create(aFileName, fmOpenRead);
  fFS.Position := 0;
  fZip := TOZipArchive.CreateFromStream(fFS, '');
  fZip.Load;

  LoadText;
end;

procedure TCustomOFiller.LoadFromStream(aStream: TStream);
begin
  Close;

  fZip := TOZipArchive.CreateFromStream(aStream, '');
  fZip.Load;

  LoadText;
end;

procedure TCustomOFiller.LoadText;
var
  xSS: TStringStream;
begin
  {$IFDEF UNICODE}
  xSS := TStringStream.Create('', TEncoding.UTF8);
  {$ELSE}
  xSS := TStringStream.Create('');
  {$ENDIF}
  try
    fZip.ExtractToStream(TextFileName, xSS);
    xSS.Position := 0;
    {$IFDEF UNICODE}
    fText := xSS.DataString;
    {$ELSE}
    fText := UTF8Decode(xSS.DataString);
    {$ENDIF}
  finally
    xSS.Free;
  end;
end;

procedure TCustomOFiller.ReplaceText(const OldPattern, NewPattern: OWideString);
  function StringToWideXML(aStr: OWideString): OWideString; {$IFDEF O_DELPHI_2007UP}inline;{$ENDIF}
  begin
    result := aStr;
    result := OWideStringReplace(result, '&', '&amp;', [rfReplaceAll]);
    result := OWideStringReplace(result, '<', '&lt;', [rfReplaceAll]);
    result := OWideStringReplace(result, '>', '&gt;', [rfReplaceAll]);
    result := OWideStringReplace(result, '''', '&apos;', [rfReplaceAll]);
    result := OWideStringReplace(result, '"', '&quot;', [rfReplaceAll]);
  end;
begin
  fText := OWideStringReplace(fText, OldPattern, StringToWideXML(NewPattern), [rfReplaceAll, rfIgnoreCase]);
end;

procedure TCustomOFiller.ReplaceTextRAW(const OldPattern,
  NewPattern: OWideString);
begin
  fText := OWideStringReplace(fText, OldPattern, NewPattern, [rfReplaceAll, rfIgnoreCase]);
end;

procedure TCustomOFiller.Save;
var xSS: TStringStream;
begin
  {$IFDEF UNICODE}
  xSS := TStringStream.Create(fText, TEncoding.UTF8);
  {$ELSE}
  xSS := TStringStream.Create(UTF8Encode(fText));
  {$ENDIF}
  try
    fZip.StoreOptions := fZip.StoreOptions + [soReplace];
    fZip.AddFromStream(TextFileName, xSS);

    fZip.Save;
  finally
    xSS.Free;
  end;
end;

{ TOFillerXLSX }

class function TOFillerXLSX.TextFileName: String;
begin
  Result := 'xl/sharedStrings.xml';
end;

{ TOFillerDOCX }

class function TOFillerDOCX.TextFileName: String;
begin
  Result := 'word/document.xml';
end;

{ TOFillerODS }

class function TOFillerODS.TextFileName: String;
begin
  Result := 'content.xml';
end;

{ TOFillerODT }

class function TOFillerODT.TextFileName: String;
begin
  Result := 'content.xml';
end;

end.
