unit OOmniXML;

interface

uses SysUtils, OmniXML_Types, OmniXML;

function CreateElement(aParentNode: IXMLNode; aTagName: XmlString): IXMLElement;
function FindCreateElement(aParentNode: IXMLNode; aTagName: XmlString): IXMLElement;
procedure RemoveElementSave(aParentNode: IXMLNode; aTagName: XmlString); overload;
procedure RemoveElementSave(aParentNode: IXMLNode; aChildNode: IXMLNode); overload;
procedure RemoveAttributeSave(aParentNode: IXMLElement; aAttributeName: XmlString);

const
  DefXMLHeader = 'version="1.0" encoding="UTF-8" standalone="yes"';
  DefXMLHeaderBrackets = '<?xml '+DefXMLHeader+'?>';

implementation

function CreateElement(aParentNode: IXMLNode; aTagName: XmlString): IXMLElement;
begin
  Result := aParentNode.AppendChild(aParentNode.OwnerDocument.CreateElement(aTagName)) as IXMLElement;
end;

function FindCreateElement(aParentNode: IXMLNode; aTagName: XmlString): IXMLElement;
var xResult: IXMLNode;
begin
  xResult := aParentNode.SelectSingleNode(aTagName);
  if Assigned(xResult) and Supports(xResult, IXMLElement) then
    Result := xResult as IXMLElement
  else
    Result := aParentNode.AppendChild(aParentNode.OwnerDocument.CreateElement(aTagName)) as IXMLElement;
end;

procedure RemoveElementSave(aParentNode: IXMLNode; aTagName: XmlString);
var xNode: IXMLNode;
begin
  try
    xNode := aParentNode.SelectSingleNode(aTagName);
    if Assigned(xNode) then
      aParentNode.RemoveChild(xNode)
  except
    //nothing
  end;
end;

procedure RemoveElementSave(aParentNode: IXMLNode; aChildNode: IXMLNode);
begin
  try
    if (aParentNode.ChildNodes.IndexOf(aChildNode) >= 0) then
      aParentNode.RemoveChild(aChildNode)
  except
    //nothing
  end;
end;

procedure RemoveAttributeSave(aParentNode: IXMLElement; aAttributeName: XmlString);
var xAttr: IXMLNode;
begin
  try
    xAttr := aParentNode.GetAttributeNode(aAttributeName);
    if Assigned(xAttr) then
      aParentNode.Attributes.Remove(xAttr)
  except
    //nothing
  end;
end;

end.
