unit OZip;

{
    Copyright (C) 2011 Ondrej Pokorny
      http://www.kluug.at
}

interface

uses Classes, AbArcTyp, AbZipTyp, AbUnzPrc;

type
  TOZipArchive = class(TAbZipArchive)
  private
    procedure ZipProc(Sender: TObject; Item: TAbArchiveItem; OutStream: TStream);
    procedure ZipFromStreamProc(Sender: TObject; Item: TAbArchiveItem; OutStream, InStream: TStream );

    procedure UnzipProc(Sender: TObject; Item: TAbArchiveItem; const NewName: String);
    procedure UnzipToStreamProc(Sender: TObject; Item: TAbArchiveItem; OutStream: TStream);
    procedure TestItemProc(Sender: TObject; Item: TAbArchiveItem);
  public
    constructor CreateFromStream(aStream: TStream; const ArchiveName: String); override;
  public
    procedure AddFromStreamEx(const NewName, NewPath: string; aStream : TStream);
  end;

implementation

uses AbExcept, AbZipPrc, AbUtils, AbConst, OFileUtils;

{ TOZipArchive }

procedure TOZipArchive.AddFromStreamEx(const NewName, NewPath: string;
  aStream: TStream);
begin
  AddFromStream(NormPath(NewPath, True, '\')+NewName, aStream);
end;

constructor TOZipArchive.CreateFromStream(aStream: TStream;
  const ArchiveName: String);
begin
  inherited CreateFromStream(aStream, ArchiveName);

  FInsertHelper           := ZipProc;
  FInsertFromStreamHelper := ZipFromStreamProc;
  FTestHelper             := TestItemProc;
  FExtractHelper          := UnzipProc;
  FExtractToStreamHelper  := UnzipToStreamProc;
end;

procedure TOZipArchive.TestItemProc(Sender: TObject; Item: TAbArchiveItem);
begin
  AbTestZipItem(TAbZipArchive(Sender), TAbZipItem(Item));
end;

procedure TOZipArchive.UnzipProc(Sender: TObject; Item: TAbArchiveItem;
  const NewName: String);
begin
  AbUnzip( TAbZipArchive(Sender), TAbZipItem(Item), NewName);
end;

procedure TOZipArchive.UnzipToStreamProc(Sender: TObject; Item: TAbArchiveItem;
  OutStream: TStream);
begin
  AbUnzipToStream(TAbZipArchive(Sender), TAbZipItem(Item), OutStream);
end;

procedure TOZipArchive.ZipFromStreamProc(Sender: TObject; Item: TAbArchiveItem;
  OutStream, InStream: TStream);
begin
  if Assigned(InStream) then
    AbZipFromStream(TAbZipArchive(Sender), TAbZipItem(Item),
                    OutStream, InStream)
  else
    raise EAbZipNoInsertion.Create;
end;

procedure TOZipArchive.ZipProc(Sender: TObject; Item: TAbArchiveItem;
  OutStream: TStream);
begin
  AbZip(TAbZipArchive(Sender), TAbZipItem(Item), OutStream);
end;

end.

