unit oExport;

{
    oExport

    Copyright (C) 2011-2012 Ondrej Pokorny
      http://www.kluug.at


  *** BEGIN LICENSE BLOCK *****
  licensed under a MPL/GPL/LGPL tri-license
  Version: MPL 1.1/GPL 2.0/LGPL 2.1

  The contents of this file are subject to the Mozilla Public License Version
  1.1 (the "License"); you may not use this file except in compliance with
  the License. You may obtain a copy of the License at
  http://www.mozilla.org/MPL

  Software distributed under the License is distributed on an "AS IS" basis,
  WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
  for the specific language governing rights and limitations under the License.

  The Initial Developer of the Original Code is Ondrej Pokorny.

  Portions created by the Initial Developer are Copyright (C) 2011
  the Initial Developer. All Rights Reserved.

  Contributor(s):
  Alternatively, the contents of this file may be used under the terms of
  either the GNU General Public License Version 2 or later (the "GPL"), or
  the GNU Lesser General Public License Version 2.1 or later (the "LGPL"),
  in which case the provisions of the GPL or the LGPL are applicable instead
  of those above. If you wish to allow use of your version of this file only
  under the terms of either the GPL or the LGPL, and not to allow others to
  use your version of this file under the terms of the MPL, indicate your
  decision by deleting the provisions above and replace them with the notice
  and other provisions required by the GPL or the LGPL. If you do not delete
  the provisions above, a recipient may use your version of this file under
  the terms of any one of the MPL, the GPL or the LGPL.

  ***** END LICENSE BLOCK *****



  SMALL INFORMATION:
  ------------------
  Excel 2007 asks for saving changes of the generated document even if you don't
  change anything. This happens only when you use TExportCellFormulaNumber.



  RELEASE INFORMATION:
  --------------------
  You find release information in "release.txt".
  
}

{$I OWideSupp.inc}

interface

uses Classes, {$IFDEF O_DELPHI_2007UP}WideStrings, {$ENDIF}OwnedList, OWideSupp,
  {$IFDEF O_DELPHI_XE2}
  Vcl.Graphics,
  {$ELSE}
  Graphics,
  {$ENDIF}
  {$IFDEF O_USE_TNT}TntClasses,{$ENDIF}
  AbZipTyp, AbArcTyp, Contnrs, OmniXML, FormatXML;

type
  //TExportType = (etTXT, etXLS);

  TDigitFormat = (df0, df1, df2, df3, dfPercent0, dfPercent1);
  TNumberFormat = -1..3;

  TCellBorder = (cbTop, cbRight, cbBottom, cbLeft);
  TCellBorders = Set of TCellBorder;

  TExportXLSX_Chain = class;
  TExportXLSX_StyleSheet = class;

  TExportODS_ColStyles = class;
  TExportODS_RowStyles = class;
  TExportODS_CellStyles = class;
  TExportODS_NumberStyles = class;

  TExportObject = class;

  TExportCellClass = class of TExportCell;
  TExportCell = class(TOwnedListObject)
  private
    fAlignment: TAlignment;
    fVAlignment: TVerticalAlignment;
    fWrapText: Boolean;
    fFont: TFont;
    //fFontStyle: TFontStyles;
    //fFontColor: TColor;
    fBGColor: TColor;

    fColSpan: Integer;
    fRowSpan: Integer;
    //fCSSClass: OWideString;
    fWidth: Integer;
    fHeight: Integer;
    fBorders: TCellBorders;

  protected
    procedure DoCreate; override;

    procedure AssignTo(Dest: TPersistent); override;

    procedure DoAsHTML(var OutText: OWideString; var OutCSSClass: OWideString;
      var OutCSSStyle: OWideString; var OutParameters: OWideString); virtual;

    procedure DoAsXLSX(aCellNode, aColsNode, aRowNode, aMergeCellsNode: IXMLElement;
      aCellName, aWorkSheetId: OWideString;
      aSharedStrings: TOWideStrings; aStyleSheet: TExportXLSX_StyleSheet;
      aChain: TExportXLSX_Chain; aRow, aCol: Integer; out outStyleId: Integer); virtual;

    procedure DoAsODS(aRowNode, aCellNode: IXMLElement; aColNodes: TInterfaceList;
      aColStyles: TExportODS_ColStyles; aRowStyles: TExportODS_RowStyles;
      aCellStyles: TExportODS_CellStyles; aNumberStyles: TExportODS_NumberStyles;
      aRealCol: Integer); virtual;

    function GetSqlText: OWideString; virtual; abstract;
    procedure SetSqlText(const Value: OWideString); virtual; abstract;
  public
    function GetExportObject: TExportObject;
    function AsHTML: OWideString;

    property Alignment: TAlignment read fAlignment write fAlignment;
    property VAlignment: TVerticalAlignment read fVAlignment write fVAlignment;
    property WrapText: Boolean read fWrapText write fWrapText;
    property Font: TFont read fFont;
    //property FontStyle: TFontStyles read fFontStyle write fFontStyle;
    //property FontColor: TColor read fFontColor write fFontColor;
    property BGColor: TColor read fBGColor write fBGColor;

    property ColSpan: Integer read fColSpan write fColSpan;
    property RowSpan: Integer read fRowSpan write fRowSpan;
    property Width: Integer read fWidth write fWidth;
    property Height: Integer read fHeight write fHeight;
    property Borders: TCellBorders read fBorders write fBorders;

    property SqlText: OWideString read GetSqlText write SetSqlText;
    //property CSSClass: OWideString read fCSSClass write fCSSClass;
  end;

  TExportRow = class;
  TExportCells = class(TOwnedList)
  private
    fExportRow: TExportRow;
  private
    function GetC(Index: Integer): TExportCell;
    procedure SetC(Index: Integer; const Value: TExportCell);
  protected
    procedure DoCreate; override;
  public
    function GetExportObject: TExportObject;

    property Items[Index: Integer]: TExportCell read GetC write SetC; default;
  end;

  TExportCellString = class(TExportCell)
  private
    fText: OWideString;
    procedure SetText(const Value: OWideString);
  protected
    procedure DoCreate; override;

    procedure AssignTo(Dest: TPersistent); override;

    procedure DoAsHTML(var OutText: OWideString; var OutCSSClass: OWideString;
      var OutCSSStyle: OWideString; var OutParameters: OWideString); override;

    procedure DoAsXLSX(aCellNode, aColsNode, aRowNode, aMergeCellsNode: IXMLElement;
      aCellName, aWorkSheetId: OWideString;
      aSharedStrings: TOWideStrings; aStyleSheet: TExportXLSX_StyleSheet;
      aChain: TExportXLSX_Chain; aRow, aCol: Integer; out outStyleId: Integer); override;

    procedure DoAsODS(aRowNode, aCellNode: IXMLElement; aColNodes: TInterfaceList;
      aColStyles: TExportODS_ColStyles; aRowStyles: TExportODS_RowStyles;
      aCellStyles: TExportODS_CellStyles; aNumberStyles: TExportODS_NumberStyles;
      aRealCol: Integer); override;

    function GetSqlText: OWideString; override;
    procedure SetSqlText(const Value: OWideString); override;
  public
    property Text: OWideString read fText write SetText;
  end;

  TExportCellCustomNumber = class(TExportCell)
  private
    fValue: Double;
  protected
    procedure DoCreate; override;

    procedure AssignTo(Dest: TPersistent); override;
  public
    function GetNumberFormatXLSXStr: OWideString; virtual; abstract;
    procedure CreateStyleNodeODS(aStylesNode: IXMLElement; aStyleName: OWideString); virtual; abstract;
  public
    property Value: Double read fValue write fValue;
  end;

  TExportCellDate = class(TExportCellCustomNumber)
  private
    //fDateTime: TDateTime;
    function GetDateTime: TDateTime;
    procedure SetDateTime(const aValue: TDateTime);
  protected
    procedure DoAsXLSX(aCellNode, aColsNode, aRowNode, aMergeCellsNode: IXMLElement;
      aCellName, aWorkSheetId: OWideString;
      aSharedStrings: TOWideStrings; aStyleSheet: TExportXLSX_StyleSheet;
      aChain: TExportXLSX_Chain; aRow, aCol: Integer; out outStyleId: Integer); override;

    procedure DoAsODS(aRowNode, aCellNode: IXMLElement; aColNodes: TInterfaceList;
      aColStyles: TExportODS_ColStyles; aRowStyles: TExportODS_RowStyles;
      aCellStyles: TExportODS_CellStyles; aNumberStyles: TExportODS_NumberStyles;
      aRealCol: Integer); override;
  public
    property DateTime: TDateTime read GetDateTime write SetDateTime;

    function GetNumberFormatXLSXStr: OWideString; override;
    procedure CreateStyleNodeODS(aStylesNode: IXMLElement; aStyleName: OWideString); override;

    function GetSqlText: OWideString; override;
    procedure SetSqlText(const Value: OWideString); override;
  end;

  TExportCellDateTime = class(TExportCellDate)
  protected
    procedure DoAsXLSX(aCellNode, aColsNode, aRowNode, aMergeCellsNode: IXMLElement;
      aCellName, aWorkSheetId: OWideString;
      aSharedStrings: TOWideStrings; aStyleSheet: TExportXLSX_StyleSheet;
      aChain: TExportXLSX_Chain; aRow, aCol: Integer; out outStyleId: Integer); override;

    procedure DoAsODS(aRowNode, aCellNode: IXMLElement; aColNodes: TInterfaceList;
      aColStyles: TExportODS_ColStyles; aRowStyles: TExportODS_RowStyles;
      aCellStyles: TExportODS_CellStyles; aNumberStyles: TExportODS_NumberStyles;
      aRealCol: Integer); override;
  public
    function GetNumberFormatXLSXStr: OWideString; override;
    procedure CreateStyleNodeODS(aStylesNode: IXMLElement; aStyleName: OWideString); override;

    function GetSqlText: OWideString; override;
    procedure SetSqlText(const Value: OWideString); override;
  end;
  TExportCellDateTimeWS = class(TExportCellDateTime)//without seconds
  public
    function GetNumberFormatXLSXStr: OWideString; override;
    procedure CreateStyleNodeODS(aStylesNode: IXMLElement; aStyleName: OWideString); override;
  end;

  TExportCellTime = class(TExportCellDate)
  protected
    procedure DoAsXLSX(aCellNode, aColsNode, aRowNode, aMergeCellsNode: IXMLElement;
      aCellName, aWorkSheetId: OWideString;
      aSharedStrings: TOWideStrings; aStyleSheet: TExportXLSX_StyleSheet;
      aChain: TExportXLSX_Chain; aRow, aCol: Integer; out outStyleId: Integer); override;

    procedure DoAsODS(aRowNode, aCellNode: IXMLElement; aColNodes: TInterfaceList;
      aColStyles: TExportODS_ColStyles; aRowStyles: TExportODS_RowStyles;
      aCellStyles: TExportODS_CellStyles; aNumberStyles: TExportODS_NumberStyles;
      aRealCol: Integer); override;
  public
    function GetNumberFormatXLSXStr: OWideString; override;
    procedure CreateStyleNodeODS(aStylesNode: IXMLElement; aStyleName: OWideString); override;

    function GetSqlText: OWideString; override;
    procedure SetSqlText(const Value: OWideString); override;
  end;
  TExportCellTimeWS = class(TExportCellTime)//without seconds
  public
    function GetNumberFormatXLSXStr: OWideString; override;
    procedure CreateStyleNodeODS(aStylesNode: IXMLElement; aStyleName: OWideString); override;
  end;

  TExportCellNumber = class(TExportCellCustomNumber)
  private
    fFormat: TNumberFormat;
  protected
    procedure DoCreate; override;

    function NumberFormatPostfix: OWideString; virtual;

    procedure DoAsHTML(var OutText: OWideString; var OutCSSClass: OWideString;
      var OutCSSStyle: OWideString; var OutParameters: OWideString); override;

    procedure DoAsXLSX(aCellNode, aColsNode, aRowNode, aMergeCellsNode: IXMLElement;
      aCellName, aWorkSheetId: OWideString;
      aSharedStrings: TOWideStrings; aStyleSheet: TExportXLSX_StyleSheet;
      aChain: TExportXLSX_Chain; aRow, aCol: Integer; out outStyleId: Integer); override;

    procedure DoAsODS(aRowNode, aCellNode: IXMLElement; aColNodes: TInterfaceList;
      aColStyles: TExportODS_ColStyles; aRowStyles: TExportODS_RowStyles;
      aCellStyles: TExportODS_CellStyles; aNumberStyles: TExportODS_NumberStyles;
      aRealCol: Integer); override;

    function GetSqlText: OWideString; override;
    procedure SetSqlText(const Value: OWideString); override;
  public
    property Format: TNumberFormat read fFormat write fFormat;

    function GetNumberFormatXLSXStr: OWideString; override;
    procedure CreateStyleNodeODS(aStylesNode: IXMLElement; aStyleName: OWideString); override;
  end;
  TExportCellPercent = class(TExportCellNumber)
  protected
    function NumberFormatPostfix: OWideString; override;

    procedure DoAsODS(aRowNode, aCellNode: IXMLElement; aColNodes: TInterfaceList;
      aColStyles: TExportODS_ColStyles; aRowStyles: TExportODS_RowStyles;
      aCellStyles: TExportODS_CellStyles; aNumberStyles: TExportODS_NumberStyles;
      aRealCol: Integer); override;
  public
    function GetNumberFormatXLSXStr: OWideString; override;
    procedure CreateStyleNodeODS(aStylesNode: IXMLElement; aStyleName: OWideString); override;
  end;

  TExportCellFormulaNumber = class(TExportCellNumber)
  private
    fFormula: OWideString;
    procedure SetFormula(const Value: OWideString);
  protected
    procedure DoCreate; override;

    procedure AssignTo(Dest: TPersistent); override;

    procedure DoAsHTML(var OutText: OWideString; var OutCSSClass: OWideString;
      var OutCSSStyle: OWideString; var OutParameters: OWideString); override;

    procedure DoAsXLSX(aCellNode, aColsNode, aRowNode, aMergeCellsNode: IXMLElement;
      aCellName, aWorkSheetId: OWideString;
      aSharedStrings: TOWideStrings; aStyleSheet: TExportXLSX_StyleSheet;
      aChain: TExportXLSX_Chain; aRow, aCol: Integer; out outStyleId: Integer); override;

    procedure DoAsODS(aRowNode, aCellNode: IXMLElement; aColNodes: TInterfaceList;
      aColStyles: TExportODS_ColStyles; aRowStyles: TExportODS_RowStyles;
      aCellStyles: TExportODS_CellStyles; aNumberStyles: TExportODS_NumberStyles;
      aRealCol: Integer); override;
  public
    property Formula: OWideString read fFormula write SetFormula;
  end;

  TExportCellFormulaPercent = class(TExportCellFormulaNumber)
  protected
    function NumberFormatPostfix: OWideString; override;

    procedure DoAsODS(aRowNode, aCellNode: IXMLElement; aColNodes: TInterfaceList;
      aColStyles: TExportODS_ColStyles; aRowStyles: TExportODS_RowStyles;
      aCellStyles: TExportODS_CellStyles; aNumberStyles: TExportODS_NumberStyles;
      aRealCol: Integer); override;
  public
    function GetNumberFormatXLSXStr: OWideString; override;
    procedure CreateStyleNodeODS(aStylesNode: IXMLElement; aStyleName: OWideString); override;
  end;

  TExportRow = class(TOwnedListObject)
  private
    fCells: TExportCells;
  protected
    procedure DoCreate; override;
    procedure DoDestroy; override;
  public
    function GetExportObject: TExportObject;

    property Cells: TExportCells read fCells;

    function AsHTML: OWideString; virtual;
  public
    function AddCellString(const aText: OWideString;
      const aFontStyle: TFontStyles = []; const aBorders: TCellBorders = [cbTop, cbRight, cbBottom, cbLeft]): TExportCellString;
    function AddCellNumber(const aValue: Double; const aFormat: TNumberFormat = 2;
      const aFontStyle: TFontStyles = []; const aBorders: TCellBorders = [cbTop, cbRight, cbBottom, cbLeft]): TExportCellNumber;
    function AddCellPercent(const aValue: Double; const aFormat: TNumberFormat = 2;
      const aFontStyle: TFontStyles = []; const aBorders: TCellBorders = [cbTop, cbRight, cbBottom, cbLeft]): TExportCellPercent;
    function AddCellFormulaNumber(const aFormula: OWideString; const aFormat: TNumberFormat = 2;
      const aFontStyle: TFontStyles = []; const aBorders: TCellBorders = [cbTop, cbRight, cbBottom, cbLeft]): TExportCellFormulaNumber;
    function AddCellFormulaPercent(const aFormula: OWideString; const aFormat: TNumberFormat = 2;
      const aFontStyle: TFontStyles = []; const aBorders: TCellBorders = [cbTop, cbRight, cbBottom, cbLeft]): TExportCellFormulaPercent;
    function AddCellDate(const aDate: TDateTime;
      const aFontStyle: TFontStyles = []; const aBorders: TCellBorders = [cbTop, cbRight, cbBottom, cbLeft]): TExportCellDate;
    function AddCellDateTime(const aDateTime: TDateTime;
      const aFontStyle: TFontStyles = []; const aBorders: TCellBorders = [cbTop, cbRight, cbBottom, cbLeft]): TExportCellDateTime;
    function AddCellDateTimeWS(const aDateTime: TDateTime;
      const aFontStyle: TFontStyles = []; const aBorders: TCellBorders = [cbTop, cbRight, cbBottom, cbLeft]): TExportCellDateTimeWS;
    function AddCellTime(const aTime: TDateTime;
      const aFontStyle: TFontStyles = []; const aBorders: TCellBorders = [cbTop, cbRight, cbBottom, cbLeft]): TExportCellTime;
    function AddCellTimeWS(const aTime: TDateTime;
      const aFontStyle: TFontStyles = []; const aBorders: TCellBorders = [cbTop, cbRight, cbBottom, cbLeft]): TExportCellTimeWS;

    function AddCellCustom(const aCellClass: TExportCellClass;
      const aFontStyle: TFontStyles = []; const aBorders: TCellBorders = [cbTop, cbRight, cbBottom, cbLeft]): TExportCell;
  end;

  TExportWorkSheet = class;
  TExportRows = class(TOwnedList)
  private
    fWorksheet: TExportWorkSheet;
  private
    function GetER(Index: Integer): TExportRow;
    procedure SetER(Index: Integer; const Value: TExportRow);
  protected
    procedure DoCreate; override;
  public
    function GetExportObject: TExportObject;

    property Items[Index: Integer]: TExportRow read GetER write SetER; default;
  end;

  TExportHeader = class;
  TExportHeaders = class;
  TExportWorkSheet = class(TOwnedListObject)
  private
    fTitle: OWideString;
    fRows: TExportRows;

    fHeader: TExportHeaders;
    fFooter: TExportHeaders;
  protected
    procedure DoCreate; override;
    procedure DoDestroy; override;
  public
    function GetExportObject: TExportObject;

    property Rows: TExportRows read fRows;
    property Title: OWideString read fTitle write fTitle;

    property Header: TExportHeaders read fHeader;
    property Footer: TExportHeaders read fFooter;
  public
    function AddRow: TExportRow;
  end;

  TExportWorkSheets = class(TOwnedList)
  private
    fExportObject: TExportObject;
  private
    function GetWS(Index: Integer): TExportWorkSheet;
    procedure SetWS(Index: Integer; const Value: TExportWorkSheet);
  protected
    procedure DoCreate; override;
  public
    function GetRelId(Index: Integer): OWideString;
    function GetId(Index: Integer): OWideString;
  public
    function GetExportObject: TExportObject;

    property Items[Index: Integer]: TExportWorkSheet read GetWS write SetWS; default;
  end;

  TExportHeader = class(TObject)
  private
    fText: OWideString;
    fFont: TFont;
  public
    constructor Create(aExportObject: TExportObject); virtual;
    destructor Destroy; override;
  public
    function DoAsXLSX: OWideString;
    procedure DoAsODS(const aNode, aOfficeAutomaticStyleNode: IXMLElement; const aStyleName: OWideString);
  public
    //USE THESE VARIABLES IN TEXT: [PAGENUMBER], [PAGECOUNT], [SHEETNAME], [DATE], [TIME], [FILENAME]
    property Text: OWideString read fText write fText;
    property Font: TFont read fFont;
  end;

  TExportHeaders = class(TObject)
  private
    fLeft: TExportHeader;
    fCenter: TExportHeader;
    fRight: TExportHeader;
  public
    constructor Create(aExportObject: TExportObject); virtual;
    destructor Destroy; override;
  public
    function Used: Boolean;
    procedure DoAsXLSX(const aNode: IXMLElement);
    procedure DoAsODS(const aNode, aOfficeAutomaticStyleNode: IXMLElement; const aStyleName: OWideString);
  public
    property Left: TExportHeader read fLeft;
    property Center: TExportHeader read fCenter;
    property Right: TExportHeader read fRight;
  end;

  TExportObject = class(TOwnedListObject)
  private
    fWorkSheets: TExportWorkSheets;
    fTitle: OWideString;
    fAuthor: OWideString;
    fApplication: OWideString;
    fDefaultFont: TFont;
    //fFontName: OWideString;//supported only in TExportXLS
    //fFontSize: Integer;
    fActiveWorkSheet: TExportWorkSheet;
    function GetActiveWorkSheet: TExportWorkSheet;//supported only in TExportXLS
    procedure WorkSheetsNotify(Sender: TOwnedList; Item: TOwnedListObject; Action: TListNotification);
  protected
    procedure DoCreate; override;
    procedure DoDestroy; override;

    function DoExport: Boolean; virtual;
  public
    function Export: Boolean;
  public//used for import
    procedure ImportFromXLSXStream(aStream: TStream);
    procedure ImportFromXLSXFile(aFileName: OWideString);
    procedure ImportFromODSStream(aStream: TStream);
    procedure ImportFromODSFile(aFileName: OWideString);
  public
    property WorkSheets: TExportWorkSheets read fWorkSheets;
    property ActiveWorkSheet: TExportWorkSheet read GetActiveWorkSheet write fActiveWorkSheet;
    property Title: OWideString read fTitle write fTitle;
    property Author: OWideString read fAuthor write fAuthor;
    property Application: OWideString read fApplication write fApplication;
    //property FontName: OWideString read fFontName write fFontName;
    //property FontSize: Integer read fFontSize write fFontSize;
    property DefaultFont: TFont read fDefaultFont;
  public
    function AddWorkSheet(const aTitle: OWideString = ''): TExportWorkSheet;
  end;
  TExportClass = class of TExportObject;

  TExportXLSX = class(TExportObject)
  private
    fExportStream: TStream;

    procedure CreateContentTypes(aZip: TAbArchive);
    procedure CreateRels(aZip: TAbArchive);
    procedure CreateWorkBookRels(aZip: TAbArchive);
    procedure CreateWorkBook(aZip: TAbArchive);
    procedure CreateDocPropsCore(aZip: TAbArchive);
    procedure CreateDocPropsApp(aZip: TAbArchive);
    procedure CreateTheme(aZip: TAbArchive);

    procedure CreateContents(aZip: TAbArchive);
    procedure CreateSheet(aZip: TAbArchive; aWorkSheetId: OWideString;
      aWorkSheet: TExportWorkSheet; aSharedStrings: TOWideStrings;
      aStyleSheet: TExportXLSX_StyleSheet; aChain: TExportXLSX_Chain);
  protected
    function DoExport: Boolean; override;

    procedure DoCreate; override;
  public
    property ExportStream: TStream read fExportStream write fExportStream;
  end;

  TExportXLSXFile = class(TExportXLSX)
  private
    fFileName: OWideString;
    fExecuteFileAfterExport: Boolean;
  protected
    function DoExport: Boolean; override;
    procedure DoCreate; override;
  public
    constructor Create2(const aFileName: OWideString; const aExecuteFileAfterExport: Boolean);
  public
    property FileName: OWideString read fFileName write fFileName;
    property ExecuteFileAfterExport: Boolean read fExecuteFileAfterExport write fExecuteFileAfterExport;
  end;

  TExportXLSXSaveDialog = class(TExportXLSXFile)
  protected
    function DoExport: Boolean; override;
  end;

  TExportXLSX_ChainItem = class(TOwnedListObject)
  private
    fCellName: OWideString;//A4, B5 ...
    fWorkSheetId: OWideString;//1, 2, 3 ...
  end;

  TExportXLSX_Chain = class(TOwnedList)
  private
    function GetI(Index: Integer): TExportXLSX_ChainItem;
    procedure SetI(Index: Integer; const Value: TExportXLSX_ChainItem);
  public
    procedure Export(aZip: TAbArchive);
  public
    property Items[Index: Integer]: TExportXLSX_ChainItem read GetI write SetI; default;
  end;

  TExportXLSX_Style = class(TOwnedListObject)
  private
    fFontId: Integer;
    fFillId: Integer;
    fNumFmtId: Integer;
    fBorderId: Integer;
    fAlignment: TAlignment;
    fVAlignment: TVerticalAlignment;
    fWrapText: Boolean;
  end;

  TExportXLSX_Styles = class(TOwnedList)
  private
    function GetI(Index: Integer): TExportXLSX_Style;
    procedure SetI(Index: Integer; const Value: TExportXLSX_Style);
  public
    property Items[Index: Integer]: TExportXLSX_Style read GetI write SetI; default;
  public
    procedure Export(aCellXfsNode: IXMLElement);
  end;

  TExportXLSX_Fonts = class(TObjectList)
  private
    function GetI(Index: Integer): TFont;
    procedure SetI(Index: Integer; const Value: TFont);
  public
    constructor Create;
  public
    property Items[Index: Integer]: TFont read GetI write SetI; default;
  public
    function CreateAddFontUnique(aFont: TFont): Integer;
    procedure Export(aFontsNode: IXMLElement);
  end;

  TExportXLSX_Fill = class(TOwnedListObject)
  private
    fColor: TColor;
  end;

  TExportXLSX_Fills = class(TOwnedList)
  private
    function GetI(Index: Integer): TExportXLSX_Fill;
    procedure SetI(Index: Integer; const Value: TExportXLSX_Fill);
  protected
    procedure DoCreate; override;
  public
    property Items[Index: Integer]: TExportXLSX_Fill read GetI write SetI; default;
  public
    function CreateAddFillUnique(aColor: TColor): Integer;
    procedure Export(aFillsNode: IXMLElement);
  end;

  TExportXLSX_NumberFormat = class(TOwnedListObject)
  private
    fNumberFormatStr: OWideString;
  end;

  TExportXLSX_NumberFormats = class(TOwnedList)
  private
    function GetI(Index: Integer): TExportXLSX_NumberFormat;
    procedure SetI(Index: Integer; const Value: TExportXLSX_NumberFormat);
  public
    property Items[Index: Integer]: TExportXLSX_NumberFormat read GetI write SetI; default;
  public
    function CreateAddNumberFormatUnique(aNumberFormatStr: OWideString): Integer;
    procedure Export(aNumberFormatsNode: IXMLElement);
  end;

  TExportXLSX_Border = class(TOwnedListObject)
  private
    fBorders: TCellBorders;
  end;

  TExportXLSX_Borders = class(TOwnedList)
  private
    function GetI(Index: Integer): TExportXLSX_Border;
    procedure SetI(Index: Integer; const Value: TExportXLSX_Border);
  protected
    procedure DoCreate; override;
  public
    property Items[Index: Integer]: TExportXLSX_Border read GetI write SetI; default;
  public
    function CreateAddBorderUnique(aBorder: TCellBorders): Integer;
    procedure Export(aBordersNode: IXMLElement);
  end;

  TExportXLSX_StyleSheet = class(TObject)
  private
    fFonts: TExportXLSX_Fonts;
    fFills: TExportXLSX_Fills;
    fNumFmts: TExportXLSX_NumberFormats;
    fBorders: TExportXLSX_Borders;
    fCellXfs: TExportXLSX_Styles;
  public
    constructor Create(aDefaultFont: TFont); virtual;
    destructor Destroy; override;
  public
    property Fonts: TExportXLSX_Fonts read fFonts;
    property Fills: TExportXLSX_Fills read fFills;
    property NumFmts: TExportXLSX_NumberFormats read fNumFmts;
    property Borders: TExportXLSX_Borders read fBorders;
    property CellXfs: TExportXLSX_Styles read fCellXfs;
  public
    function CreateAddStyleUnique(aCell: TExportCell): Integer;

    procedure Export(aZip: TAbArchive);
  end;

  TExportXLSX_SharedStrings = class(TOWideStringList)
  public
    procedure Export(aZip: TAbArchive);
  end;

  TExportODS = class(TExportObject)
  private
    fExportStream: TStream;

    procedure CreateMimeType(aZip: TAbArchive);
    procedure CreateStyles(aZip: TAbArchive);
    procedure CreateMeta(aZip: TAbArchive);
    procedure CreateSettings(aZip: TAbArchive);
    procedure CreateManifest(aZip: TAbArchive);

    procedure CreateContent(aZip: TAbArchive);

    procedure ExportWorkSheetStyles(aStylesNode: IXMLElement);
    function ExportWorkSheetsMasterStyles: OWideString;
  protected
    function DoExport: Boolean; override;

    procedure DoCreate; override;
  public
    property ExportStream: TStream read fExportStream write fExportStream;
  end;

  TExportODSFile = class(TExportODS)
  private
    fFileName: OWideString;
    fExecuteFileAfterExport: Boolean;
  protected
    function DoExport: Boolean; override;
    procedure DoCreate; override;
  public
    constructor Create2(const aFileName: OWideString; const aExecuteFileAfterExport: Boolean);
  public
    property FileName: OWideString read fFileName write fFileName;
    property ExecuteFileAfterExport: Boolean read fExecuteFileAfterExport write fExecuteFileAfterExport;
  end;

  TExportODSSaveDialog = class(TExportODSFile)
  protected
    function DoExport: Boolean; override;
  end;

  TExportODS_ColStyle = class(TOwnedListObject)
  private
    fWidth: Integer;
  end;
  TExportODS_ColStyles = class(TOwnedList)
  private
    function GetC(Index: Integer): TExportODS_ColStyle;
    procedure SetC(Index: Integer; const Value: TExportODS_ColStyle);
  public
    property Items[Index: Integer]: TExportODS_ColStyle read GetC write SetC; default;

    function CreateAddUnique(aWidth: Integer): Integer;
    procedure Export(aStylesNode: IXMLElement);
    function StyleName(Index: Integer): OWideString;
  end;

  TExportODS_RowStyle = class(TOwnedListObject)
  private
    fHeight: Integer;
  end;
  TExportODS_RowStyles = class(TOwnedList)
  private
    function GetC(Index: Integer): TExportODS_RowStyle;
    procedure SetC(Index: Integer; const Value: TExportODS_RowStyle);
  public
    property Items[Index: Integer]: TExportODS_RowStyle read GetC write SetC; default;

    function CreateAddUnique(aHeight: Integer): Integer;
    procedure Export(aStylesNode: IXMLElement);
    function StyleName(Index: Integer): OWideString;
  end;

  TExportODS_NumberStyle = class(TOwnedListObject)
  private
    fXLSXNumberFormat: OWideString;
    fParentCell: TExportCellCustomNumber;
  end;

  TExportODS_NumberStyles = class(TOwnedList)
  private
    function GetC(Index: Integer): TExportODS_NumberStyle;
    procedure SetC(Index: Integer; const Value: TExportODS_NumberStyle);
  public
    property Items[Index: Integer]: TExportODS_NumberStyle read GetC write SetC; default;

    function CreateAddNumberStyleUnique(aCell: TExportCellCustomNumber): Integer;
    procedure Export(aStylesNode: IXMLElement);
    function StyleName(Index: Integer): OWideString;
  end;

  TExportODS_CellStyle = class(TOwnedListObject)
  private
    fFont: TFont;
    fBorders: TCellBorders;
    fBGColor: TColor;
    fAlignment: TAlignment;
    fVAlignment: TVerticalAlignment;
    fWrapText: Boolean;
    fNumberStyleId: Integer;
  protected
    procedure DoCreate; override;
    procedure DoDestroy; override;
  end;
  TExportODS_CellStyles = class(TOwnedList)
  private
    function GetC(Index: Integer): TExportODS_CellStyle;
    procedure SetC(Index: Integer; const Value: TExportODS_CellStyle);
  public
    property Items[Index: Integer]: TExportODS_CellStyle read GetC write SetC; default;

    function CreateAddStyleUnique(aCell: TExportCell;
      aNumberStyles: TExportODS_NumberStyles): Integer;
    procedure Export(aStylesNode, aFontFaceDeclsNode: IXMLElement;
      aNumberStyles: TExportODS_NumberStyles);
    function StyleName(Index: Integer): OWideString;
  end;



  TCustomGenerateExport = class;

  TExportHeadProcEx = procedure (Sender: TCustomGenerateExport) of Object;
  TExportRowProcEx = procedure(Sender: TCustomGenerateExport; const Row: Integer) of Object;

  IOCustomProgressDlg = interface
  ['{05377B31-EEB3-4123-A03B-8532D2729772}']
    procedure IncPosition;
  end;

  TCustomGenerateExport = class(TObject)
  private
    fTitle: OWideString;
    fRowCount: Integer;
    fTitleColSpan: Integer;
    fExportObject: TExportObject;

    fOnExportHead: TExportHeadProcEx;
    fOnExportFoot: TExportHeadProcEx;
  protected
    procedure DoCreate; virtual;
    procedure DoExecute(aClearRows: Boolean;
      aFormProcess: IOCustomProgressDlg); virtual;
    procedure DoExportRow(const Row: Integer); virtual;

  public
    constructor Create; overload;
    constructor Create(aExportClass: TExportClass); overload;
  public
    procedure Execute(aClearRows: Boolean = true;
      aFormProcess: IOCustomProgressDlg = nil);
    procedure ExecuteExport;

    procedure DefaultExportHead(Sender: TCustomGenerateExport);
    procedure DefaultExportFoot(Sender: TCustomGenerateExport);
  public
    property Title: OWideString read fTitle write fTitle;
    property RowCount: Integer read fRowCount write fRowCount;
    property TitleColSpan: Integer read fTitleColSpan write fTitleColSpan;
    property ExportObject: TExportObject read fExportObject write fExportObject;

    property OnExportHead: TExportHeadProcEx read fOnExportHead write fOnExportHead;
    property OnExportFoot: TExportHeadProcEx read fOnExportFoot write fOnExportFoot;
  end;

  TGenerateExportRows = class(TCustomGenerateExport)
  private
    fOnExportRow: TExportRowProcEx;
  protected
    procedure DoExecute(aClearRows: Boolean;
      aFormProcess: IOCustomProgressDlg); override;
    procedure DoCreate; override;
    procedure DoExportRow(const Row: Integer); override;
  public
    property OnExportRow: TExportRowProcEx read fOnExportRow write fOnExportRow;
  end;

const
  DigitFormatDigits: Array[TDigitFormat] of Integer = (0, 1, 2, 3, 0, 1);
  DigitFormatCSS: Array[TDigitFormat] of OWideString = ('no0', 'no1', 'no2', 'no3', 'per0', 'per1');

  EXP_HEADER_ROW = -1;
  EXP_FOOTER_ROW = -2;

  DefExportPageWidth = 1000;

  OOoCalcSizeParameter = 19.8;

const
  CrLF = #13#10;
  XLSExportHeader =
    '<html xmlns:o="urn:schemas-microsoft-com:office:office" xmlns:x="urn:schemas-microsoft-com:office:excel" xmlns="http://www.w3.org/TR/REC-html40">'+CrLF+
    '<head>'+CrLF+
      '<meta http-equiv="Content-Type" content="text/html; charset=utf-8">'+CrLF+
      '<style>'+CrLF+
        'table{mso-displayed-decimal-separator:"\,"; mso-displayed-thousand-separator:" "; font-family: "{FONTNAME}"; font-size:{FONTSIZE}pt;} '+CrLF+ //For Excel
        'td{border: 0.5pt windowtext solid} '+CrLF+
        //'td{border: 0} '+CrLF+
        '.no0{mso-number-format:0} '+CrLF+
        '.no1{mso-number-format:"0\.0"} '+CrLF+
        '.no2{mso-number-format:"0\.00"} '+CrLF+
        '.no3{mso-number-format:"0\.000"} '+CrLF+
        '.per0{mso-number-format:"0%"} '+CrLF+
        '.per1{mso-number-format:"0\.0%"} '+CrLF+
        '.per2{mso-number-format:"0\.00%"} '+CrLF+
        '.per3{mso-number-format:"0\.000%"} '+CrLF+
        '.text{mso-number-format:"\@"} '+CrLF+
        //'.noborder, .noborder TD{border: 0} '+CrLF+
        (*'.ncbTop, .ncbTop TD{border-top: 0} '+CrLF+
        '.ncbRight, .ncbRight TD{border-right: 0} '+CrLF+
        '.ncbBottom, .ncbBottom TD{border-bottom: 0} '+CrLF+
        '.ncbLeft, .ncbLeft TD{border-left: 0} '+CrLF+*)
      '</style>'+CrLF+
      '<!--[if gte mso 9]><xml><x:ExcelWorkbook><x:ExcelWorksheets><x:ExcelWorksheet><x:Name>{TITLE}</x:Name>'+'<x:WorksheetOptions><x:Selected/><x:ProtectContents>True</x:ProtectContents></x:WorksheetOptions></x:ExcelWorksheet></x:ExcelWorksheets></x:ExcelWorkbook></xml><![endif]-->'+CrLF+
    '</head>'+CrLF+
    '<body>'+CrLF+
    '<table>'+CrLF+
    '<!--StartFragment-->'+CrLF;
  XLSExportFooter = '<!--EndFragment-->'+CrLF+'</table>'+CrLF+'</body>'+CrLF+'</html>';
  ExcelNumberFormat: Array[TNumberFormat] of OWideString = ('General', '0', '0,0', '0,00', '0,000');
  XLSXNumberFormat: Array[TNumberFormat] of OWideString = ('', '0', '0.0', '0.00', '0.000');
  OOoCalcNumberFormat: Array[TNumberFormat] of OWideString = ('', '0', '0,0', '0,00', '0,000');
  ExcelNumberFormatCSS: Array[TNumberFormat] of OWideString = ('General', '0', '0\.0', '0\.00', '0\.000');

function GetXLSExportHeader(aTitle: OWideString; aFontName: OWideString; aFontSize: Integer): OWideString;
function GetExcelNumberFormat(aFormat: TNumberFormat; aPostFix: OWideString): OWideString;
function GetOOoCalcNumberFormat(aFormat: TNumberFormat; aPostFix: OWideString): OWideString;
function GetExcelNumberFormatCSS(aFormat: TNumberFormat; aPostFix: OWideString): OWideString;
function GetXLSXNumberFormat(aFormat: TNumberFormat; aPostFix: OWideString): OWideString;
function XLSNumberCell(const Number: Extended; const DigitFormat: TDigitFormat; const CSSStyle: OWideString; const Params: OWideString): OWideString;
function XLSFormulaCell(const Number: Extended; const Formula: OWideString; const DigitFormat: TDigitFormat; const CSSStyle: OWideString; const Params: OWideString): OWideString;

procedure ReadXLSXContentTypes(aZip: TAbArchive; aWorkSheets: TOWideStrings; var aSharedStrings: OWideString);
procedure ReadDOCXContentTypes(aZip: TAbArchive; var aDocument: OWideString);

function ExcelRange(const Col, Row: Integer; const ColSpan: Integer = 0; const RowSpan: Integer = 0): String;
function ExcelColToInt(aCol: String): Integer;

procedure FontToODS(const aFont: TFont; const aToNode: IXMLElement);

implementation

uses SysUtils,
  {$IFDEF O_DELPHI_XE2}
  Vcl.Dialogs,
  {$ELSE}
  Dialogs,
  {$ENDIF}
  {$IFDEF O_USE_TNT}TntDialogs,{$ENDIF} UtfConversion, OOmniXML,
  {$IFDEF O_DELPHI_2007UP}WideStrUtils, Windows, {$ENDIF}Math, CursorFunc, FormatSQL, oExportLNG,
  AbZipKit, OZip, AbUnzper, oFileUtils, StrUtils;


function ExcelColToInt(aCol: String): Integer;
var I, xA, xCharCount: Integer;
begin
  aCol := UpperCase(aCol);
  //delete ":"
  I := Pos(':', aCol);
  if I > 0 then
    Delete(aCol, I, Length(aCol));

  //delete all non-alpha characters from the string
  for I := Length(aCol) downto 1 do
  if not (AnsiChar(aCol[I]) in ['A'..'Z']) then
    Delete(aCol, I, 1);

  xA := Ord('A');
  xCharCount := Ord('Z')-xA+1;

  Result := 0;
  for I := 1 to Length(aCol) do begin
    Result := Result + (Ord(aCol[I])-xA+1)*Round(IntPower(xCharCount, (Length(aCol)-I)));
  end;

  Dec(Result);
end;

function ExcelRange(const Col, Row, ColSpan, RowSpan: Integer): String;
var
  xCol1, xCol2, xA: Integer;
  xCharCount: Integer;
begin
  Result := '';

  xA := Ord('A');
  xCharCount := Ord('Z')-xA+1;

  //Column
  xCol1 := Col div xCharCount;
  xCol2 := Col mod xCharCount;
  if xCol1 > 0 then
    Result := Result + Char(xA+xCol1-1);
  Result := Result + Char(xA+xCol2);
  //Row
  Result := Result + IntToStr(Row+1);

  //Span
  if (ColSpan > 1) or (RowSpan > 1) then begin
    Result := Result + ':';

    //Column
    xCol1 := (Col+ColSpan-1) div xCharCount;
    xCol2 := (Col+ColSpan-1) mod xCharCount;
    if xCol1 > 0 then
      Result := Result + Char(xA+xCol1-1);
    Result := Result + Char(xA+xCol2);
    //Row
    Result := Result + IntToStr(Row+RowSpan);
  end;
end;

procedure FontToODS(const aFont: TFont; const aToNode: IXMLElement);
begin
  aToNode.SetAttribute('style:font-name', aFont.Name);
  aToNode.SetAttribute('fo:color', ColorToHTML(aFont.Color));
  aToNode.SetAttribute('fo:font-size', IntToStr(aFont.Size)+'pt');
  if fsItalic in aFont.Style then
    aToNode.SetAttribute('fo:font-style', 'italic');
  if fsBold in aFont.Style then
    aToNode.SetAttribute('fo:font-weight', 'bold');
  if fsUnderline in aFont.Style then
    aToNode.SetAttribute('style:text-underline-style', 'solid');
  if fsStrikeOut in aFont.Style then
    aToNode.SetAttribute('style:text-line-through-style', 'solid');
end;

procedure ReadXLSXContentTypes(aZip: TAbArchive; aWorkSheets: TOWideStrings; var aSharedStrings: OWideString);
var
  xXML: IXMLDocument;
  xOverrideNodes: IXMLNodeList;
  I: Integer;
  xFileName, xContentType: OWideString;
  xMS: TMemoryStream;
begin
  xXML := CreateXMLDoc;
  xMS := TMemoryStream.Create;
  try
    xMS.Clear;
    aZip.ExtractToStream('[Content_Types].xml', xMS);
    xMS.Position := 0;

    xXML.LoadFromStream(xMS);
    xOverrideNodes := xXML.SelectNodes('Types/Override');
    if Assigned(xOverrideNodes) then begin
      for I := 0 to xOverrideNodes.Length-1 do
      if Supports(xOverrideNodes.Item[I], IXMLElement) then
      with (xOverrideNodes.Item[I] as IXMLElement) do begin
        xContentType := GetAttribute('ContentType');
        if Assigned(aWorkSheets) and AnsiSameStr(xContentType, 'application/vnd.openxmlformats-officedocument.spreadsheetml.worksheet+xml')
        then begin
          xFileName := GetAttribute('PartName');
          if (xFileName <> '') and (AnsiChar(xFileName[1]) in ['\', '/']) then
            Delete(xFileName, 1, 1);
          aWorkSheets.Add(xFileName);
        end else if AnsiSameStr(xContentType, 'application/vnd.openxmlformats-officedocument.spreadsheetml.sharedStrings+xml')
        then begin
          aSharedStrings := GetAttribute('PartName');
          if (aSharedStrings <> '') and (AnsiChar(aSharedStrings[1]) in ['\', '/']) then
            Delete(aSharedStrings, 1, 1);
        end;
      end;
    end;
  finally
    xMS.Free;
    xXML := nil;
  end;
end;

procedure ReadDOCXContentTypes(aZip: TAbArchive; var aDocument: OWideString);
var
  xXML: IXMLDocument;
  xOverrideNodes: IXMLNodeList;
  I: Integer;
  xFileName: OWideString;
  xMS: TMemoryStream;
begin
  xXML := CreateXMLDoc;
  xMS := TMemoryStream.Create;
  try
    xMS.Clear;
    aZip.ExtractToStream('[Content_Types].xml', xMS);
    xMS.Position := 0;

    xXML.LoadFromStream(xMS);
    xOverrideNodes := xXML.SelectNodes('Types/Override');
    if Assigned(xOverrideNodes) then begin
      for I := 0 to xOverrideNodes.Length-1 do
      if Supports(xOverrideNodes.Item[I], IXMLElement) then
      with (xOverrideNodes.Item[I] as IXMLElement) do
      if AnsiSameStr(GetAttribute('ContentType'), 'application/vnd.openxmlformats-officedocument.wordprocessingml.document.main+xml')
      then begin
        xFileName := GetAttribute('PartName');
        if (xFileName <> '') and (AnsiChar(xFileName[1]) in ['\', '/']) then
          Delete(xFileName, 1, 1);
        aDocument := xFileName;
        break;
      end;
    end;
  finally
    xMS.Free;
    xXML := nil;
  end;
end;

function GetXLSExportHeader(aTitle: OWideString; aFontName: OWideString; aFontSize: Integer): OWideString;
begin
  Result := XLSExportHeader;
  Result := OWideStringReplace(Result, '{TITLE}', StringToWideHTML(aTitle), [rfReplaceAll, rfIgnoreCase]);
  Result := OWideStringReplace(Result, '{FONTNAME}', StringToWideHTML(aFontName), [rfReplaceAll, rfIgnoreCase]);
  Result := OWideStringReplace(Result, '{FONTSIZE}', IntToStr(aFontSize), [rfReplaceAll, rfIgnoreCase]);
end;

function GetExcelNumberFormat(aFormat: TNumberFormat; aPostFix: OWideString): OWideString;
begin
  Result := ExcelNumberFormat[aFormat];
  if aFormat >= 0 then
    Result := Result + aPostFix;
end;
function GetOOoCalcNumberFormat(aFormat: TNumberFormat; aPostFix: OWideString): OWideString;
begin
  Result := OOoCalcNumberFormat[aFormat];
  if aFormat >= 0 then
    Result := Result + aPostFix;
end;
function GetExcelNumberFormatCSS(aFormat: TNumberFormat; aPostFix: OWideString): OWideString;
begin
  Result := ExcelNumberFormatCSS[aFormat];
  if aFormat >= 0 then
    Result := Result + aPostFix;
end;
function GetXLSXNumberFormat(aFormat: TNumberFormat; aPostFix: OWideString): OWideString;
begin
  Result := XLSXNumberFormat[aFormat];
  if aFormat >= 0 then
    Result := Result + aPostFix;
end;

function XLSNumberCell(const Number: Extended; const DigitFormat: TDigitFormat; const CSSStyle: OWideString; const Params: OWideString): OWideString;
var
  FS, FSDisplay: TFormatSettings;
begin
  GetLocaleFormatSettings(0, FS);
  GetLocaleFormatSettings(0, FSDisplay);
  FS.DecimalSeparator := '.';
  FS.ThousandSeparator := #0;
  FSDisplay.DecimalSeparator := ',';
  FSDisplay.ThousandSeparator := ' ';
  Result := '<td class="' + DigitFormatCSS[DigitFormat];
  if (length(CSSStyle) > 0) then
    Result := Result + ' ';
  Result := Result + CSSStyle + '" ';
  if (length(Params) > 0) then
    Result := Result + Params + ' ';
  Result := Result + 'x:num="' + FloatToStr(Number, FS) + '">' + FloatToStrF(Number, ffFixed, 18, DigitFormatDigits[DigitFormat], FSDisplay) + '</td>';
end;

function XLSFormulaCell(const Number: Extended; const Formula: OWideString; const DigitFormat: TDigitFormat; const CSSStyle: OWideString; const Params: OWideString): OWideString;
var
  FS, FSDisplay: TFormatSettings;
begin
  GetLocaleFormatSettings(0, FS);
  GetLocaleFormatSettings(0, FSDisplay);
  FS.DecimalSeparator := '.';
  FS.ThousandSeparator := #0;
  FSDisplay.DecimalSeparator := ',';
  FSDisplay.ThousandSeparator := ' ';
  Result := '<td class="' + DigitFormatCSS[DigitFormat];
  if (length(CSSStyle) > 0) then
    Result := Result + ' ';
  Result := Result + CSSStyle + '" ';
  if (length(Params) > 0) then
    Result := Result + Params + ' ';
  Result := Result + 'x:num="' + FloatToStr(Number, FS) + '" x:fmla="' + Formula + '">' +
    //FloatToStrF(Number, ffFixed, 18, DigitFormatDigits[DigitFormat], FSDisplay) +
    OWideStringReplace(Formula, '.', ',', [rfReplaceAll, rfIgnoreCase])+ //In Order to function in OpenOffice
    '</td>';
end;

{ TExportRow }

function TExportRow.AddCellCustom(const aCellClass: TExportCellClass;
  const aFontStyle: TFontStyles; const aBorders: TCellBorders): TExportCell;
begin
  Result := aCellClass.CreateAdd(Self.Cells);
  with Result do begin
    Font.Style := aFontStyle;
    Borders := aBorders;
  end;
end;

function TExportRow.AddCellDate(const aDate: TDateTime;
  const aFontStyle: TFontStyles; const aBorders: TCellBorders): TExportCellDate;
begin
  Result := TExportCellDate(AddCellCustom(TExportCellDate, aFontStyle, aBorders));
  with Result do begin
    DateTime := aDate;
  end;
end;

function TExportRow.AddCellDateTime(const aDateTime: TDateTime;
  const aFontStyle: TFontStyles; const aBorders: TCellBorders): TExportCellDateTime;
begin
  Result := TExportCellDateTime(AddCellCustom(TExportCellDateTime, aFontStyle, aBorders));
  with Result do begin
    DateTime := aDateTime;
  end;
end;

function TExportRow.AddCellDateTimeWS(const aDateTime: TDateTime;
  const aFontStyle: TFontStyles; const aBorders: TCellBorders): TExportCellDateTimeWS;
begin
  Result := TExportCellDateTimeWS(AddCellCustom(TExportCellDateTimeWS, aFontStyle, aBorders));
  with Result do begin
    DateTime := aDateTime;
  end;
end;

function TExportRow.AddCellFormulaNumber(const aFormula: OWideString;
  const aFormat: TNumberFormat; const aFontStyle: TFontStyles;
  const aBorders: TCellBorders): TExportCellFormulaNumber;
begin
  Result := TExportCellFormulaNumber(AddCellCustom(TExportCellFormulaNumber, aFontStyle, aBorders));
  with Result do begin
    Formula := aFormula;
    Format := aFormat;
  end;
end;

function TExportRow.AddCellFormulaPercent(const aFormula: OWideString;
  const aFormat: TNumberFormat; const aFontStyle: TFontStyles;
  const aBorders: TCellBorders): TExportCellFormulaPercent;
begin
  Result := TExportCellFormulaPercent(AddCellCustom(TExportCellFormulaPercent, aFontStyle, aBorders));
  with Result do begin
    Formula := aFormula;
    Format := aFormat;
  end;
end;

function TExportRow.AddCellNumber(const aValue: Double;
  const aFormat: TNumberFormat; const aFontStyle: TFontStyles;
  const aBorders: TCellBorders): TExportCellNumber;
begin
  Result := TExportCellNumber(AddCellCustom(TExportCellNumber, aFontStyle, aBorders));
  with Result do begin
    Value := aValue;
    Format := aFormat;
  end;
end;

function TExportRow.AddCellPercent(const aValue: Double;
  const aFormat: TNumberFormat; const aFontStyle: TFontStyles;
  const aBorders: TCellBorders): TExportCellPercent;
begin
  Result := TExportCellPercent(AddCellCustom(TExportCellPercent, aFontStyle, aBorders));
  with Result do begin
    Value := aValue;
    Format := aFormat;
  end;
end;

function TExportRow.AddCellString(const aText: OWideString;
  const aFontStyle: TFontStyles;
  const aBorders: TCellBorders): TExportCellString;
begin
  Result := TExportCellString(AddCellCustom(TExportCellString, aFontStyle, aBorders));
  with Result do begin
    Text := aText;
  end;
end;

function TExportRow.AddCellTime(const aTime: TDateTime;
  const aFontStyle: TFontStyles; const aBorders: TCellBorders): TExportCellTime;
begin
  Result := TExportCellTime(AddCellCustom(TExportCellTime, aFontStyle, aBorders));
  with Result do begin
    DateTime := aTime;
  end;
end;

function TExportRow.AddCellTimeWS(const aTime: TDateTime;
  const aFontStyle: TFontStyles;
  const aBorders: TCellBorders): TExportCellTimeWS;
begin
  Result := TExportCellTimeWS(AddCellCustom(TExportCellTimeWS, aFontStyle, aBorders));
  with Result do begin
    DateTime := aTime;
  end;
end;

function TExportRow.AsHTML: OWideString;
var I: Integer;
begin
  result := '<tr>';
  for I := 0 to Cells.Count-1 do begin
    result := result + TExportCell(Cells[I]).AsHTML;
  end;
  result := result + '</tr>';
end;

procedure TExportRow.DoCreate;
begin
  inherited;

  fCells := TExportCells.Create;
  fCells.fExportRow := Self;
end;

procedure TExportRow.DoDestroy;
begin
  fCells.Free;

  inherited;
end;

function TExportRow.GetExportObject: TExportObject;
begin
  if OwnerList is TExportRows then
    Result := TExportRows(OwnerList).GetExportObject
  else
    Result := nil;
end;

{ TExportObject }

function TExportObject.AddWorkSheet(
  const aTitle: OWideString): TExportWorkSheet;
begin
  Result := TExportWorkSheet.CreateAdd(Self.WorkSheets);
  if aTitle <> '' then
    Result.Title := aTitle;
  ActiveWorkSheet := Result;
end;

procedure TExportObject.DoCreate;
begin
  inherited;
  fWorkSheets := TExportWorkSheets.Create;
  fWorkSheets.fExportObject := Self;
  fWorkSheets.OnNotify := WorkSheetsNotify;

  fTitle := '';
  fAuthor := '';
  fApplication := '';
  fDefaultFont := TFont.Create;
  fDefaultFont.Size := 10;
  fDefaultFont.Name := 'Tahoma';
  //fFontSize := 10;
  //fFontName := 'Tahoma';
end;

procedure TExportObject.DoDestroy;
begin
  fWorkSheets.Free;

  inherited;
end;

function TExportObject.DoExport: Boolean;
begin
  //nothing
  Result := false;
end;

function TExportObject.Export: Boolean;
begin
  Result := DoExport;
end;

function TExportObject.GetActiveWorkSheet: TExportWorkSheet;
begin
  if not Assigned(fActiveWorkSheet) then begin
    if fWorkSheets.Count = 0 then
      TExportWorkSheet.CreateAdd(fWorkSheets);

    fActiveWorkSheet := fWorkSheets[fWorkSheets.Count-1];
  end;
  Result := fActiveWorkSheet;
end;

procedure TExportObject.ImportFromODSFile(aFileName: OWideString);
var xFS: TOFileStream;
begin
  xFS := TOFileStream.Create(aFileName, fmOpenRead or fmShareDenyNone);
  try
    ImportFromODSStream(xFS);
  finally
    xFS.Free;
  end;
end;

procedure TExportObject.ImportFromODSStream(aStream: TStream);
  procedure AddWorkSheet(aWorkSheet: IXMLElement);
  var
    xRowNodes, xCellNodes: IXMLNodeList;
    xRow: IXMLNode;
    xText, xTextParts: IXMLNodeList;
    xCell: IXMLElement;
    xExpWS: TExportWorkSheet;
    xExpRow: TExportRow;
    xExpCell: TExportCell;
    IRow, ICell, xRepeat, IRepeat, sR: Integer;
    xValueType: OWideString;
  begin
    xExpWS := TExportWorkSheet.CreateAdd(WorkSheets);
    xExpWS.Title := aWorkSheet.GetAttribute('table:name');
    xRowNodes := aWorkSheet.SelectNodes('table:table-row');
    for IRow := 0 to xRowNodes.Length-1 do begin
      xRow := xRowNodes.Item[IRow];
      //if xRow.ChildNodes.Length <= 1 then
      //  continue;//the last cell is always "table:number-columns-repeated="1024"... ignore them!
      xExpRow := TExportRow.CreateAdd(xExpWS.Rows);

      xCellNodes := xRow.SelectNodes('table:table-cell');
      for ICell := 0 to xCellNodes.Length-1 do
      if Supports(xCellNodes.Item[ICell], IXMLElement) then begin
        xCell := (xCellNodes.Item[ICell] as IXMLElement);

        xExpCell := nil;
        xValueType := xCell.GetAttribute('office:value-type');
        if AnsiSameStr(xValueType, 'float') or AnsiSameStr(xValueType, 'percentage') then begin
          xExpCell := TExportCellNumber.CreateAdd(xExpRow.Cells);
          TExportCellNumber(xExpCell).Value := SqlStrToFloat(xCell.GetAttribute('office:value'));
        end else if AnsiSameStr(xValueType, 'date') then begin
          xExpCell := TExportCellDateTime.CreateAdd(xExpRow.Cells);
          TExportCellDateTime(xExpCell).Value := ODSStrToDateTime(xCell.GetAttribute('office:date-value'));
        end else if AnsiSameStr(xValueType, 'time') then begin
          xExpCell := TExportCellTime.CreateAdd(xExpRow.Cells);
          TExportCellTime(xExpCell).Value := ODSStrToTime(xCell.GetAttribute('office:time-value'));
        end else {if AnsiSameStr(xValueType, 'string') then} begin//empty cells do not have a type -> just handle them as String cells with no "text:p" node
          xCell.SelectNodes('text:p', xText);

          if Assigned(xText) or not (ICell = xCellNodes.Length-1) then begin//delete last empty cell
            xExpCell := TExportCellString.CreateAdd(xExpRow.Cells);
            TExportCellString(xExpCell).Text := '';
            if Assigned(xText) then
            for IRepeat := 0 to xText.Length-1 do begin
              if IRepeat > 0 then
                TExportCellString(xExpCell).Text := TExportCellString(xExpCell).Text + CrLF;
              if Supports(xCellNodes.Item[ICell], IXMLElement) then begin
                xTextParts := (xText.Item[IRepeat] as IXMLElement).ChildNodes;
                for xRepeat := 0 to xTextParts.Length-1 do begin
                  if (Supports(xTextParts.Item[xRepeat], IXMLElement)) and (xTextParts.Item[xRepeat].NodeName = 'text:s') then begin
                    for sR := 1 to SqlStrToIntDef((xTextParts.Item[xRepeat] as IXMLElement).GetAttribute('text:c'), 1) do
                      TExportCellString(xExpCell).Text := TExportCellString(xExpCell).Text + ' ';
                  end else begin
                    TExportCellString(xExpCell).Text := TExportCellString(xExpCell).Text + xTextParts.Item[xRepeat].Text;
                  end;
                end;
              end;
            end;
          end;
        end;

        if not (ICell = xCellNodes.Length-1) and Assigned(xExpCell) then begin
          xRepeat := SqlStrToInt(xCell.GetAttribute('table:number-columns-repeated'));
          for IRepeat := 2 to xRepeat do
          with TExportCellClass(xExpCell.ClassType).CreateAdd(xExpRow.Cells) do
            Assign(xExpCell);
        end;
      end;
    end;
  end;

  procedure ReadContent(aStream: TStream);
  var
    xXML: IXMLDocument;
    xTableNodes: IXMLNodeList;
    I: Integer;
  begin
    xXML := CreateXMLDoc;
    try
      xXML.LoadDocumentWithPreserveWhiteSpace := True;
      xXML.LoadFromStream(aStream);

      xTableNodes := xXML.SelectNodes('office:document-content/office:body/office:spreadsheet/table:table');
      for I := 0 to xTableNodes.Length-1 do
      if Supports(xTableNodes.Item[I], IXMLElement) then begin
        AddWorkSheet(xTableNodes.Item[I] as IXMLElement);
      end;
    finally
      xXML := nil;
    end;
  end;
var
  xZip: TOZipArchive;
  //xZip: TAbUnZipper;
  xMS: TMemoryStream;
begin
  xZip := TOZipArchive.CreateFromStream(aStream, '');
  //xZip := TAbUnZipper.Create(nil);
  //xZip.FileName := 'c:\Users\Ondrej Pokorny\Documents\a\Import2.xlsx';
  //xZip := TAbUnZipper.Create(nil);
  xMS := TMemoryStream.Create;
  try
    xZip.Load;
    xMS.Clear;
    xZip.ExtractToStream('content.xml', xMS);
    xMS.Position := 0;
    WorkSheets.Clear;
    ReadContent(xMS);
  finally
    xZip.Free;
    xMS.Free;
  end;
end;

procedure TExportObject.ImportFromXLSXFile(aFileName: OWideString);
var xFS: TOFileStream;
begin
  xFS := TOFileStream.Create(aFileName, fmOpenRead or fmShareDenyNone);
  try
    xFS.Position := 0;
    ImportFromXLSXStream(xFS);
  finally
    xFS.Free;
  end;
end;

procedure TExportObject.ImportFromXLSXStream(aStream: TStream);
  procedure LoadSharedStrings(aStream: TStream; aSharedStrings: TOWideStrings);
  var
    xXML: IXMLDocument;
    xStringNodes: IXMLNodeList;
    I: Integer;
  begin
    xXML := CreateXMLDoc;
    try
      xXML.LoadFromStream(aStream);
      xStringNodes := xXML.SelectNodes('sst/si');
      for I := 0 to xStringNodes.Length-1 do begin
        aSharedStrings.Add(xStringNodes.Item[I].SelectSingleNode('t').Text);
      end;
    finally
      xXML := nil;
      //(xXML as TXMLDocument).Free;
    end;
  end;

  procedure AddWorkSheet(aStream: TStream; aSharedStrings: TOWideStrings);
  var
    xWS: TExportWorkSheet;
    xRow: TExportRow;
    xCell: TExportCell;
    xXML: IXMLDocument;
    xRowNodes: IXMLNodeList;
    xRowNode, xValueNode: IXMLNode;
    xCellNode: IXMLElement;
    I, L, xColI, xRowI, xStringValueIndex: Integer;
  begin
    xWS := TExportWorkSheet.CreateAdd(Self.WorkSheets);
    xXML := CreateXMLDoc;
    try
      xXML.LoadFromStream(aStream);
      xRowNodes := xXML.SelectNodes('worksheet/sheetData/row');
      for I := 0 to xRowNodes.Length-1 do begin
        xRowNode := xRowNodes.Item[I];
        if not Supports(xRowNode, IXMLElement) then continue;
        xRowI := SqlStrToInt((xRowNode as IXMLElement).GetAttribute('r'));//xRowI starts from 1
        if xRowI < 1 then begin//error
          xRow := TExportRow.CreateAdd(xWS.Rows);
        end else begin
          while (xRowI > xWS.Rows.Count) do
            TExportRow.CreateAdd(xWS.Rows);
          xRow := xWS.Rows[xRowI-1];
        end;
        for L := 0 to xRowNode.ChildNodes.Length-1 do
        if Supports(xRowNode.ChildNodes.Item[L], IXMLElement) then begin
          xCellNode := (xRowNode.ChildNodes.Item[L] as IXMLElement);

          xColI := ExcelColToInt(xCellNode.GetAttribute('r'));
          while (xColI > xRow.Cells.Count) do
            TExportCellString.CreateAdd(xRow.Cells);

          xValueNode := xCellNode.SelectSingleNode('v');
          if AnsiSameStr(xCellNode.GetAttribute('t'), 's') then begin
            xCell := TExportCellString.CreateAdd(xRow.Cells);
            if Assigned(xValueNode) then
              xStringValueIndex := StrToIntDef(xValueNode.Text, -1)
            else
              xStringValueIndex := -1;
            if (0 <= xStringValueIndex) and (xStringValueIndex <= aSharedStrings.Count-1) then
              TExportCellString(xCell).Text := aSharedStrings[xStringValueIndex];
          end else begin
            xCell := TExportCellNumber.CreateAdd(xRow.Cells);
            if Assigned(xValueNode) then
              TExportCellNumber(xCell).Value := SqlStrToFloat(xValueNode.Text);
          end;
        end;
      end;
    finally
      xXML := nil;
    end;
  end;

  procedure LoadSheetTitles(aStream: TStream);
  var
    xXML: IXMLDocument;
    xSheetNodes: IXMLNodeList;
    xSheet: IXMLElement;
    I, xSheetId: Integer;
  begin
    xXML := CreateXMLDoc;
    try
      xXML.LoadFromStream(aStream);
      xSheetNodes := xXML.SelectNodes('workbook/sheets/sheet');
      for I := 0 to xSheetNodes.Length-1 do
      if Supports(xSheetNodes.Item[I], IXMLElement) then begin
        xSheet := (xSheetNodes.Item[I] as IXMLElement);
        xSheetId := SqlStrToInt(xSheet.GetAttribute('sheetId'))-1;
        if (xSheetId >= 0) and (xSheetId < Self.WorkSheets.Count) then begin
          Self.WorkSheets[xSheetId].Title := xSheet.GetAttribute('name');
        end;
      end;
    finally
      xXML := nil;
      //(xXML as TXMLDocument).Free;
    end;
  end;
var
  xZip: TOZipArchive;
  //xZip: TAbUnZipper;
  xMS: TMemoryStream;
  xWorkSheets, xSharedStrings: TOWideStringList;
  xSharedStringsFile: OWideString;
  I: Integer;
begin
  xZip := TOZipArchive.CreateFromStream(aStream, '');
  //xZip := TAbUnZipper.Create(nil);
  //xZip.FileName := 'c:\Users\Ondrej Pokorny\Documents\a\Import2.xlsx';
  //xZip := TAbUnZipper.Create(nil);
  xMS := TMemoryStream.Create;
  xSharedStrings := TOWideStringList.Create;
  xWorkSheets := TOWideStringList.Create;
  try
    xZip.Load;
    ReadXLSXContentTypes(xZip, xWorkSheets, xSharedStringsFile);

    xMS.Clear;
    xZip.ExtractToStream('xl/sharedStrings.xml'{xSharedStringsFile}, xMS);
    xMS.Position := 0;
    LoadSharedStrings(xMS, xSharedStrings);

    WorkSheets.Clear;
    xWorkSheets.Sort;
    for I := 0 to xWorkSheets.Count-1 do begin
      xMS.Clear;
      xZip.ExtractToStream(xWorkSheets[I], xMS);
      xMS.Position := 0;
      AddWorkSheet(xMS, xSharedStrings);
    end;

    xMS.Clear;
    xZip.ExtractToStream('xl/workbook.xml', xMS);
    xMS.Position := 0;
    LoadSheetTitles(xMS);
  finally
    xZip.Free;
    xMS.Free;
    xSharedStrings.Free;
    xWorkSheets.Free;
  end;
end;

procedure TExportObject.WorkSheetsNotify(Sender: TOwnedList;
  Item: TOwnedListObject; Action: TListNotification);
begin
  if Item = fActiveWorkSheet then
    fActiveWorkSheet := nil;
end;

{ TExportCell }

function TExportCell.AsHTML: OWideString;
var OutText, OutCSSClass, OutCSSStyle, OutParameters: OWideString;
begin
  OutParameters := '';
  OutCSSClass := '';
  OutCSSStyle := '';
  OutText := '';
  DoAsHTML(OutText, OutCSSClass, OutCSSStyle, OutParameters);

  {OutCSSStyle := OutCSSStyle + ' font-family:'''+StringToWideHTML(aFontName)+''';';
  OutCSSStyle := OutCSSStyle + ' font-size:'''+IntToStr(aFontSize)+'pt'';';}

  OutCSSClass := Trim(OutCSSClass);
  if OutCSSClass <> '' then
    OutCSSClass := ' class="'+OutCSSClass+'"';
  if OutCSSStyle <> '' then
    OutCSSStyle := ' style="'+OutCSSStyle+'"';
  OutParameters := Trim(OutParameters);
  if OutParameters <> '' then
    OutParameters := ' '+OutParameters;
  result := '<td'+OutCSSClass+OutCSSStyle+OutParameters+'>'+OutText+'</td>';
end;

procedure TExportCell.AssignTo(Dest: TPersistent);
begin
  if Dest is TExportCell then
  with TExportCell(Dest) do begin
    fAlignment := Self.fAlignment;
    fVAlignment := Self.fVAlignment;
    fWrapText := Self.fWrapText;
    fFont.Assign(Self.fFont);
    fBGColor := Self.fBGColor;

    fColSpan := Self.fColSpan;
    fRowSpan := Self.fRowSpan;

    fWidth := Self.fWidth;
    fHeight := Self.fHeight;
    fBorders := Self.fBorders;
  end;

  inherited;
end;

procedure TExportCell.DoAsODS(aRowNode, aCellNode: IXMLElement; aColNodes: TInterfaceList;
  aColStyles: TExportODS_ColStyles; aRowStyles: TExportODS_RowStyles;
  aCellStyles: TExportODS_CellStyles; aNumberStyles: TExportODS_NumberStyles;
  aRealCol: Integer);
var
  xStyleIndex: Integer;
begin
  if ColSpan > 1 then
    aCellNode.SetAttribute('table:number-columns-spanned', IntToStr(ColSpan));

  if Width > 0 then begin
    xStyleIndex := aColStyles.CreateAddUnique(Width);
    (aColNodes[aRealCol] as IXMLElement).SetAttribute('table:style-name', aColStyles.StyleName(xStyleIndex));//getNamedAttributeNew
  end;
  if Height > 0 then begin
    xStyleIndex := aRowStyles.CreateAddUnique(Height);
    aRowNode.SetAttribute('table:style-name', aRowStyles.StyleName(xStyleIndex));//getNamedAttributeNew
  end;

  xStyleIndex := aCellStyles.CreateAddStyleUnique(Self, aNumberStyles);
  aCellNode.SetAttribute('table:style-name', aCellStyles.StyleName(xStyleIndex));//getNamedAttributeNew
end;

procedure TExportCell.DoAsXLSX(aCellNode, aColsNode, aRowNode,
  aMergeCellsNode: IXMLElement; aCellName, aWorkSheetId: OWideString;
  aSharedStrings: TOWideStrings; aStyleSheet: TExportXLSX_StyleSheet;
  aChain: TExportXLSX_Chain; aRow, aCol: Integer; out outStyleId: Integer);

  procedure AddColsNode(aColsI: Integer; aWidth: Double);
  begin
    with CreateElement(aColsNode, 'col') do begin
      SetAttribute('min', IntToStr(aColsI+1));
      SetAttribute('max', IntToStr(aColsI+1));
      SetAttribute('width', FloatToSqlStr(aWidth));
      SetAttribute('customWidth', '1');
    end;
  end;
var
  I: Integer;
begin
  if ColSpan > 1 then begin
    with CreateElement(aMergeCellsNode, 'mergeCell') do
      SetAttribute('ref', ExcelRange(aCol, aRow, ColSpan, 1));
  end;
  aCellNode.SetAttribute('r', aCellName);
  outStyleId := aStyleSheet.CreateAddStyleUnique(Self);
  aCellNode.SetAttribute('s', IntToStr(outStyleId));
  if Width > 0 then begin
    for I := aColsNode.ChildNodes.Length to aCol do
      AddColsNode(I, 11);
    (aColsNode.ChildNodes.Item[aCol] as IXMLElement).SetAttribute('width', FloatToSqlStr(Width/7.5/1.4));//getNamedAttributeNew
  end;
  if Height > 0 then begin
    aRowNode.SetAttribute('ht', FloatToSqlStr(Height/1.3));//getNamedAttributeNew
    aRowNode.SetAttribute('customHeight', '1');//getNamedAttributeNew
  end;
end;

procedure TExportCell.DoAsHTML(var OutText: OWideString; var OutCSSClass: OWideString;
  var OutCSSStyle: OWideString; var OutParameters: OWideString);
begin

  if (ColSpan > 1) then
    OutParameters := OutParameters+' colspan="'+IntToStr(ColSpan)+'"';
  if (RowSpan > 1) then
    OutParameters := OutParameters+' rowspan="'+IntToStr(RowSpan)+'"';
  if Width > 0 then
    OutParameters := OutParameters+' width='+IntToStr(Width);
  if Height > 0 then
    OutParameters := OutParameters+' height='+IntToStr(Height);

  OutParameters := OutParameters+' align="'+AlignmentToHTML(fAlignment)+'"';
  OutParameters := OutParameters+' bgcolor="'+ColorToHTML(fBGColor)+'"';
  OutText := FontStyleToHTML(OutText, Font.Style);
  OutText := '<font color="'+ColorToHTML(Font.Color)+'">'+OutText+'</font>';


  {if (cbTop in Borders) then                                        //OOo
    OutCSSStyle := OutCSSStyle+'border-top:1px solid #00000;';
  if (cbRight in Borders) then
    OutCSSStyle := OutCSSStyle+'border-right:1px solid #00000;';
  if (cbBottom in Borders) then
    OutCSSStyle := OutCSSStyle+'border-bottom:1px solid #00000;';
  if (cbLeft in Borders) then
    OutCSSStyle := OutCSSStyle+'border-left:1px solid #00000;';}
  {if (cbTop in Borders) then                                         //EXCEL
    OutCSSStyle := OutCSSStyle+'border-top:0.5pt windowtext solid;';
  if (cbRight in Borders) then
    OutCSSStyle := OutCSSStyle+'border-right:0.5pt windowtext solid;';
  if (cbBottom in Borders) then
    OutCSSStyle := OutCSSStyle+'border-bottom:0.5pt windowtext solid;';
  if (cbLeft in Borders) then
    OutCSSStyle := OutCSSStyle+'border-left:0.5pt windowtext solid;';}

  if not(cbTop in Borders) then
    OutCSSStyle := OutCSSStyle+'border-top:0;';
  if not(cbRight in Borders) then
    OutCSSStyle := OutCSSStyle+'border-right:0;';
  if not(cbBottom in Borders) then
    OutCSSStyle := OutCSSStyle+'border-bottom:0;';
  if not(cbLeft in Borders) then
    OutCSSStyle := OutCSSStyle+'border-left:0;';
end;

procedure TExportCell.DoCreate;
var xExportObject: TExportObject;
begin
  inherited;

  fFont := TFont.Create;
  xExportObject := GetExportObject;
  if Assigned(xExportObject) then
    fFont.Assign(xExportObject.DefaultFont);
  //fFontColor := clBlack;
  //fFontStyle := [];
  fBGColor := clNone;
  fAlignment := taLeftJustify;
  fVAlignment := taVerticalCenter;
  fWrapText := False;
  fColSpan := 1;
  fRowSpan := 1;
  fWidth := 0;//100
  fHeight := 0;
  fBorders := [cbTop, cbRight, cbBottom, cbLeft];
  //fCSSClass := '';
end;

function TExportCell.GetExportObject: TExportObject;
begin
  if OwnerList is TExportCells then
    Result := TExportCells(OwnerList).GetExportObject
  else
    Result := nil;
end;

{ TExportCellString }

procedure TExportCellString.AssignTo(Dest: TPersistent);
begin
  if Dest is TExportCellString then
  with TExportCellString(Dest) do begin
    fText := Self.fText;
  end;

  inherited;
end;

procedure TExportCellString.DoAsHTML(var OutText: OWideString; var OutCSSClass: OWideString;
  var OutCSSStyle: OWideString; var OutParameters: OWideString);
begin
  //FOR EXCEL
  OutCSSClass := 'text';

  //FOR OOoCalc
  OutParameters := OutParameters +
    ' SDNUM="3079;0;@"';

  OutText := StringToWideHTML(Text);

  inherited DoAsHTML(OutText, OutCSSClass, OutCSSStyle, OutParameters);
end;

procedure TExportCellString.DoAsODS(aRowNode, aCellNode: IXMLElement;
  aColNodes: TInterfaceList; aColStyles: TExportODS_ColStyles;
  aRowStyles: TExportODS_RowStyles; aCellStyles: TExportODS_CellStyles;
  aNumberStyles: TExportODS_NumberStyles; aRealCol: Integer);
begin
  inherited;

  aCellNode.SetAttribute('office:value-type', 'string');//getNamedAttributeNew
  CreateElement(aCellNode, 'text:p').Text := Text;//StringToUTF8XML(Text, False);
end;

procedure TExportCellString.DoAsXLSX(aCellNode, aColsNode, aRowNode,
  aMergeCellsNode: IXMLElement; aCellName, aWorkSheetId: OWideString;
  aSharedStrings: TOWideStrings; aStyleSheet: TExportXLSX_StyleSheet;
  aChain: TExportXLSX_Chain; aRow, aCol: Integer; out outStyleId: Integer);
begin
  inherited;

  aCellNode.SetAttribute('t', 's');//type = String
  FindCreateElement(aCellNode, 'v').Text := IntToStr(aSharedStrings.IndexOf(Text));
end;

procedure TExportCellString.DoCreate;
begin
  inherited;

  fText := '';
end;

function TExportCellString.GetSqlText: OWideString;
begin
  Result := Text;
end;

procedure TExportCellString.SetSqlText(const Value: OWideString);
begin
  Text := Value;
end;

procedure TExportCellString.SetText(const Value: OWideString);
begin
  if fText <> Value then begin
    fText := Value;
    if Pos(#13, Value) > 0 then
      WrapText := True;
  end;
end;

{ TExportCellNumber }

procedure TExportCellNumber.CreateStyleNodeODS(aStylesNode: IXMLElement; aStyleName: OWideString);
var xNSE: IXMLElement;
begin
  xNSE := CreateElement(aStylesNode, 'number:number-style');
  xNSE.SetAttribute('style:name', aStyleName);//StringToUTF8XML(aStyleName));

  with CreateElement(xNSE, 'number:number') do begin
    SetAttribute('number:decimal-places', IntToStr(Format));
    SetAttribute('number:min-integer-digits', '1');
  end;
end;

procedure TExportCellNumber.DoAsHTML(var OutText: OWideString; var OutCSSClass: OWideString;
  var OutCSSStyle: OWideString; var OutParameters: OWideString);
var
  FSExcel, FSOOoCalc, FSDisplay: TFormatSettings;
begin
  GetLocaleFormatSettings(0, FSExcel);
  GetLocaleFormatSettings(0, FSOOoCalc);
  GetLocaleFormatSettings(0, FSDisplay);
  FSExcel.DecimalSeparator := '.';
  FSExcel.ThousandSeparator := #0;
  FSOOoCalc.DecimalSeparator := ',';
  FSOOoCalc.ThousandSeparator := #0;
  FSDisplay.DecimalSeparator := ',';
  FSDisplay.ThousandSeparator := ' ';

  //FOR EXCEL
  OutCSSStyle := OutCSSStyle + ' mso-number-format:''' + GetExcelNumberFormatCSS(Format, NumberFormatPostfix) + ''';';
  OutParameters := OutParameters + ' x:num="' + FloatToStr(Value, FSExcel) + '"';

  //FOR OOoCalc
  OutCSSStyle := OutCSSStyle + '';
  OutParameters := OutParameters +
    ' SDVAL="' + FloatToStr(Value, FSOOoCalc) + '"' +
    ' SDNUM="3079;0;' + GetOOoCalcNumberFormat(Format, NumberFormatPostfix) + '"';

  //OutText := FloatToStrF(Value, ffFixed, 18, Format, FSDisplay);

  OutText := FloatToStr(Value, FSDisplay);

  inherited DoAsHTML(OutText, OutCSSClass, OutCSSStyle, OutParameters);
end;

procedure TExportCellNumber.DoAsODS(aRowNode, aCellNode: IXMLElement;
  aColNodes: TInterfaceList; aColStyles: TExportODS_ColStyles;
  aRowStyles: TExportODS_RowStyles; aCellStyles: TExportODS_CellStyles;
  aNumberStyles: TExportODS_NumberStyles; aRealCol: Integer);
begin
  inherited;

  aCellNode.SetAttribute('office:value-type', 'float');//getNamedAttributeNew
  aCellNode.SetAttribute('office:value', FloatToSqlStr(Value));//getNamedAttributeNew
end;

procedure TExportCellNumber.DoAsXLSX(aCellNode, aColsNode, aRowNode,
  aMergeCellsNode: IXMLElement; aCellName, aWorkSheetId: OWideString;
  aSharedStrings: TOWideStrings; aStyleSheet: TExportXLSX_StyleSheet;
  aChain: TExportXLSX_Chain; aRow, aCol: Integer; out outStyleId: Integer);
begin
  inherited;

  //aCellNode.SetAttribute('t', 's');//without type = number
  FindCreateElement(aCellNode, 'v').Text := FloatToSqlStr(fValue);
end;

procedure TExportCellNumber.DoCreate;
begin
  inherited;

  fFormat := 2;
  fValue := 0;

  Alignment := taRightJustify;
end;

function TExportCellNumber.GetNumberFormatXLSXStr: OWideString;
begin
  Result := GetXLSXNumberFormat(Format, '');
end;

function TExportCellNumber.GetSqlText: OWideString;
begin
  Result := FloatToSqlStr(fValue);
end;

function TExportCellNumber.NumberFormatPostfix: OWideString;
begin
  Result := '';
end;

procedure TExportCellNumber.SetSqlText(const Value: OWideString);
begin
  fValue := SqlStrToFloat(Value);
end;

{ TExportCellPercent }

procedure TExportCellPercent.CreateStyleNodeODS(aStylesNode: IXMLElement;
  aStyleName: OWideString);
var xPSN: IXMLElement;
begin
  xPSN := CreateElement(aStylesNode, 'number:percentage-style');
  xPSN.SetAttribute('style:name', aStyleName);//StringToUTF8XML(aStyleName));
  with CreateElement(xPSN, 'number:number') do begin
    SetAttribute('number:decimal-places', IntToStr(Format));
  end;
  CreateElement(xPSN, 'number:text').Text := '%';
end;

procedure TExportCellPercent.DoAsODS(aRowNode, aCellNode: IXMLElement;
  aColNodes: TInterfaceList; aColStyles: TExportODS_ColStyles;
  aRowStyles: TExportODS_RowStyles; aCellStyles: TExportODS_CellStyles;
  aNumberStyles: TExportODS_NumberStyles; aRealCol: Integer);
begin
  inherited;

  aCellNode.SetAttribute('office:value-type', 'percentage');//getNamedAttributeNew
end;

function TExportCellPercent.GetNumberFormatXLSXStr: OWideString;
begin
  Result := GetXLSXNumberFormat(Format, '%');
end;

function TExportCellPercent.NumberFormatPostfix: OWideString;
begin
  result := '%';
end;

{ TExportCellFormulaNumber }

procedure TExportCellFormulaNumber.AssignTo(Dest: TPersistent);
begin
  if Dest is TExportCellFormulaNumber then
  with TExportCellFormulaNumber(Dest) do begin
    fFormula := Self.fFormula;
  end;
  inherited;
end;

procedure TExportCellFormulaNumber.DoAsHTML(var OutText, OutCSSClass,
  OutCSSStyle, OutParameters: OWideString);
begin
  inherited DoAsHTML(OutText, OutCSSClass, OutCSSStyle, OutParameters);

  OutParameters := OutParameters + ' x:fmla="' + StringToWideHTML(Formula) + '"';
end;

procedure TExportCellFormulaNumber.DoAsODS(aRowNode, aCellNode: IXMLElement;
  aColNodes: TInterfaceList; aColStyles: TExportODS_ColStyles;
  aRowStyles: TExportODS_RowStyles; aCellStyles: TExportODS_CellStyles;
  aNumberStyles: TExportODS_NumberStyles; aRealCol: Integer);
begin
  inherited;

  RemoveAttributeSave(aCellNode, 'office:value');
  aCellNode.SetAttribute('table:formula', Formula);//StringToUTF8XML(Formula));//getNamedAttributeNew
end;

procedure TExportCellFormulaNumber.DoAsXLSX(aCellNode, aColsNode, aRowNode,
  aMergeCellsNode: IXMLElement; aCellName, aWorkSheetId: OWideString;
  aSharedStrings: TOWideStrings; aStyleSheet: TExportXLSX_StyleSheet;
  aChain: TExportXLSX_Chain; aRow, aCol: Integer; out outStyleId: Integer);
begin
  CreateElement(aCellNode, 'f').Text := Copy(Formula, 2, High(Integer));//MUST BE BEFORE inherited;

  inherited;

  RemoveElementSave(aCellNode, 'v');

  with TExportXLSX_ChainItem.CreateAdd(aChain) do begin
    fCellName := aCellName;
    fWorkSheetId := aWorkSheetId;
  end;
end;

procedure TExportCellFormulaNumber.DoCreate;
begin
  inherited;

  fFormula := '';
end;

procedure TExportCellFormulaNumber.SetFormula(const Value: OWideString);
begin
  if (Length(Value) > 0) and (Value[1] <> '=') then
    fFormula := '='+Value
  else
    fFormula := Value;
end;

{ TExportCellFormulaPercent }

procedure TExportCellFormulaPercent.CreateStyleNodeODS(aStylesNode: IXMLElement;
  aStyleName: OWideString);
var xPSE: IXMLElement;
begin
  xPSE := CreateElement(aStylesNode, 'number:percentage-style');
  xPSE.SetAttribute('style:name', aStyleName);//StringToUTF8XML(aStyleName));
  with CreateElement(xPSE, 'number:number') do begin
    SetAttribute('number:decimal-places', IntToStr(Format));
  end;
  CreateElement(xPSE, 'number:text').Text := '%';
end;

procedure TExportCellFormulaPercent.DoAsODS(aRowNode, aCellNode: IXMLElement;
  aColNodes: TInterfaceList; aColStyles: TExportODS_ColStyles;
  aRowStyles: TExportODS_RowStyles; aCellStyles: TExportODS_CellStyles;
  aNumberStyles: TExportODS_NumberStyles; aRealCol: Integer);
begin
  inherited;

  aCellNode.SetAttribute('office:value-type', 'percentage');//getNamedAttributeNew
end;

function TExportCellFormulaPercent.GetNumberFormatXLSXStr: OWideString;
begin
  Result := GetXLSXNumberFormat(Format, '%');
end;

function TExportCellFormulaPercent.NumberFormatPostfix: OWideString;
begin
  result := '%';
end;

{ TCustomGenerateExport }

constructor TCustomGenerateExport.Create;
begin
  Create(TExportXLSXSaveDialog);
end;

constructor TCustomGenerateExport.Create(aExportClass: TExportClass);
begin
  inherited Create;

  fExportObject := aExportClass.Create;

  DoCreate;
end;

procedure TCustomGenerateExport.DefaultExportFoot(
  Sender: TCustomGenerateExport);
begin
  //nothing
end;

procedure TCustomGenerateExport.DefaultExportHead(
  Sender: TCustomGenerateExport);
begin
  with TExportRow.CreateInsert(Sender.ExportObject.ActiveWorkSheet.Rows, 0) do
  with TExportCellString.CreateAdd(Cells) do begin
    Text := Title;
    ColSpan := fTitleColSpan;
    Font.Style := [fsBold];
    Borders := [];
  end;
  with TExportRow.CreateInsert(Sender.ExportObject.ActiveWorkSheet.Rows, 1) do
  with TExportCellString.CreateAdd(Cells) do begin
    ColSpan := fTitleColSpan;
    Borders := [];
  end;
end;

procedure TCustomGenerateExport.DoCreate;
begin
  fTitle := '';
  fRowCount := 0;
  fTitleColSpan := 1;

  fOnExportHead := DefaultExportHead;
  fOnExportFoot := DefaultExportFoot;
end;

procedure TCustomGenerateExport.DoExecute(aClearRows: Boolean;
  aFormProcess: IOCustomProgressDlg);
var
  I: Integer;
begin
  ScreenCursors.SetHourGlassCursor;
  try
    fExportObject.Title := fTitle;
    if aClearRows then
      fExportObject.ActiveWorkSheet.Rows.Clear;

    if Assigned(OnExportHead) then
      fOnExportHead(Self);

    DoExportRow(EXP_HEADER_ROW);
    for I := 0 to RowCount-1 do begin
      DoExportRow(I);

      if Assigned(aFormProcess) then
        aFormProcess.IncPosition;
    end;
    DoExportRow(EXP_FOOTER_ROW);

    if Assigned(OnExportFoot) then
      fOnExportFoot(Self);

    {if Assigned(FormProcess) then
      FormProcess.Hide;}

  finally
    ScreenCursors.ReleaseLastCursor;
  end;
end;

procedure TCustomGenerateExport.DoExportRow(const Row: Integer);
begin
  //nothing here
end;

procedure TCustomGenerateExport.Execute(aClearRows: Boolean;
  aFormProcess: IOCustomProgressDlg);
begin
  DoExecute(aClearRows, aFormProcess);
end;

procedure TCustomGenerateExport.ExecuteExport;
begin
  fExportObject.Export;
end;

{ TGenerateExportRows }

procedure TGenerateExportRows.DoCreate;
begin
  inherited;

  fOnExportRow := nil;
end;

procedure TGenerateExportRows.DoExecute;
begin
  inherited;
end;

procedure TGenerateExportRows.DoExportRow(const Row: Integer);
begin
  inherited;

  if Assigned(OnExportRow) then
    fOnExportRow(Self, Row);
end;

{ TExportXLSX }

procedure TExportXLSX.CreateContents(aZip: TAbArchive);
var
  xChain: TExportXLSX_Chain;
  xStyleSheet: TExportXLSX_StyleSheet;
  xSharedStrings: TExportXLSX_SharedStrings;
  I, L, K: Integer;
begin
  xChain := TExportXLSX_Chain.Create;
  xSharedStrings := TExportXLSX_SharedStrings.Create;
  xStyleSheet := TExportXLSX_StyleSheet.Create(DefaultFont);
  try
    xSharedStrings.Sorted := True;
    xSharedStrings.Duplicates := dupIgnore;
    for I := 0 to WorkSheets.Count-1 do
    for L := 0 to WorkSheets[I].Rows.Count-1 do
    for K := 0 to WorkSheets[I].Rows[L].Cells.Count-1 do
    if WorkSheets[I].Rows[L].Cells[K] is TExportCellString then begin
      xSharedStrings.Add(TExportCellString(WorkSheets[I].Rows[L].Cells[K]).Text);
    end;

    for I := 0 to WorkSheets.Count-1 do
      CreateSheet(aZip, WorkSheets.GetId(I), WorkSheets[I], xSharedStrings, xStyleSheet, xChain);

    xChain.Export(aZip);
    xStyleSheet.Export(aZip);
    xSharedStrings.Export(aZip);
  finally
    xChain.Free;
    xSharedStrings.Free;
    xStyleSheet.Free;
  end;
end;

procedure TExportXLSX.CreateContentTypes(aZip: TAbArchive);
var
  xSS: TStringStream;
  s: OWideString;
  I: Integer;
begin
  s :=
    DefXMLHeaderBrackets+CrLf+
    '<Types xmlns="http://schemas.openxmlformats.org/package/2006/content-types">'+
      '<Override PartName="/xl/theme/theme1.xml" ContentType="application/vnd.openxmlformats-officedocument.theme+xml"/>'+
      '<Override PartName="/xl/styles.xml" ContentType="application/vnd.openxmlformats-officedocument.spreadsheetml.styles+xml"/>'+
      '<Default Extension="rels" ContentType="application/vnd.openxmlformats-package.relationships+xml"/>'+
      '<Default Extension="xml" ContentType="application/xml"/>'+
      '<Override PartName="/xl/workbook.xml" ContentType="application/vnd.openxmlformats-officedocument.spreadsheetml.sheet.main+xml"/>'+
      '<Override PartName="/docProps/app.xml" ContentType="application/vnd.openxmlformats-officedocument.extended-properties+xml"/>';
  for I := 0 to WorkSheets.Count-1 do
    s := s + '<Override PartName="/xl/worksheets/sheet'+WorkSheets.GetId(I)+'.xml" ContentType="application/vnd.openxmlformats-officedocument.spreadsheetml.worksheet+xml"/>';
  s := s +
      '<Override PartName="/xl/calcChain.xml" ContentType="application/vnd.openxmlformats-officedocument.spreadsheetml.calcChain+xml"/>'+
      '<Override PartName="/xl/sharedStrings.xml" ContentType="application/vnd.openxmlformats-officedocument.spreadsheetml.sharedStrings+xml"/>'+
      '<Override PartName="/docProps/core.xml" ContentType="application/vnd.openxmlformats-package.core-properties+xml"/>'+
    '</Types>';
  xSS := TStringStream.Create(s);
  try
    xSS.Position := 0;
    aZip.AddFromStream('[Content_Types].xml', xSS);
  finally
    xSS.Free;
  end;
end;

procedure TExportXLSX.CreateDocPropsApp(aZip: TAbArchive);
var
  xSS: TStringStream;
begin
  xSS := TStringStream.Create(
    DefXMLHeaderBrackets+CrLf+
    '<Properties xmlns="http://schemas.openxmlformats.org/officeDocument/2006/extended-properties" xmlns:vt="http://schemas.openxmlformats.org/officeDocument/2006/docPropsVTypes">'+
      '<Application>'+Application+'</Application>'+
    '</Properties>');
  try
    xSS.Position := 0;
    aZip.AddFromStream('docProps\app.xml', xSS);
  finally
    xSS.Free;
  end;
end;

procedure TExportXLSX.CreateDocPropsCore(aZip: TAbArchive);
var
  xSS: TStringStream;
begin
  xSS := TStringStream.Create(
    DefXMLHeaderBrackets+CrLf+
    '<cp:coreProperties xmlns:cp="http://schemas.openxmlformats.org/package/2006/metadata/core-properties" xmlns:dc="http://purl.org/dc/elements/1.1/" xmlns:dcterms="http://purl.org/dc/terms/"'+' xmlns:dcmitype="http://purl.org/dc/dcmitype/" xmlns:xsi="http://www.w3.org/2001/XMLSchema-instance">'+
      '<dc:creator>'+StringToUTF8XML(Author, False)+'</dc:creator>'+
    '</cp:coreProperties>');
  try
    xSS.Position := 0;
    aZip.AddFromStream('docProps\core.xml', xSS);
  finally
    xSS.Free;
  end;
end;

procedure TExportXLSX.CreateRels(aZip: TAbArchive);
var
  xSS: TStringStream;
begin
  xSS := TStringStream.Create(
    DefXMLHeaderBrackets+CrLf+
    '<Relationships xmlns="http://schemas.openxmlformats.org/package/2006/relationships">'+
      '<Relationship Id="rId3" Type="http://schemas.openxmlformats.org/officeDocument/2006/relationships/extended-properties" Target="docProps/app.xml"/>'+
      '<Relationship Id="rId2" Type="http://schemas.openxmlformats.org/package/2006/relationships/metadata/core-properties" Target="docProps/core.xml"/>'+
      '<Relationship Id="rId1" Type="http://schemas.openxmlformats.org/officeDocument/2006/relationships/officeDocument" Target="xl/workbook.xml"/>'+
    '</Relationships>');
  try
    xSS.Position := 0;
    aZip.AddFromStream('_rels\.rels', xSS);
  finally
    xSS.Free;
  end;
end;

procedure TExportXLSX.CreateSheet(aZip: TAbArchive; aWorkSheetId: OWideString;
  aWorkSheet: TExportWorkSheet; aSharedStrings: TOWideStrings;
  aStyleSheet: TExportXLSX_StyleSheet; aChain: TExportXLSX_Chain);
var
  xMSSheet: TMemoryStream;
  xXML: IXMLDocument;
  xDimensionNode, xSheetFormatPrNode, xSheetDataNode, xRowNode, xMergeCellsNode, xHeaderFooterNode, xColsNode, xCellNode: IXMLElement;
  I, L, K, CurrentL, maxI, maxL, xLastStyleId: Integer;
  xPI: IXMLProcessingInstruction;
begin
  xXML := CreateXMLDoc;//TjanXMLTree.Create('worksheet', '', nil);
  try
    xPI := xXML.CreateProcessingInstruction('xml', DefXMLHeader);
    xXML.AppendChild(xPI);

    xXML.DocumentElement := xXML.CreateElement('worksheet');
    xXML.DocumentElement.SetAttribute('xmlns', 'http://schemas.openxmlformats.org/spreadsheetml/2006/main');
    xXML.DocumentElement.SetAttribute('xmlns:r', 'http://schemas.openxmlformats.org/officeDocument/2006/relationships');

    xDimensionNode := CreateElement(xXML.DocumentElement, 'dimension');
    xSheetFormatPrNode := CreateElement(xXML.DocumentElement, 'sheetFormatPr');
    xSheetFormatPrNode.SetAttribute('baseColWidth', '10');
    xSheetFormatPrNode.SetAttribute('defaultRowHeight', '15');

    xColsNode := CreateElement(xXML.DocumentElement, 'cols');
    xSheetDataNode := CreateElement(xXML.DocumentElement, 'sheetData');
    xMergeCellsNode := CreateElement(xXML.DocumentElement, 'mergeCells');
    maxI := aWorkSheet.Rows.Count;
    maxL := 0;
    for I := 0 to aWorkSheet.Rows.Count-1 do begin
      xRowNode := CreateElement(xSheetDataNode, 'row');
      xRowNode.SetAttribute('r', IntToStr(I+1));
      xRowNode.SetAttribute('spans', '1:'+IntToStr(Max(1, aWorkSheet.Rows[I].Cells.Count)));

      maxL := Max(maxL, aWorkSheet.Rows[I].Cells.Count);
      CurrentL := 0;
      for L := 0 to aWorkSheet.Rows[I].Cells.Count-1 do begin
        xCellNode := CreateElement(xRowNode, 'c');
        aWorkSheet.Rows[I].Cells[L].DoAsXLSX(xCellNode, xColsNode, xRowNode, xMergeCellsNode,
          ExcelRange(CurrentL, I, 1, 1), aWorkSheetId, aSharedStrings, aStyleSheet, aChain, I, CurrentL, xLastStyleId);
        for K := 1 to aWorkSheet.Rows[I].Cells[L].ColSpan-1 do begin
          Inc(CurrentL);
          xCellNode := CreateElement(xRowNode, 'c');
          xCellNode.SetAttribute('r', ExcelRange(CurrentL, I, 1, 1));
          xCellNode.SetAttribute('s', IntToStr(xLastStyleId));
        end;
        Inc(CurrentL);
      end;
    end;
    xDimensionNode.SetAttribute('ref', ExcelRange(0, 0, maxL, maxI));
    if xMergeCellsNode.ChildNodes.Length = 0 then
      RemoveElementSave(xXML.DocumentElement, xMergeCellsNode)
    else
      xMergeCellsNode.SetAttribute('count', IntToStr(xMergeCellsNode.ChildNodes.Length));
    if xColsNode.ChildNodes.Length = 0 then
      RemoveElementSave(xXML.DocumentElement, xColsNode);

    if aWorkSheet.Header.Used or aWorkSheet.Footer.Used then begin
      xHeaderFooterNode := CreateElement(xXML.DocumentElement, 'headerFooter');
      if aWorkSheet.Header.Used then
        aWorkSheet.Header.DoAsXLSX(CreateElement(xHeaderFooterNode, 'oddHeader'));
      if aWorkSheet.Footer.Used then
        aWorkSheet.Footer.DoAsXLSX(CreateElement(xHeaderFooterNode, 'oddFooter'));
    end;

    xMSSheet := TMemoryStream.Create;
    try
      xXML.SaveToStream(xMSSheet);
      xMSSheet.Position := 0;

      aZip.AddFromStream('xl\worksheets\sheet'+aWorkSheetId+'.xml', xMSSheet);
    finally
      xMSSheet.Free;
    end;
  finally
    xXML := nil;
  end;
end;

procedure TExportXLSX.CreateTheme(aZip: TAbArchive);
var
  xSS: TStringStream;
begin
  xSS := TStringStream.Create(
    DefXMLHeaderBrackets+CrLf+
    '<a:theme xmlns:a="http://schemas.openxmlformats.org/drawingml/2006/main" name="Larissa-Design"></a:theme>');
  try
    xSS.Position := 0;
    aZip.AddFromStream('xl\theme\theme1.xml', xSS);
  finally
    xSS.Free;
  end;
end;

procedure TExportXLSX.CreateWorkBook(aZip: TAbArchive);
var
  xSS: TStringStream;
  s: UTF8String;
  I: Integer;
begin
  s :=
    DefXMLHeaderBrackets+CrLf+
    '<workbook xmlns="http://schemas.openxmlformats.org/spreadsheetml/2006/main" xmlns:r="http://schemas.openxmlformats.org/officeDocument/2006/relationships">'+
      '<fileVersion appName="xl" lastEdited="4" lowestEdited="4" rupBuild="4506"/>'+
      '<workbookPr defaultThemeVersion="124226" />'+
      '<sheets>';
  for I := 0 to WorkSheets.Count-1 do
    s := s + '<sheet name="'+StringToUTF8XML(WorkSheets[I].Title, False)+'" '+
      'sheetId="'+StringToUTF8XML(WorkSheets.GetId(I))+'" '+
      'r:id="'+StringToUTF8XML(WorkSheets.GetRelId(I))+'"/>';
  s := s +
      '</sheets>'+
      '<calcPr calcId="125725" />'+
    '</workbook>';
  xSS := TStringStream.Create(s);
  try
    xSS.Position := 0;
    aZip.AddFromStream('xl\workbook.xml', xSS);
  finally
    xSS.Free;
  end;
end;

procedure TExportXLSX.CreateWorkBookRels(aZip: TAbArchive);
var
  xSS: TStringStream;
  s: OWideString;
  I: Integer;
begin
  s :=
    DefXMLHeaderBrackets+CrLf+
    '<Relationships xmlns="http://schemas.openxmlformats.org/package/2006/relationships">'+
      '<Relationship Id="rId1" Type="http://schemas.openxmlformats.org/officeDocument/2006/relationships/styles" Target="styles.xml"/>'+
      '<Relationship Id="rId2" Type="http://schemas.openxmlformats.org/officeDocument/2006/relationships/theme" Target="theme/theme1.xml"/>'+
      '<Relationship Id="rId3" Type="http://schemas.openxmlformats.org/officeDocument/2006/relationships/calcChain" Target="calcChain.xml"/>'+
      '<Relationship Id="rId4" Type="http://schemas.openxmlformats.org/officeDocument/2006/relationships/sharedStrings" Target="sharedStrings.xml"/>';
  for I := 0 to WorkSheets.Count-1 do
    s := s + '<Relationship Id="'+WorkSheets.GetRelId(I)+'" Type="http://schemas.openxmlformats.org/officeDocument/2006/relationships/worksheet" Target="worksheets/sheet'+WorkSheets.GetId(I)+'.xml"/>';
  s := s +
    '</Relationships>';
  xSS := TStringStream.Create(s);
  try
    xSS.Position := 0;
    aZip.AddFromStream('xl\_rels\workbook.xml.rels', xSS);
  finally
    xSS.Free;
  end;
end;

procedure TExportXLSX.DoCreate;
begin
  inherited;

  fExportStream := nil;
end;

type
  TAbArchiveItemClass = class of TAbArchiveItem;

function TExportXLSX.DoExport: Boolean;
var
  xZip: TOZipArchive;
begin
  xZip := TOZipArchive.CreateFromStream(fExportStream, '');
  try
    CreateContentTypes(xZip);
    CreateRels(xZip);
    CreateWorkBookRels(xZip);
    CreateWorkBook(xZip);
    CreateDocPropsCore(xZip);
    CreateDocPropsApp(xZip);
    CreateTheme(xZip);
    CreateContents(xZip);

    xZip.Save;
  finally
    xZip.Free;
  end;
  Result := True;
end;

{ TExportRows }

procedure TExportRows.DoCreate;
begin
  inherited;

  fWorksheet := nil;
end;

function TExportRows.GetER(Index: Integer): TExportRow;
begin
  Result := TExportRow(inherited Items[Index]);
end;

function TExportRows.GetExportObject: TExportObject;
begin
  if Assigned(fWorksheet) then
    Result := fWorksheet.GetExportObject
  else
    Result := nil;
end;

procedure TExportRows.SetER(Index: Integer; const Value: TExportRow);
begin
  inherited Items[Index] := Value;
end;

{ TExportWorkSheet }

function TExportWorkSheet.AddRow: TExportRow;
begin
  Result := TExportRow.CreateAdd(Self.Rows);
end;

procedure TExportWorkSheet.DoCreate;
begin
  inherited;

  fTitle := oexport_worksheet+' '+IntToStr(IndexInList+1);
  fRows := TExportRows.Create;
  fRows.fWorksheet := Self;
  fHeader := TExportHeaders.Create(GetExportObject);
  fFooter := TExportHeaders.Create(GetExportObject);
end;

procedure TExportWorkSheet.DoDestroy;
begin
  fHeader.Free;
  fFooter.Free;
  inherited;

end;

function TExportWorkSheet.GetExportObject: TExportObject;
begin
  if (OwnerList is TExportWorkSheets) then
    Result := TExportWorkSheets(OwnerList).GetExportObject
  else
    Result := nil;
end;

{ TExportWorkSheets }

procedure TExportWorkSheets.DoCreate;
begin
  inherited;

  fExportObject := nil;
end;

function TExportWorkSheets.GetExportObject: TExportObject;
begin
  Result := fExportObject;
end;

function TExportWorkSheets.GetId(Index: Integer): OWideString;
begin
  Result := IntToStr(1+Index);
end;

function TExportWorkSheets.GetRelId(Index: Integer): OWideString;
begin
  Result := 'rId'+IntToStr(5+Index);
end;

function TExportWorkSheets.GetWS(Index: Integer): TExportWorkSheet;
begin
  Result := TExportWorkSheet(inherited Items[Index]);
end;

procedure TExportWorkSheets.SetWS(Index: Integer;
  const Value: TExportWorkSheet);
begin
  inherited Items[Index] := Value;
end;

{ TExportXLSX_Chain }

procedure TExportXLSX_Chain.Export(aZip: TAbArchive);
var
  xSS: TStringStream;
  s: UTF8String;
  I: Integer;
begin
  if Count = 0 then
    exit;
  s :=
    DefXMLHeaderBrackets+CrLf+
    '<calcChain xmlns="http://schemas.openxmlformats.org/spreadsheetml/2006/main">';
  for I := 0 to Count-1 do
    s := s + '<c r="'+StringToUTF8Xml(Items[I].fCellName)+'" i="'+StringToUTF8Xml(Items[I].fWorkSheetId)+'"/>';
  s := s +
    '</calcChain>';
  xSS := TStringStream.Create(s);
  try
    xSS.Position := 0;
    aZip.AddFromStream('xl\calcChain.xml', xSS);
  finally
    xSS.Free;
  end;
end;

function TExportXLSX_Chain.GetI(Index: Integer): TExportXLSX_ChainItem;
begin
  Result := TExportXLSX_ChainItem(inherited Items[Index]);
end;

procedure TExportXLSX_Chain.SetI(Index: Integer;
  const Value: TExportXLSX_ChainItem);
begin
  inherited Items[Index] := Value;
end;

{ TExportXLSX_Styles }

function GetXLSXNumberFormatId(Index: Integer): OWideString;
begin
  Result := IntToStr(171+Index);
end;

procedure TExportXLSX_Styles.Export(aCellXfsNode: IXMLElement);
var I: Integer;
  xStyle: TExportXLSX_Style;
  xStyleNode: IXMLElement;
begin
  aCellXfsNode.SetAttribute('count', IntToStr(Count));
  for I := 0 to Count-1 do begin
    xStyle := Items[I];
    xStyleNode := CreateElement(aCellXfsNode, 'xf');
    if xStyle.fNumFmtId = -1 then
      xStyleNode.SetAttribute('numFmtId', '0')
    else
      xStyleNode.SetAttribute('numFmtId', GetXLSXNumberFormatId(xStyle.fNumFmtId));
    xStyleNode.SetAttribute('fontId', IntToStr(xStyle.fFontId));
    xStyleNode.SetAttribute('fillId', IntToStr(xStyle.fFillId));
    xStyleNode.SetAttribute('borderId', IntToStr(xStyle.fBorderId));
    xStyleNode.SetAttribute('xfId', '0');
    xStyleNode.SetAttribute('applyFont', '1');
    xStyleNode.SetAttribute('applyFill', '1');
    //if (xStyle.fAlignment <> taLeftJustify) or (xStyle.fVAlignment <> taAlignTop) then begin
      xStyleNode.SetAttribute('applyAlignment', '1');
      with CreateElement(xStyleNode, 'alignment') do begin
        //if (xStyle.fAlignment <> taLeftJustify) then
          SetAttribute('horizontal', AlignmentToHTML(xStyle.fAlignment));
        //if (xStyle.fVAlignment <> taAlignTop) then
          SetAttribute('vertical', VAlignmentToHTML(xStyle.fVAlignment));
          if xStyle.fWrapText then
            SetAttribute('wrapText', '1');
      end;
    //end;
  end;
end;

function TExportXLSX_Styles.GetI(Index: Integer): TExportXLSX_Style;
begin
  Result := TExportXLSX_Style(inherited Items[Index]);
end;

procedure TExportXLSX_Styles.SetI(Index: Integer;
  const Value: TExportXLSX_Style);
begin
  inherited Items[Index] := Value;
end;

{ TExportXLSX_SharedStrings }

procedure TExportXLSX_SharedStrings.Export(aZip: TAbArchive);
  function FormatStringForExcel(aStr: OWideString): OWideString;
  begin
    Result := aStr;
    Result := OWideStringReplace(Result, #13#10, #10, [rfReplaceAll]);
    Result := OWideStringReplace(Result, #13, #10, [rfReplaceAll]);//if somebody used only #13 as line break
  end;
var
  xSS: TStringStream;
  s: UTF8String;
  xAttr: UTF8String;
  I: Integer;
begin
  s :=
    DefXMLHeaderBrackets+CrLf+
    '<sst xmlns="http://schemas.openxmlformats.org/spreadsheetml/2006/main" '+
    'count="'+StringToUTF8XML(IntToStr(Count))+'" '+
    'uniqueCount="'+StringToUTF8XML(IntToStr(Count))+'">';
  for I := 0 to Count-1 do begin
    if (Strings[I] <> '') and ((Strings[I][1] = ' ') or (Strings[I][Length(Strings[I])] = ' ')) then
      xAttr := ' xml:space="preserve"'
    else
      xAttr := '';
    s := s + '<si><t'+xAttr+'>'+StringToUTF8Xml(FormatStringForExcel(Strings[I]), False, False)+'</t></si>';
  end;
  s := s +
    '</sst>';
  xSS := TStringStream.Create(s);
  try
    xSS.Position := 0;
    aZip.AddFromStream('xl\sharedStrings.xml', xSS);
  finally
    xSS.Free;
  end;
end;

{ TExportCells }

procedure TExportCells.DoCreate;
begin
  inherited;

  fExportRow := nil;
end;

function TExportCells.GetC(Index: Integer): TExportCell;
begin
  Result := TExportCell(inherited Items[Index]);
end;

function TExportCells.GetExportObject: TExportObject;
begin
  if Assigned(fExportRow) then
    Result := fExportRow.GetExportObject
  else
    Result := nil;
end;

procedure TExportCells.SetC(Index: Integer; const Value: TExportCell);
begin
  Items[Index] := Value;
end;

{ TExportXLSX_Fonts }

constructor TExportXLSX_Fonts.Create;
begin
  inherited;
  OwnsObjects := True;
end;

function TExportXLSX_Fonts.CreateAddFontUnique(aFont: TFont): Integer;
var I : Integer;
  NewFont: TFont;
begin
  for I := 0 to Count-1 do
  if Items[I] is TFont then begin
    if SameFont(TFont(Items[I]), aFont) then begin
      Result := I;
      exit;
    end;
  end;

  //not found
  NewFont := TFont.Create;
  NewFont.Assign(aFont);
  Result := Add(NewFont);
end;

procedure TExportXLSX_Fonts.Export(aFontsNode: IXMLElement);
var I: Integer;
  xFont: TFont;
  xFontNode: IXMLElement;
begin
  aFontsNode.SetAttribute('count', IntToStr(Count));
  for I := 0 to Count-1 do
  if Items[I] is TFont then begin
    xFont := TFont(Items[I]);
    xFontNode := CreateElement(aFontsNode, 'font');
    if fsBold in xFont.Style then
      CreateElement(xFontNode, 'b');
    if fsItalic in xFont.Style then
      CreateElement(xFontNode, 'i');
    if fsUnderline in xFont.Style then
      CreateElement(xFontNode, 'u');
    if fsStrikeOut in xFont.Style then
      CreateElement(xFontNode, 'strike');
    CreateElement(xFontNode, 'sz').SetAttribute('val', IntToStr(xFont.Size));
    CreateElement(xFontNode, 'color').SetAttribute('rgb', UpperCase('FF'+ColorToHTML(xFont.Color, False)));
    CreateElement(xFontNode, 'name').SetAttribute('val', xFont.Name);//StringToUTF8XML(xFont.Name));
  end;
end;

function TExportXLSX_Fonts.GetI(Index: Integer): TFont;
begin
  Result := TFont(inherited Items[Index]);
end;

procedure TExportXLSX_Fonts.SetI(Index: Integer; const Value: TFont);
begin
  inherited Items[Index] := Value;
end;

{ TExportXLSX_Fills }

function TExportXLSX_Fills.CreateAddFillUnique(aColor: TColor): Integer;
var I : Integer;
begin
  for I := 0 to Count-1 do
  if Items[I].fColor = aColor then begin
    Result := I;
    exit;
  end;

  //not found
  with TExportXLSX_Fill.CreateAdd(Self) do begin
    fColor := aColor;
  end;
  Result := Count-1;
end;

procedure TExportXLSX_Fills.DoCreate;
begin
  inherited;

  //MUST BE HERE: IS NEEDED BY EXCEL
  with TExportXLSX_Fill.CreateAdd(Self) do begin
    fColor := clNone;
  end;
  with TExportXLSX_Fill.CreateAdd(Self) do begin
    fColor := clBtnFace;
  end;
end;

procedure TExportXLSX_Fills.Export(aFillsNode: IXMLElement);
var I: Integer;
  xFillNode: IXMLElement;
begin
  aFillsNode.SetAttribute('count', IntToStr(Count));
  for I := 0 to Count-1 do begin
    xFillNode := CreateElement(CreateElement(aFillsNode, 'fill'), 'patternFill');
    if Items[I].fColor = clNone then begin
      xFillNode.SetAttribute('patternType', 'none');
    end else if Items[I].fColor = clBtnFace then begin
      xFillNode.SetAttribute('patternType', 'gray125');
    end else begin
      xFillNode.SetAttribute('patternType', 'solid');
      CreateElement(xFillNode, 'fgColor').SetAttribute('rgb', UpperCase('FF'+ColorToHTML(Items[I].fColor, False)));
      CreateElement(xFillNode, 'bgColor').SetAttribute('indexed', '64');
    end;
  end;
end;

function TExportXLSX_Fills.GetI(Index: Integer): TExportXLSX_Fill;
begin
  Result := TExportXLSX_Fill(inherited Items[Index]);
end;

procedure TExportXLSX_Fills.SetI(Index: Integer; const Value: TExportXLSX_Fill);
begin
  inherited Items[Index] := Value;
end;

{ TExportXLSX_StyleSheet }

constructor TExportXLSX_StyleSheet.Create(aDefaultFont: TFont);
var xDefaultFont: TFont;
begin
  inherited Create;

  fFonts := TExportXLSX_Fonts.Create;
  fFills := TExportXLSX_Fills.Create;
  fNumFmts := TExportXLSX_NumberFormats.Create;
  fBorders := TExportXLSX_Borders.Create;
  fCellXfs := TExportXLSX_Styles.Create;


  //DEFAULT STYLE - MUST BE HERE; NEEDED BY EXCEL
  with TExportXLSX_Style.CreateAdd(CellXfs) do begin
    xDefaultFont := TFont.Create;
    try
      if Assigned(aDefaultFont) then begin
        xDefaultFont.Assign(aDefaultFont);
      end else begin
        xDefaultFont.Name := 'Calibri';
        xDefaultFont.Size := 11;
        xDefaultFont.Color := clBlack;
      end;
      fFontId := Fonts.CreateAddFontUnique(xDefaultFont);
    finally
      xDefaultFont.Free;
    end;
    fFillId := Fills.CreateAddFillUnique(clNone);
    fNumFmtId := -1;
    fBorderId := Borders.CreateAddBorderUnique([]);
    fAlignment := taLeftJustify;
    fVAlignment := taAlignBottom;
    fWrapText := False;
  end;
end;

function TExportXLSX_StyleSheet.CreateAddStyleUnique(
  aCell: TExportCell): Integer;
var I: Integer;
begin
  for I := 0 to CellXfs.Count-1 do
  if SameFont(Fonts[CellXfs[I].fFontId], aCell.Font) and
    (Fills[CellXfs[I].fFillId].fColor = aCell.fBGColor) and
    (Borders[CellXfs[I].fBorderId].fBorders = aCell.fBorders) and
    (CellXfs[I].fAlignment = aCell.Alignment) and
    (CellXfs[I].fVAlignment = aCell.VAlignment) and
    (CellXfs[I].fWrapText = aCell.WrapText) and
    (not (aCell is TExportCellCustomNumber) or (
      not(CellXfs[I].fNumFmtId = -1) and
      (NumFmts[CellXfs[I].fNumFmtId].fNumberFormatStr = TExportCellCustomNumber(aCell).GetNumberFormatXLSXStr))
    )
  then begin
    Result := I;
    exit;
  end;

  //NEW STYLE
  with TExportXLSX_Style.CreateAdd(CellXfs) do begin
    fFontId := Fonts.CreateAddFontUnique(aCell.Font);
    fFillId := Fills.CreateAddFillUnique(aCell.BGColor);
    if (aCell is TExportCellCustomNumber) then
      fNumFmtId := NumFmts.CreateAddNumberFormatUnique(TExportCellCustomNumber(aCell).GetNumberFormatXLSXStr)
    else
      fNumFmtId := -1;
    fBorderId := Borders.CreateAddBorderUnique(aCell.Borders);
    fAlignment := aCell.Alignment;
    fVAlignment := aCell.VAlignment;
    fWrapText := aCell.WrapText;
  end;
  Result := CellXfs.Count-1;
end;

destructor TExportXLSX_StyleSheet.Destroy;
begin
  fFonts.Free;
  fFills.Free;
  fNumFmts.Free;
  fBorders.Free;
  fCellXfs.Free;

  inherited;
end;

procedure TExportXLSX_StyleSheet.Export(aZip: TAbArchive);
var
  xMS: TMemoryStream;
  xXML: IXMLDocument;
  xCSE: IXMLElement;
  xPI: IXMLProcessingInstruction;
begin
  xXML := CreateXMLDoc;//TjanXMLTree.Create('styleSheet', '', nil);
  xMS := TMemoryStream.Create;
  try
    xPI := xXML.CreateProcessingInstruction('xml', DefXMLHeader);
    xXML.AppendChild(xPI);

    xXML.DocumentElement := xXML.CreateElement('styleSheet');
    xXML.DocumentElement.SetAttribute('xmlns', 'http://schemas.openxmlformats.org/spreadsheetml/2006/main');
    fNumFmts.Export(CreateElement(xXML.DocumentElement, 'numFmts'));
    fFonts.Export(CreateElement(xXML.DocumentElement, 'fonts'));
    fFills.Export(CreateElement(xXML.DocumentElement, 'fills'));
    fBorders.Export(CreateElement(xXML.DocumentElement, 'borders'));

    xCSE := CreateElement(xXML.DocumentElement, 'cellStyleXfs');
    xCSE.SetAttribute('count', '1');
    with CreateElement(xCSE, 'xf') do begin
      SetAttribute('numFmtId', '0');
      SetAttribute('fontId', '0');
      SetAttribute('fillId', '0');
      SetAttribute('borderId', '0');
    end;

    fCellXfs.Export(CreateElement(xXML.DocumentElement, 'cellXfs'));

    xXML.SaveToStream(xMS);
    xMS.Position := 0;
    aZip.AddFromStream('xl\styles.xml', xMS);
  finally
    xXML := nil;
    xMS.Free;
  end;
end;

{ TExportXLSX_Borders }

function TExportXLSX_Borders.CreateAddBorderUnique(
  aBorder: TCellBorders): Integer;
var I : Integer;
begin
  for I := 0 to Count-1 do
  if Items[I].fBorders = aBorder then begin
    Result := I;
    exit;
  end;

  //not found
  with TExportXLSX_Border.CreateAdd(Self) do begin
    fBorders := aBorder;
  end;
  Result := Count-1;
end;

procedure TExportXLSX_Borders.DoCreate;
begin
  inherited;

  //MUST BE HERE: IS NEEDED BY EXCEL
  with TExportXLSX_Border.CreateAdd(Self) do begin
    fBorders := [];
  end;
end;

procedure TExportXLSX_Borders.Export(aBordersNode: IXMLElement);
var I: Integer;
  xBorderNode: IXMLElement;

  procedure AddBorder(aBorderNode: IXMLElement; aName: OWideString; aHasBorder: Boolean);
  var xNBE: IXMLElement;
  begin
    xNBE := CreateElement(aBorderNode, aName);
    if aHasBorder then begin
      xNBE.SetAttribute('style', 'thin');
      CreateElement(xNBE, 'color').SetAttribute('indexed', '64');//black
    end;
  end;
begin
  aBordersNode.SetAttribute('count', IntToStr(Count));
  for I := 0 to Count-1 do begin
    xBorderNode := CreateElement(aBordersNode, 'border');
    AddBorder(xBorderNode, 'left', cbLeft in Items[I].fBorders);
    AddBorder(xBorderNode, 'right', cbRight in Items[I].fBorders);
    AddBorder(xBorderNode, 'top', cbTop in Items[I].fBorders);
    AddBorder(xBorderNode, 'bottom', cbBottom in Items[I].fBorders);
    AddBorder(xBorderNode, 'diagonal', False);
  end;
end;

function TExportXLSX_Borders.GetI(Index: Integer): TExportXLSX_Border;
begin
  Result := TExportXLSX_Border(inherited Items[Index]);
end;

procedure TExportXLSX_Borders.SetI(Index: Integer;
  const Value: TExportXLSX_Border);
begin
  inherited Items[Index] := Value;
end;

{ TExportXLSX_NumberFormats }

function TExportXLSX_NumberFormats.CreateAddNumberFormatUnique(
  aNumberFormatStr: OWideString): Integer;
var I: Integer;
begin
  for I := 0 to Count-1 do
  if Items[I].fNumberFormatStr = aNumberFormatStr then begin
    Result := I;
    exit;
  end;

  //not found
  with TExportXLSX_NumberFormat.CreateAdd(Self) do begin
    fNumberFormatStr := aNumberFormatStr;
  end;
  Result := Count-1;
end;

procedure TExportXLSX_NumberFormats.Export(aNumberFormatsNode: IXMLElement);
var I: Integer;
  xNumberFormatNode: IXMLElement;
begin
  aNumberFormatsNode.SetAttribute('count', IntToStr(Count));
  for I := 0 to Count-1 do begin
    xNumberFormatNode := CreateElement(aNumberFormatsNode, 'numFmt');
    xNumberFormatNode.SetAttribute('numFmtId', GetXLSXNumberFormatId(I));
    xNumberFormatNode.SetAttribute('formatCode', Items[I].fNumberFormatStr);
  end;
end;

function TExportXLSX_NumberFormats.GetI(
  Index: Integer): TExportXLSX_NumberFormat;
begin
  Result := TExportXLSX_NumberFormat(inherited Items[Index]);
end;

procedure TExportXLSX_NumberFormats.SetI(Index: Integer;
  const Value: TExportXLSX_NumberFormat);
begin
  inherited Items[Index] := Value;
end;

{ TExportXLSXSaveDialog }

function TExportXLSXSaveDialog.DoExport: Boolean;
var
  SD: TOWideSaveDialog;
begin
  SD := TOWideSaveDialog.Create(nil);
  try
    SD.DefaultExt := xlsxexport_defaultextension;
    SD.Filter := xlsxexport_dialogfilter;
    SD.Title := xlsxexport_title;
    if fFileName <> '' then
      SD.FileName := fFileName
    else
      SD.FileName := NormFileName(Title);
    SD.Options := SD.Options + [ofOverwritePrompt];

    Result := SD.Execute;
    if Result then begin
      fFileName := SD.FileName;

      Result := inherited DoExport;
    end;
  finally
    SD.Free;
  end;
end;

{ TExportCellDate }

procedure TExportCellDate.CreateStyleNodeODS(aStylesNode: IXMLElement; aStyleName: OWideString);
var xDSE: IXMLElement;
begin
  xDSE := CreateElement(aStylesNode, 'number:date-style');
  xDSE.SetAttribute('style:name', aStyleName);//StringToUTF8XML(aStyleName));

  CreateElement(xDSE, 'number:day').SetAttribute('number:style', 'long');
  CreateElement(xDSE, 'number:text').Text := '.';
  CreateElement(xDSE, 'number:month').SetAttribute('number:style', 'long');
  CreateElement(xDSE, 'number:text').Text := '.';
  CreateElement(xDSE, 'number:year').SetAttribute('number:style', 'long');
end;

procedure TExportCellDate.DoAsODS(aRowNode, aCellNode: IXMLElement;
  aColNodes: TInterfaceList; aColStyles: TExportODS_ColStyles;
  aRowStyles: TExportODS_RowStyles; aCellStyles: TExportODS_CellStyles;
  aNumberStyles: TExportODS_NumberStyles; aRealCol: Integer);
begin
  inherited;

  if not SameValue(DateTime, 0) then begin
    aCellNode.SetAttribute('office:value-type', 'date');//getNamedAttributeNew
    aCellNode.SetAttribute('office:date-value', DateToODS(Value))//getNamedAttributeNew
  end;
end;

procedure TExportCellDate.DoAsXLSX(aCellNode, aColsNode, aRowNode,
  aMergeCellsNode: IXMLElement; aCellName, aWorkSheetId: OWideString;
  aSharedStrings: TOWideStrings; aStyleSheet: TExportXLSX_StyleSheet;
  aChain: TExportXLSX_Chain; aRow, aCol: Integer; out outStyleId: Integer);
begin
  inherited;

  if SameValue(DateTime, 0) then
    RemoveElementSave(aCellNode, 'v')
  else
    FindCreateElement(aCellNode, 'v').Text := IntToStr(Trunc(DateTime));
end;

function TExportCellDate.GetDateTime: TDateTime;
begin
  Result := Value;
end;

function TExportCellDate.GetNumberFormatXLSXStr: OWideString;
begin
  Result := 'd/m/yyyy;@';
end;

function TExportCellDate.GetSqlText: OWideString;
begin
  Result := DateToSqlStr(DateTime);
end;

procedure TExportCellDate.SetDateTime(const aValue: TDateTime);
begin
  Value := aValue;
end;

procedure TExportCellDate.SetSqlText(const Value: OWideString);
begin
  DateTime := SqlStrToDate(Value);
end;

{ TExportCellDateTime }

procedure TExportCellDateTime.CreateStyleNodeODS(aStylesNode: IXMLElement; aStyleName: OWideString);
var xNDE: IXMLElement;
begin
  xNDE := CreateElement(aStylesNode, 'number:date-style');
  xNDE.SetAttribute('style:name', aStyleName);//StringToUTF8XML(aStyleName));

  CreateElement(xNDE, 'number:day').SetAttribute('number:style', 'long');
  CreateElement(xNDE, 'number:text').Text := '.';
  CreateElement(xNDE, 'number:month').SetAttribute('number:style', 'long');
  CreateElement(xNDE, 'number:text').Text := '.';
  CreateElement(xNDE, 'number:year').SetAttribute('number:style', 'long');
  CreateElement(xNDE, 'number:text').Text := ' ';
  CreateElement(xNDE, 'number:hours').SetAttribute('number:style', 'long');
  CreateElement(xNDE, 'number:text').Text := ':';
  CreateElement(xNDE, 'number:minutes').SetAttribute('number:style', 'long');
  CreateElement(xNDE, 'number:text').Text := ':';
  CreateElement(xNDE, 'number:seconds').SetAttribute('number:style', 'long');
end;

procedure TExportCellDateTime.DoAsODS(aRowNode, aCellNode: IXMLElement;
  aColNodes: TInterfaceList; aColStyles: TExportODS_ColStyles;
  aRowStyles: TExportODS_RowStyles; aCellStyles: TExportODS_CellStyles;
  aNumberStyles: TExportODS_NumberStyles; aRealCol: Integer);
begin
  inherited;

  if not SameValue(Value, 0) then begin
    aCellNode.SetAttribute('office:date-value', DateTimeToODS(Value));//getNamedAttributeNew
  end;
end;

procedure TExportCellDateTime.DoAsXLSX(aCellNode, aColsNode, aRowNode,
  aMergeCellsNode: IXMLElement; aCellName, aWorkSheetId: OWideString;
  aSharedStrings: TOWideStrings; aStyleSheet: TExportXLSX_StyleSheet;
  aChain: TExportXLSX_Chain; aRow, aCol: Integer; out outStyleId: Integer);
begin
  inherited;

  if SameValue(DateTime, 0) then
    RemoveElementSave(aCellNode, 'v')
  else
    FindCreateElement(aCellNode, 'v').Text := FloatToSqlStr(DateTime);
end;

function TExportCellDateTime.GetNumberFormatXLSXStr: OWideString;
begin
  Result := 'dd/mm/yyyy\ hh:mm:ss';
end;

function TExportCellDateTime.GetSqlText: OWideString;
begin
  Result := DateTimeToSqlStr(DateTime);
end;

procedure TExportCellDateTime.SetSqlText(const Value: OWideString);
begin
  DateTime := SqlStrToDateTime(Value);
end;

{ TExportCellDateTimeWS }

procedure TExportCellDateTimeWS.CreateStyleNodeODS(aStylesNode: IXMLElement; aStyleName: OWideString);
var xDSE: IXMLElement;
begin
  xDSE := CreateElement(aStylesNode, 'number:date-style');
  xDSE.SetAttribute('style:name', aStyleName);//StringToUTF8XML(aStyleName));

  CreateElement(xDSE, 'number:day').SetAttribute('number:style', 'long');
  CreateElement(xDSE, 'number:text').Text := '.';
  CreateElement(xDSE, 'number:month').SetAttribute('number:style', 'long');
  CreateElement(xDSE, 'number:text').Text := '.';
  CreateElement(xDSE, 'number:year').SetAttribute('number:style', 'long');
  CreateElement(xDSE, 'number:text').Text := ' ';
  CreateElement(xDSE, 'number:hours').SetAttribute('number:style', 'long');
  CreateElement(xDSE, 'number:text').Text := ':';
  CreateElement(xDSE, 'number:minutes').SetAttribute('number:style', 'long');
end;

function TExportCellDateTimeWS.GetNumberFormatXLSXStr: OWideString;
begin
  Result := 'dd/mm/yyyy\ hh:mm';
end;

{ TExportCellTime }

procedure TExportCellTime.CreateStyleNodeODS(aStylesNode: IXMLElement; aStyleName: OWideString);
var xTSE: IXMLElement;
begin
  xTSE := CreateElement(aStylesNode, 'number:time-style');
  xTSE.SetAttribute('style:name', aStyleName);//StringToUTF8XML(aStyleName));

  CreateElement(xTSE, 'number:hours').SetAttribute('number:style', 'long');
  CreateElement(xTSE, 'number:text').Text := ':';
  CreateElement(xTSE, 'number:minutes').SetAttribute('number:style', 'long');
  CreateElement(xTSE, 'number:text').Text := ':';
  CreateElement(xTSE, 'number:seconds').SetAttribute('number:style', 'long');
end;

procedure TExportCellTime.DoAsODS(aRowNode, aCellNode: IXMLElement;
  aColNodes: TInterfaceList; aColStyles: TExportODS_ColStyles;
  aRowStyles: TExportODS_RowStyles; aCellStyles: TExportODS_CellStyles;
  aNumberStyles: TExportODS_NumberStyles; aRealCol: Integer);
begin
  inherited;

  RemoveAttributeSave(aCellNode, 'office:date-value');
  if not SameValue(Value, 0) then begin
    aCellNode.SetAttribute('office:value-type', 'time');//getNamedAttributeNew
    aCellNode.SetAttribute('office:time-value', TimeToODS(Value));//getNamedAttributeNew
  end;
end;

procedure TExportCellTime.DoAsXLSX(aCellNode, aColsNode, aRowNode,
  aMergeCellsNode: IXMLElement; aCellName, aWorkSheetId: OWideString;
  aSharedStrings: TOWideStrings; aStyleSheet: TExportXLSX_StyleSheet;
  aChain: TExportXLSX_Chain; aRow, aCol: Integer; out outStyleId: Integer);
begin
  inherited;

  if SameValue(DateTime, 0) then
    RemoveElementSave(aCellNode, 'v')
  else
    FindCreateElement(aCellNode, 'v').Text := FloatToSqlStr(Frac(DateTime));
end;

function TExportCellTime.GetNumberFormatXLSXStr: OWideString;
begin
  Result := 'hh:mm:ss';
end;

function TExportCellTime.GetSqlText: OWideString;
begin
  Result := TimeToSqlStr(DateTime);
end;

procedure TExportCellTime.SetSqlText(const Value: OWideString);
begin
  DateTime := SqlStrToTime(Value);
end;

{ TExportCellTimeWS }

procedure TExportCellTimeWS.CreateStyleNodeODS(aStylesNode: IXMLElement; aStyleName: OWideString);
var xDSE: IXMLElement;
begin
  xDSE := CreateElement(aStylesNode, 'number:date-style');
  xDSE.SetAttribute('style:name', aStyleName);//StringToUTF8XML(aStyleName));

  CreateElement(xDSE, 'number:hours').SetAttribute('number:style', 'long');
  CreateElement(xDSE, 'number:text').Text := ':';
  CreateElement(xDSE, 'number:minutes').SetAttribute('number:style', 'long');
end;

function TExportCellTimeWS.GetNumberFormatXLSXStr: OWideString;
begin
  Result := 'hh:mm';
end;

{ TExportXLSXFile }

constructor TExportXLSXFile.Create2(const aFileName: OWideString;
  const aExecuteFileAfterExport: Boolean);
begin
  inherited Create;
  
  fFileName := aFileName;
  fExecuteFileAfterExport := aExecuteFileAfterExport;
end;

procedure TExportXLSXFile.DoCreate;
begin
  inherited;

  fFileName := '';
  fExecuteFileAfterExport := False;
end;

function TExportXLSXFile.DoExport: Boolean;
var
  xExportStream: TStream;
  xFS: TOFileStream;
begin
  xExportStream := fExportStream;
  xFS := TOFileStream.Create(fFileName, fmOpenWrite or fmCreate);
  try
    fExportStream := xFS;
    Result := inherited DoExport;
  finally
    xFS.Free;
    fExportStream := xExportStream;
  end;

  if fExecuteFileAfterExport then
    ExecuteFile(fFileName);
end;

{ TExportCellCustomNumber }

procedure TExportCellCustomNumber.AssignTo(Dest: TPersistent);
begin
  if Dest is TExportCellCustomNumber then
  with TExportCellCustomNumber(Dest) do begin
    fValue := Self.fValue;
  end;
  inherited;
end;

procedure TExportCellCustomNumber.DoCreate;
begin
  inherited;

  fValue := 0;
end;

{ TExportODS }

procedure TExportODS.CreateContent(aZip: TAbArchive);
  procedure AddDocConAttributes(aDocumentContentNode: IXMLElement);
  begin
    with aDocumentContentNode do begin
      SetAttribute('xmlns:office', 'urn:oasis:names:tc:opendocument:xmlns:office:1.0');
      SetAttribute('xmlns:style', 'urn:oasis:names:tc:opendocument:xmlns:style:1.0');
      SetAttribute('xmlns:text', 'urn:oasis:names:tc:opendocument:xmlns:text:1.0');
      SetAttribute('xmlns:table', 'urn:oasis:names:tc:opendocument:xmlns:table:1.0');
      SetAttribute('xmlns:draw', 'urn:oasis:names:tc:opendocument:xmlns:drawing:1.0');
      SetAttribute('xmlns:fo', 'urn:oasis:names:tc:opendocument:xmlns:xsl-fo-compatible:1.0');
      SetAttribute('xmlns:xlink', 'http://www.w3.org/1999/xlink');
      SetAttribute('xmlns:dc', 'http://purl.org/dc/elements/1.1/');
      SetAttribute('xmlns:meta', 'urn:oasis:names:tc:opendocument:xmlns:meta:1.0');
      SetAttribute('xmlns:number', 'urn:oasis:names:tc:opendocument:xmlns:datastyle:1.0');
      SetAttribute('xmlns:presentation', 'urn:oasis:names:tc:opendocument:xmlns:presentation:1.0');
      SetAttribute('xmlns:svg', 'urn:oasis:names:tc:opendocument:xmlns:svg-compatible:1.0');
      SetAttribute('xmlns:chart', 'urn:oasis:names:tc:opendocument:xmlns:chart:1.0');
      SetAttribute('xmlns:dr3d', 'urn:oasis:names:tc:opendocument:xmlns:dr3d:1.0');
      SetAttribute('xmlns:math', 'http://www.w3.org/1998/Math/MathML');
      SetAttribute('xmlns:form', 'urn:oasis:names:tc:opendocument:xmlns:form:1.0');
      SetAttribute('xmlns:script', 'urn:oasis:names:tc:opendocument:xmlns:script:1.0');
      SetAttribute('xmlns:ooo', 'http://openoffice.org/2004/office');
      SetAttribute('xmlns:ooow', 'http://openoffice.org/2004/writer');
      SetAttribute('xmlns:oooc', 'http://openoffice.org/2004/calc');
      SetAttribute('xmlns:dom', 'http://www.w3.org/2001/xml-events');
      SetAttribute('xmlns:xforms', 'http://www.w3.org/2002/xforms');
      SetAttribute('xmlns:xsd', 'http://www.w3.org/2001/XMLSchema');
      SetAttribute('xmlns:xsi', 'http://www.w3.org/2001/XMLSchema-instance');
      SetAttribute('xmlns:rpt', 'http://openoffice.org/2005/report');
      SetAttribute('xmlns:of', 'urn:oasis:names:tc:opendocument:xmlns:of:1.2');
      SetAttribute('xmlns:xhtml', 'http://www.w3.org/1999/xhtml');
      SetAttribute('xmlns:grddl', 'http://www.w3.org/2003/g/data-view#');
      SetAttribute('xmlns:field', 'urn:openoffice:names:experimental:ooo-ms-interop:xmlns:field:1.0');
      SetAttribute('office:version', '1.2');
      SetAttribute('grddl:transformation', 'http://docs.oasis-open.org/office/1.2/xslt/odf2rdf.xsl');
    end;
  end;

  procedure ExportWorkSheet(aWorkSheetIndex: Integer; aWorkSheet: TExportWorkSheet;
    aSpreadSheetNode: IXMLElement;
    aColStyles: TExportODS_ColStyles; aRowStyles: TExportODS_RowStyles;
    aCellStyles: TExportODS_CellStyles; aNumberStyles: TExportODS_NumberStyles);
    procedure ExportColumns(aTableNode: IXMLElement; aColNodes: TInterfaceList);
    var IRow, ICell, ICol, xColCount, xMaxColCount: Integer;
      xCell: TExportCell;
    begin
      xMaxColCount := 0;
      for IRow := 0 to aWorkSheet.Rows.Count-1 do begin
        xColCount := 0;
        for ICell := 0 to aWorkSheet.Rows[IRow].Cells.Count-1 do begin
          xCell := aWorkSheet.Rows[IRow].Cells[ICell];
           Inc(xColCount, Max(1, xCell.ColSpan));
        end;
        xMaxColCount := Max(xMaxColCount, xColCount);
      end;

      for ICol := 0 to xMaxColCount-1 do begin
        aColNodes.Add(CreateElement(aTableNode, 'table:table-column'));
      end;
    end;

    procedure ExportRow(aRow: TExportRow; aTableNode: IXMLElement; aColNodes: TInterfaceList);
    var
      xRowNode, xCellNode: IXMLElement;
      xCell: TExportCell;
      I, ColI, xRealCol: Integer;
    begin
      xRowNode := CreateElement(aTableNode, 'table:table-row');
      xRealCol := 0;
      for I := 0 to aRow.Cells.Count-1 do begin
        xCell := aRow.Cells[I];
        xCellNode := CreateElement(xRowNode, 'table:table-cell');

        xCell.DoAsODS(xRowNode, xCellNode, aColNodes, aColStyles, aRowStyles, aCellStyles, aNumberStyles, xRealCol);

        Inc(xRealCol, Max(1, xCell.ColSpan));
        for ColI := 2 to xCell.ColSpan do
          CreateElement(xRowNode, 'table:covered-table-cell');
      end;
    end;

  var
    xTableNode: IXMLElement;
    xColNodes: TInterfaceList;
    I: Integer;
  begin
    xColNodes := TInterfaceList.Create;
    try
      xTableNode := CreateElement(aSpreadSheetNode, 'table:table');
      xTableNode.SetAttribute('table:name', aWorkSheet.Title);
      xTableNode.SetAttribute('table:style-name', 'ta'+IntToStr(aWorkSheetIndex+1));
      ExportColumns(xTableNode, xColNodes);
      for I := 0 to aWorkSheet.Rows.Count-1 do
        ExportRow(aWorkSheet.Rows[I], xTableNode, xColNodes);
    finally
      xColNodes.Free;
    end;
  end;
var
  xXML: IXMLDocument;
  xMS: TMemoryStream;
  xColStyles: TExportODS_ColStyles;
  xRowStyles: TExportODS_RowStyles;
  xCellStyles: TExportODS_CellStyles;
  xNumberStyles: TExportODS_NumberStyles;
  xFontFaceDeclsNode, xStylesNode, xBodyNode, xSpreadSheetNode: IXMLElement;
  I: Integer;
  xPI: IXMLProcessingInstruction;
begin
  xXML := CreateXMLDoc;//TjanXMLTree.Create('office:document-content', '', nil);
  xColStyles := TExportODS_ColStyles.Create;
  xRowStyles := TExportODS_RowStyles.Create;
  xCellStyles := TExportODS_CellStyles.Create;
  xNumberStyles := TExportODS_NumberStyles.Create;
  try
    xPI := xXML.CreateProcessingInstruction('xml', DefXMLHeader);
    xXML.AppendChild(xPI);

    xXML.DocumentElement := xXML.CreateElement('office:document-content');
    AddDocConAttributes(xXML.DocumentElement);
    xFontFaceDeclsNode := CreateElement(xXML.DocumentElement, 'office:font-face-decls');
    xStylesNode := CreateElement(xXML.DocumentElement, 'office:automatic-styles');
    xBodyNode := CreateElement(xXML.DocumentElement, 'office:body');
    xSpreadSheetNode := CreateElement(xBodyNode, 'office:spreadsheet');

    for I := 0 to WorkSheets.Count-1 do begin
      ExportWorkSheet(I, WorkSheets[I], xSpreadSheetNode, xColStyles, xRowStyles, xCellStyles, xNumberStyles);
    end;

    xColStyles.Export(xStylesNode);
    xRowStyles.Export(xStylesNode);
    xCellStyles.Export(xStylesNode, xFontFaceDeclsNode, xNumberStyles);
    xNumberStyles.Export(xStylesNode);
    ExportWorkSheetStyles(xStylesNode);
    xMS := TMemoryStream.Create;
    try
      xXML.SaveToStream(xMS);
      xMS.Position := 0;

      aZip.AddFromStream('content.xml', xMS);
    finally
      xMS.Free;
    end;
  finally
    xNumberStyles.Free;
    xColStyles.Free;
    xRowStyles.Free;
    xCellStyles.Free;
    xXML := nil;
  end;

end;

procedure TExportODS.CreateManifest(aZip: TAbArchive);
  function AddEntry(aFileName: OWideString): OWideString;
  begin
    Result := '<manifest:file-entry manifest:media-type="text/xml" manifest:full-path="'+aFileName+'"/>';
  end;
var
  xSS: TStringStream;
  s: OWideString;
begin
  s :=
    DefXMLHeaderBrackets+CrLf+
    '<manifest:manifest xmlns:manifest="urn:oasis:names:tc:opendocument:xmlns:manifest:1.0">'+
      '<manifest:file-entry manifest:media-type="application/vnd.oasis.opendocument.spreadsheet" manifest:version="1.2" manifest:full-path="/"/>'+
      AddEntry('content.xml')+
      AddEntry('styles.xml')+
      AddEntry('meta.xml')+
      AddEntry('settings.xml')+
    '</manifest:manifest>';
  xSS := TStringStream.Create(s);
  try
    xSS.Position := 0;
    aZip.AddFromStream('META-INF\manifest.xml', xSS);
  finally
    xSS.Free;
  end;
end;

procedure TExportODS.CreateMeta(aZip: TAbArchive);
var
  xSS: TStringStream;
  s: UTF8String;
begin
  s :=
    DefXMLHeaderBrackets+CrLf+
    '<office:document-meta xmlns:office="urn:oasis:names:tc:opendocument:xmlns:office:1.0"'+
    ' xmlns:xlink="http://www.w3.org/1999/xlink" xmlns:dc="http://purl.org/dc/elements/1.1/"'+
    ' xmlns:meta="urn:oasis:names:tc:opendocument:xmlns:meta:1.0" xmlns:ooo="http://openoffice.org/2004/office"'+
    ' xmlns:grddl="http://www.w3.org/2003/g/data-view#" office:version="1.2"'+
    ' grddl:transformation="http://docs.oasis-open.org/office/1.2/xslt/odf2rdf.xsl">'+
      '<office:meta>'+
        '<meta:generator>'+StringToUTF8XML(fApplication)+'</meta:generator>'+
        //'<dc:date>'++'</dc:date>'+//2010-10-03T19:22:23.25
        '<dc:creator>'+StringToUTF8XML(fAuthor)+'</dc:creator>'+
      '</office:meta>'+
    '</office:document-meta>';
  xSS := TStringStream.Create(s);
  try
    xSS.Position := 0;
    aZip.AddFromStream('meta.xml', xSS);
  finally
    xSS.Free;
  end;
end;

procedure TExportODS.CreateMimeType(aZip: TAbArchive);
var
  xSS: TStringStream;
  s: OWideString;
begin
  s := 'application/vnd.oasis.opendocument.spreadsheet';
  xSS := TStringStream.Create(s);
  try
    xSS.Position := 0;
    aZip.AddFromStream('mimetype', xSS);
  finally
    xSS.Free;
  end;
end;

procedure TExportODS.CreateSettings(aZip: TAbArchive);
var
  xSS: TStringStream;
  s: OWideString;
begin
  s :=
    DefXMLHeaderBrackets+CrLf+
    '<office:document-settings xmlns:office="urn:oasis:names:tc:opendocument:xmlns:office:1.0"'+
    ' xmlns:xlink="http://www.w3.org/1999/xlink" xmlns:config="urn:oasis:names:tc:opendocument:xmlns:config:1.0"'+
    ' xmlns:ooo="http://openoffice.org/2004/office" office:version="1.2">'+
    '</office:document-settings>';
  xSS := TStringStream.Create(s);
  try
    xSS.Position := 0;
    aZip.AddFromStream('settings.xml', xSS);
  finally
    xSS.Free;
  end;
end;

procedure TExportODS.CreateStyles(aZip: TAbArchive);
var
  xSS: TStringStream;
  s: OWideString;
begin
  s :=
    DefXMLHeaderBrackets+CrLf+
    '<office:document-styles xmlns:office="urn:oasis:names:tc:opendocument:xmlns:office:1.0" xmlns:style="urn:oasis:names:tc:opendocument:xmlns:style:1.0"'+
    ' xmlns:text="urn:oasis:names:tc:opendocument:xmlns:text:1.0" xmlns:table="urn:oasis:names:tc:opendocument:xmlns:table:1.0"'+
    ' xmlns:draw="urn:oasis:names:tc:opendocument:xmlns:drawing:1.0" xmlns:fo="urn:oasis:names:tc:opendocument:xmlns:xsl-fo-compatible:1.0"'+
    ' xmlns:xlink="http://www.w3.org/1999/xlink" xmlns:dc="http://purl.org/dc/elements/1.1/" xmlns:meta="urn:oasis:names:tc:opendocument:xmlns:meta:1.0"'+
    ' xmlns:number="urn:oasis:names:tc:opendocument:xmlns:datastyle:1.0" xmlns:presentation="urn:oasis:names:tc:opendocument:xmlns:presentation:1.0"'+
    ' xmlns:svg="urn:oasis:names:tc:opendocument:xmlns:svg-compatible:1.0" xmlns:chart="urn:oasis:names:tc:opendocument:xmlns:chart:1.0"'+
    ' xmlns:dr3d="urn:oasis:names:tc:opendocument:xmlns:dr3d:1.0" xmlns:math="http://www.w3.org/1998/Math/MathML"'+
    ' xmlns:form="urn:oasis:names:tc:opendocument:xmlns:form:1.0" xmlns:script="urn:oasis:names:tc:opendocument:xmlns:script:1.0"'+
    ' xmlns:ooo="http://openoffice.org/2004/office" xmlns:ooow="http://openoffice.org/2004/writer" xmlns:oooc="http://openoffice.org/2004/calc"'+
    ' xmlns:dom="http://www.w3.org/2001/xml-events" xmlns:rpt="http://openoffice.org/2005/report" xmlns:of="urn:oasis:names:tc:opendocument:xmlns:of:1.2"'+
    ' xmlns:xhtml="http://www.w3.org/1999/xhtml" xmlns:grddl="http://www.w3.org/2003/g/data-view#" office:version="1.2"'+
    ' grddl:transformation="http://docs.oasis-open.org/office/1.2/xslt/odf2rdf.xsl">'+
    ExportWorkSheetsMasterStyles+
    '</office:document-styles>';
  {$IFDEF UNICODE}
  xSS := TStringStream.Create(s, TUTF8Encoding.Create);
  {$ELSE}
  xSS := TStringStream.Create(UTF8Encode(s));
  {$ENDIF}
  try
    xSS.Position := 0;
    aZip.AddFromStream('styles.xml', xSS);
  finally
    xSS.Free;
  end;
end;

procedure TExportODS.DoCreate;
begin
  inherited;

  fExportStream := nil;
end;

function TExportODS.DoExport: Boolean;
var
  xZip: TOZipArchive;
begin
  xZip := TOZipArchive.CreateFromStream(fExportStream, '');
  try
    CreateMimeType(xZip);
    CreateStyles(xZip);
    CreateMeta(xZip);
    CreateSettings(xZip);
    CreateManifest(xZip);

    CreateContent(xZip);

    xZip.Save;
  finally
    xZip.Free;
  end;
  Result := True;
end;
function TExportODS.ExportWorkSheetsMasterStyles: OWideString;
  procedure GetStyleText(aPageLayoutNode: IXMLElement; aTagName: OWideString);
  begin
    with CreateElement(CreateElement(aPageLayoutNode, 'style:'+aTagName+'-style'), 'style:header-footer-properties') do begin
      SetAttribute('fo:min-height', '0.75cm');
      SetAttribute('fo:margin-left', '0cm');
      SetAttribute('fo:margin-right', '0cm');
      SetAttribute('fo:margin-bottom', '0.25cm');
    end;
  end;
var
  xMasterStyles, xAutomaticStyles: IXMLDocument;
  xSSE: IXMLElement;
  I: Integer;
  xPageLayoutNode: IXMLElement;
begin
  xMasterStyles := CreateXMLDoc;
  xAutomaticStyles := CreateXMLDoc;

  xMasterStyles.DocumentElement := xMasterStyles.CreateElement('office:master-styles');
  xAutomaticStyles.DocumentElement := xAutomaticStyles.CreateElement('office:automatic-styles');

  for I := 0 to WorkSheets.Count-1 do
  if WorkSheets[I].Header.Used or WorkSheets[I].Footer.Used then begin

    xSSE := CreateElement(xMasterStyles.DocumentElement, 'style:master-page');
    xSSE.SetAttribute('style:name', 'PageStyle_5f_Worksheet_20_'+IntToStr(I+1));
    xSSE.SetAttribute('style:display-name', 'PageStyle_Worksheet '+IntToStr(I+1));
    xSSE.SetAttribute('style:page-layout-name', 'Mpm'+IntToStr(I+1));

    xPageLayoutNode := CreateElement(xAutomaticStyles.DocumentElement, 'style:page-layout');
    xPageLayoutNode.SetAttribute('style:name', 'Mpm'+IntToStr(I+1));
    if WorkSheets[I].Header.Used then begin
      WorkSheets[I].Header.DoAsODS(CreateElement(xSSE, 'style:header'), xAutomaticStyles.DocumentElement, 'h');
      GetStyleText(xPageLayoutNode, 'header');
    end;
    if WorkSheets[I].Footer.Used then begin
      WorkSheets[I].Footer.DoAsODS(CreateElement(xSSE, 'style:footer'), xAutomaticStyles.DocumentElement, 'f');
      GetStyleText(xPageLayoutNode, 'footer');
    end;
  end;

  Result := xAutomaticStyles.XML + xMasterStyles.XML;
end;

procedure TExportODS.ExportWorkSheetStyles(aStylesNode: IXMLElement);
var
  I: Integer;
  xSSE: IXMLElement;
begin
  for I := 0 to WorkSheets.Count-1 do
  if WorkSheets[I].Header.Used or WorkSheets[I].Footer.Used then begin
    xSSE := CreateElement(aStylesNode, 'style:style');
    xSSE.SetAttribute('style:name', 'ta'+IntToStr(I+1));
    xSSE.SetAttribute('style:family', 'table');
    xSSE.SetAttribute('style:master-page-name', 'PageStyle_5f_Worksheet_20_'+IntToStr(I+1));
    with CreateElement(xSSE, 'style:table-properties') do begin
      SetAttribute('table:display', 'true');
    end;
  end;
end;

{ TExportODS_ColStyles }

function TExportODS_ColStyles.CreateAddUnique(aWidth: Integer): Integer;
var I : Integer;
begin
  for I := 0 to Count-1 do
  if Items[I].fWidth = aWidth then begin
    Result := I;
    exit;
  end;

  //not found
  with TExportODS_ColStyle.CreateAdd(Self) do begin
    fWidth := aWidth;
  end;
  Result := Count-1;
end;

procedure TExportODS_ColStyles.Export(aStylesNode: IXMLElement);
var
  I: Integer;
  xSSE: IXMLElement;
begin
  for I := 0 to Count-1 do begin
    xSSE := CreateElement(aStylesNode, 'style:style');
    xSSE.SetAttribute('style:name', StyleName(I));
    xSSE.SetAttribute('style:family', 'table-column');
    with CreateElement(xSSE, 'style:table-column-properties') do begin
      SetAttribute('style:column-width', FloatToSqlStr(Items[I].fWidth * 0.01838)+'cm');
    end;
  end;
end;

function TExportODS_ColStyles.GetC(Index: Integer): TExportODS_ColStyle;
begin
  Result := TExportODS_ColStyle(inherited Items[Index]);
end;

procedure TExportODS_ColStyles.SetC(Index: Integer;
  const Value: TExportODS_ColStyle);
begin
  inherited Items[Index] := Value;
end;

function TExportODS_ColStyles.StyleName(Index: Integer): OWideString;
begin
  Result := 'co'+IntToStr(Index+1);
end;

{ TExportODS_RowStyles }

function TExportODS_RowStyles.CreateAddUnique(aHeight: Integer): Integer;
var I : Integer;
begin
  for I := 0 to Count-1 do
  if Items[I].fHeight = aHeight then begin
    Result := I;
    exit;
  end;

  //not found
  with TExportODS_RowStyle.CreateAdd(Self) do begin
    fHeight := aHeight;
  end;
  Result := Count-1;
end;

procedure TExportODS_RowStyles.Export(aStylesNode: IXMLElement);
var I: Integer;
  xSSE: IXMLElement;
begin
  for I := 0 to Count-1 do begin
    xSSE := CreateElement(aStylesNode, 'style:style');
    xSSE.SetAttribute('style:name', StyleName(I));
    xSSE.SetAttribute('style:family', 'table-row');
    with CreateElement(xSSE, 'style:table-row-properties') do begin
      SetAttribute('style:row-height', FloatToSqlStr(Items[I].fHeight * 0.0272)+'cm');
    end;
  end;
end;

function TExportODS_RowStyles.GetC(Index: Integer): TExportODS_RowStyle;
begin
  Result := TExportODS_RowStyle(inherited Items[Index]);
end;

procedure TExportODS_RowStyles.SetC(Index: Integer;
  const Value: TExportODS_RowStyle);
begin
  inherited Items[Index] := Value;
end;

function TExportODS_RowStyles.StyleName(Index: Integer): OWideString;
begin
  Result := 'ro'+IntToStr(Index+1);
end;

{ TExportODS_CellStyles }

function TExportODS_CellStyles.CreateAddStyleUnique(
  aCell: TExportCell; aNumberStyles: TExportODS_NumberStyles): Integer;
var I: Integer;
begin
  for I := 0 to Count-1 do
  if SameFont(Items[I].fFont, aCell.Font) and
    (Items[I].fBGColor = aCell.fBGColor) and
    (Items[I].fBorders = aCell.fBorders) and
    (Items[I].fAlignment = aCell.Alignment) and
    (Items[I].fVAlignment = aCell.VAlignment) and
    (Items[I].fWrapText = aCell.WrapText) and
    (not (aCell is TExportCellCustomNumber) or (
      not(Items[I].fNumberStyleId = -1) and
      (aNumberStyles[Items[I].fNumberStyleId].fXLSXNumberFormat = TExportCellCustomNumber(aCell).GetNumberFormatXLSXStr))
    )
  then begin
    Result := I;
    exit;
  end;

  //NEW STYLE
  with TExportODS_CellStyle.CreateAdd(Self) do begin
    fFont.Assign(aCell.Font);
    fBGColor := aCell.BGColor;
    fBorders := aCell.Borders;
    fAlignment := aCell.Alignment;
    fVAlignment := aCell.VAlignment;
    fWrapText := aCell.WrapText;
    if (aCell is TExportCellCustomNumber) then
      fNumberStyleId := aNumberStyles.CreateAddNumberStyleUnique(TExportCellCustomNumber(aCell))
    else
      fNumberStyleId := -1;
  end;
  Result := Count-1;
end;

procedure TExportODS_CellStyles.Export(aStylesNode,
  aFontFaceDeclsNode: IXMLElement; aNumberStyles: TExportODS_NumberStyles);

  procedure AddBorderAttribute(aCellPropNode: IXMLElement; aAddBorder: Boolean; aBorderName: OWideString);
  begin
    if aAddBorder then
      aCellPropNode.SetAttribute('fo:border-'+aBorderName, '0.002cm solid #000000')
    else
      aCellPropNode.SetAttribute('fo:border-'+aBorderName, 'none');
  end;

var I: Integer;
  xFontFaceDecls: TOWideStringList;
  xStyleNode, xTextPropNode, xCellPropNode, xParaPropNode: IXMLElement;
begin

  xFontFaceDecls := TOWideStringList.Create;
  try
    xFontFaceDecls.Duplicates := dupIgnore;
    xFontFaceDecls.Sorted := True;
    for I := 0 to Count-1 do begin
      xFontFaceDecls.Add(Items[I].fFont.Name);

      xStyleNode := CreateElement(aStylesNode, 'style:style');
      xStyleNode.SetAttribute('style:name', StyleName(I));
      xStyleNode.SetAttribute('style:family', 'table-cell');
      if Items[I].fNumberStyleId >= 0 then
        xStyleNode.SetAttribute('style:data-style-name', aNumberStyles.StyleName(Items[I].fNumberStyleId));

      xCellPropNode := CreateElement(xStyleNode, 'style:table-cell-properties');
      xCellPropNode.SetAttribute('style:vertical-align', VAlignmentToODS(Items[I].fVAlignment));
      if Items[I].fBGColor <> clNone then begin
        xCellPropNode.SetAttribute('fo:background-color', ColorToHTML(Items[I].fBGColor));
      end;
      if Items[I].fBorders = [cbTop, cbRight, cbBottom, cbLeft] then begin
        xCellPropNode.SetAttribute('fo:border', '0.002cm solid #000000');
      end else begin
        AddBorderAttribute(xCellPropNode, cbTop in Items[I].fBorders, 'top');
        AddBorderAttribute(xCellPropNode, cbBottom in Items[I].fBorders, 'bottom');
        AddBorderAttribute(xCellPropNode, cbLeft in Items[I].fBorders, 'left');
        AddBorderAttribute(xCellPropNode, cbRight in Items[I].fBorders, 'right');
      end;

      xParaPropNode := CreateElement(xStyleNode, 'style:paragraph-properties');
      xParaPropNode.SetAttribute('fo:text-align', AlignmentToODS(Items[I].fAlignment));

      xTextPropNode := CreateElement(xStyleNode, 'style:text-properties');
      FontToODS(Items[I].fFont, xTextPropNode);
    end;


    for I := 0 to xFontFaceDecls.Count-1 do
    with CreateElement(aFontFaceDeclsNode, 'style:font-face') do begin
      SetAttribute('style:name', xFontFaceDecls[I]);
      SetAttribute('svg:font-family', xFontFaceDecls[I]);
    end;
  finally
    xFontFaceDecls.Free;
  end;
end;

function TExportODS_CellStyles.GetC(Index: Integer): TExportODS_CellStyle;
begin
  Result := TExportODS_CellStyle(inherited Items[Index]);
end;

procedure TExportODS_CellStyles.SetC(Index: Integer;
  const Value: TExportODS_CellStyle);
begin
  inherited Items[Index] := Value;
end;

function TExportODS_CellStyles.StyleName(Index: Integer): OWideString;
begin
  Result := 'ce'+IntToStr(Index+1);
end;

{ TExportODS_CellStyle }

procedure TExportODS_CellStyle.DoCreate;
begin
  inherited;

  fFont := TFont.Create;
  fBGColor := clNone;
  fBorders := [cbTop, cbRight, cbBottom, cbLeft];
  fWrapText := False;
  fNumberStyleId := -1;
end;

procedure TExportODS_CellStyle.DoDestroy;
begin
  fFont.Free;

  inherited;
end;

{ TExportODS_NumberStyles }

function TExportODS_NumberStyles.CreateAddNumberStyleUnique(
  aCell: TExportCellCustomNumber): Integer;
var I : Integer;
begin
  for I := 0 to Count-1 do
  if Items[I].fXLSXNumberFormat = aCell.GetNumberFormatXLSXStr then begin
    Result := I;
    exit;
  end;

  //not found
  with TExportODS_NumberStyle.CreateAdd(Self) do begin
    fXLSXNumberFormat := aCell.GetNumberFormatXLSXStr;
    fParentCell := aCell;
  end;
  Result := Count-1;
end;

procedure TExportODS_NumberStyles.Export(aStylesNode: IXMLElement);
var I: Integer;
begin
  for I := 0 to Count-1 do
    Items[I].fParentCell.CreateStyleNodeODS(aStylesNode, StyleName(I));
end;

function TExportODS_NumberStyles.GetC(Index: Integer): TExportODS_NumberStyle;
begin
  Result := TExportODS_NumberStyle(inherited Items[Index]);
end;

procedure TExportODS_NumberStyles.SetC(Index: Integer;
  const Value: TExportODS_NumberStyle);
begin
  inherited Items[Index] := Value;
end;

function TExportODS_NumberStyles.StyleName(Index: Integer): OWideString;
begin
  Result := 'N'+IntToStr(501+Index);
end;

{ TExportODSSaveDialog }

function TExportODSSaveDialog.DoExport: Boolean;
var
  SD: TOWideSaveDialog;
begin
  SD := TOWideSaveDialog.Create(nil);
  try
    SD.DefaultExt := odsexport_defaultextension;
    SD.Filter := odsexport_dialogfilter;
    SD.Title := odsexport_title;
    if fFileName <> '' then
      SD.FileName := fFileName
    else
      SD.FileName := NormFileName(Title);
    SD.Options := SD.Options + [ofOverwritePrompt];

    Result := SD.Execute;
    if Result then begin
      fFileName := SD.FileName;

      Result := inherited DoExport;
    end;
  finally
    SD.Free;
  end;
end;

{ TExportODSFile }

constructor TExportODSFile.Create2(const aFileName: OWideString;
  const aExecuteFileAfterExport: Boolean);
begin
  inherited Create;
  
  fFileName := aFileName;
  fExecuteFileAfterExport := aExecuteFileAfterExport;
end;

procedure TExportODSFile.DoCreate;
begin
  inherited;

  fFileName := '';
  fExecuteFileAfterExport := False;
end;

function TExportODSFile.DoExport: Boolean;
var
  xExportStream: TStream;
  xFS: TOFileStream;
begin
  xExportStream := fExportStream;
  xFS := TOFileStream.Create(fFileName, fmOpenWrite or fmCreate);
  try
    fExportStream := xFS;
    Result := inherited DoExport;
  finally
    xFS.Free;
    fExportStream := xExportStream;
  end;

  if fExecuteFileAfterExport then
    ExecuteFile(fFileName);
end;

{ TExportHeader }

constructor TExportHeader.Create(aExportObject: TExportObject);
begin
  inherited Create;

  fFont := TFont.Create;
  fFont.Assign(aExportObject.DefaultFont);
  fText := '';
end;

destructor TExportHeader.Destroy;
begin
  fFont.Free;
  
  inherited;
end;

procedure TExportHeader.DoAsODS(const aNode, aOfficeAutomaticStyleNode: IXMLElement; const aStyleName: OWideString);
var xP: IXMLElement;
  function AddTextElement(aTagName, aTagValue: OWideString): IXMLElement;
    function CreateStyleElement: IXMLElement;
    begin
      Result := CreateElement(xP, 'text:span');
      Result.SetAttribute('text:style-name', 'MT'+aStyleName);
    end;
  var xS, xT: IXMLElement;
    xPos, xNextPos: Integer;
  begin
    if aTagName <> '' then begin
      xS := CreateStyleElement;
      xT := CreateElement(xS, 'text:'+aTagName);
      if aTagValue <> '' then
        xT.Text := aTagValue
      else
        xT.Text := '???';
      Result := xT;
    end else begin
      aTagValue := ChangeChar(aTagValue, #13#10, #13);
      aTagValue := ChangeChar(aTagValue, #10, #13);
      xPos := 1;
      while xPos <= Length(aTagValue) do begin
        xNextPos := PosEx(#13, aTagValue, xPos);
        if xNextPos > 0 then begin
          xS := CreateStyleElement;
          xS.Text := Copy(aTagValue, xPos, xNextPos-xPos);
          xP := CreateElement(aNode, 'text:p');
          xPos := xNextPos+1;
        end else begin
          break;
        end;
      end;//while
      if xPos <= Length(aTagValue) then begin
        xNextPos := Length(aTagValue)+1;
        xS := CreateStyleElement;
        xS.Text := Copy(aTagValue, xPos, xNextPos-xPos);
      end;
    end;
  end;
var
  xBeginPos, xPos, xNextPos, xClosePos, xOpen2Pos: Integer;
  xKey, xKeyTag: OWideString;
  xStyleNode, xTextPropNode, xKeyNode: IXMLElement;
begin
  xStyleNode := CreateElement(aOfficeAutomaticStyleNode, 'style:style');
  xStyleNode.SetAttribute('style:name', 'MT'+aStyleName);
  xStyleNode.SetAttribute('style:family', 'text');
  xTextPropNode := CreateElement(xStyleNode, 'style:text-properties');
  FontToODS(Font, xTextPropNode);

  xP := CreateElement(aNode, 'text:p');

  xPos := 1;
  xBeginPos := 1;
  while xPos <= Length(Text) do begin
    xNextPos := PosEx('[', Text, xPos);
    if xNextPos > 0 then begin
      xClosePos := PosEx(']', Text, xNextPos+1);
      xOpen2Pos := PosEx('[', Text, xNextPos+1);
      if (xClosePos > 0) and (xOpen2Pos > 0) and (xOpen2Pos < xClosePos) then begin
        xPos := xOpen2Pos;
        continue;
      end;

      if xClosePos > 0 then begin
        xPos := xClosePos+1;
        xKey := Copy(Text, xNextPos+1, xClosePos-xNextPos-1);
        xKeyTag := '';
        if xKey = 'PAGENUMBER' then begin
          xKeyTag := 'page-number';
        end else if xKey = 'PAGECOUNT' then begin
          xKeyTag := 'page-count';
        end else if xKey = 'SHEETNAME' then begin
          xKeyTag := 'sheet-name';
        end else if xKey = 'DATE' then begin
          xKeyTag := 'date';
        end else if xKey = 'TIME' then begin
          xKeyTag := 'time';
        end else if xKey = 'FILENAME' then begin
          xKeyTag := 'file-name';
        end;

        if xKeyTag <> '' then begin
          if xNextPos > xBeginPos then
            AddTextElement('', Copy(Text, xBeginPos, xNextPos-xBeginPos));
          xKeyNode := AddTextElement(xKeyTag, '');
          if xKeyTag = 'file-name' then
            xKeyNode.SetAttribute('text:display', 'name-and-extension');

          xBeginPos := xPos;
        end;
      end else begin//if xClosePos > 0 then begin
        break;
      end;
    end else begin//if xNextPos > 0 then begin
      break;
    end;
  end;//while

  if xBeginPos <= Length(Text) then begin
    xNextPos := Length(Text)+1;
    AddTextElement('', Copy(Text, xBeginPos, xNextPos-xBeginPos));
  end;
end;

function TExportHeader.DoAsXLSX: OWideString;
var xText: OWideString;
begin
  Result := '&"'+Font.Name;
  if (fsBold in Font.Style) or (fsItalic in Font.Style) then begin
    Result := Result + ',';
    if (fsBold in Font.Style) then
      Result := Result + 'Bold ';
    if (fsItalic in Font.Style) then
      Result := Result + 'Italic ';
  end;
  Result := Trim(Result);//must be here
  Result := Result + '"&'+IntToStr(Font.Size);
  if fsStrikeOut in Font.Style then
    Result := Result + '&S';
  if fsUnderline in Font.Style then
    Result := Result + '&U';

  Result := Result + '&K'+ColorToHTML(Font.Color, False);

  xText := Text;
  xText := ChangeChar(xText, '&', '_');
  xText := ChangeChar(xText, '[PAGENUMBER]', '&P');
  xText := ChangeChar(xText, '[PAGECOUNT]', '&N');
  xText := ChangeChar(xText, '[SHEETNAME]', '&A');
  xText := ChangeChar(xText, '[DATE]', '&D');
  xText := ChangeChar(xText, '[TIME]', '&T');
  xText := ChangeChar(xText, '[FILENAME]', '&F');
  
  Result := Result + xText;
end;

{ TExportHeaders }

constructor TExportHeaders.Create(aExportObject: TExportObject);
begin
  fLeft := TExportHeader.Create(aExportObject);
  fCenter := TExportHeader.Create(aExportObject);
  fRight := TExportHeader.Create(aExportObject);
end;

destructor TExportHeaders.Destroy;
begin
  fLeft.Free;
  fCenter.Free;
  fRight.Free;
  
  inherited;
end;

procedure TExportHeaders.DoAsODS(const aNode, aOfficeAutomaticStyleNode: IXMLElement; const aStyleName: OWideString);
begin
  if fLeft.Text <> '' then
    fLeft.DoAsODS(CreateElement(aNode, 'style:region-left'), aOfficeAutomaticStyleNode, aStyleName+'1');
  if fCenter.Text <> '' then
    fCenter.DoAsODS(CreateElement(aNode, 'style:region-center'), aOfficeAutomaticStyleNode, aStyleName+'2');
  if fRight.Text <> '' then
    fRight.DoAsODS(CreateElement(aNode, 'style:region-right'), aOfficeAutomaticStyleNode, aStyleName+'3');
end;

procedure TExportHeaders.DoAsXLSX(const aNode: IXMLElement);
begin
  if fLeft.Text <> '' then
    aNode.Text := aNode.Text + '&L' + fLeft.DoAsXLSX; 
  if fCenter.Text <> '' then
    aNode.Text := aNode.Text + '&C' + fCenter.DoAsXLSX; 
  if fRight.Text <> '' then
    aNode.Text := aNode.Text + '&R' + fRight.DoAsXLSX;
end;

function TExportHeaders.Used: Boolean;
begin
  Result := (fLeft.Text <> '') or (fCenter.Text <> '') or (fRight.Text <> '');
end;

end.