unit oFileUtils;

{
    Copyright (C) 2011 Ondrej Pokorny
      http://www.kluug.at
}

{$I OWideSupp.inc}

interface

uses Windows, Classes, {$IFDEF O_USE_TNT}TntClasses,{$ENDIF} OWideSupp;

procedure WideFindFiles(Path, Mask: OWideString; const Attr: Integer; const outStrList: TOWideStrings;
  const FullPath: Boolean = true; const MaxCount: Integer = 0); overload;
procedure WideFindFilesMulti(Path: OWideString; const Masks: Array of OWideString; const Attr: Integer; const outStrList: TOWideStrings;
  const FullPath: Boolean = true; const MaxCount: Integer = 0); overload;
function WideFileSize(FileName: OWideString): Int64;
function WideFileSizeText(FileName: OWideString): OWideString; overload;
function WideFileSizeText(FileSize: Int64): OWideString; overload;
function NormPathA(Path: AnsiString; const AddSlash: Boolean = true; const Slash: AnsiChar = '/'): AnsiString;
function NormPath(Path: OWideString; const AddSlash: Boolean = true; const Slash: WideChar = '/'): OWideString;
function NormFileName(FileName: OWideString; CharReplaceTo: WideChar = '-'): OWideString ;
function WideExtractFileNameOnly(FileName: OWideString): OWideString;
function WideExtractFileExtOnly(FileName: OWideString): OWideString;

function OExtractFilePath(aFileName: OWideString): OWideString; {$IFDEF O_DELPHI_2007UP}inline;{$ENDIF}
function OExtractFileDir(aFileName: OWideString): OWideString; {$IFDEF O_DELPHI_2007UP}inline;{$ENDIF}
function OExtractFileName(aFileName: OWideString): OWideString; {$IFDEF O_DELPHI_2007UP}inline;{$ENDIF}
function OExtractFileExt(aFileName: OWideString): OWideString; {$IFDEF O_DELPHI_2007UP}inline;{$ENDIF}
function OFileExists(aFileName: OWideString): Boolean; {$IFDEF O_DELPHI_2007UP}inline;{$ENDIF}

function ExecuteFile(const FileName, Params, DefaultDir: OWideString;
  ShowCmd: Integer): THandle; overload;
function ExecuteFile(const FileName: OWideString): THandle; overload;
function ExecuteFileLong(const FileName, Params, DefaultDir: OWideString): LongBool;
function ExecuteFileAndWaitLong(const FileName, Params, DefaultDir: OWideString): LongBool;
procedure RunDosWithResults(aDosApp: OWideString; aResults: TOWideStrings);

function ShowFileInExplorer(const aFileName: OWideString; const aExploreDirs: Boolean = False; const aShowCmd: Integer = SW_SHOWNORMAL): THandle;

function GetWindowsTempDir: OWideString;
function GetWindowsTempDirA: AnsiString;
function GetSystemRootDir: OWideString;
function GetSystem32Dir: OWideString;
function GetUniqueFileName(Path, Prefix, Ext: OWideString): OWideString;
function GetUniqueFileName2(aFileName: OWideString): OWideString;

function WidePathCanonicalize(Path: OWideString): OWideString;
function WideRelToAbs(const RelPath, BasePath: OWideString): OWideString;
function WideAbsToRel(const AbsPath, BasePath: OWideString): OWideString;

function DeleteFileToTrash(aFileName: OWideString; aShowConfirmation: Boolean): Boolean;
function DeleteFilesToTrash(aFileNames: TOWideStrings; aShowConfirmation: Boolean): Boolean;
function RenameFileShell(aOldFileName, aNewFileName: OWideString): Boolean;

function GetShortName(sLongName: OWideString): OWideString;

type
  TFilesToDelete = class(TOWideStringList)
  public
    destructor Destroy; override;
  end;


  TOHMemoryStream = {$IFDEF O_DELPHI_2007UP}class helper for TMemoryStream{$ELSE}class(TMemoryStream){$ENDIF}
  public
    procedure LoadFromFileAllowShare(const FileName: OWideString);

    procedure LoadFromFileW(const FileName: OWideString);
    procedure SaveToFileW(const FileName: OWideString);
  end;


implementation

uses SysUtils, {$IFDEF O_USE_TNT}TntSysUtils, TntWindows,{$ENDIF} {$IFDEF O_DELPHI_2007UP}WideStrUtils, {$ENDIF}ShellApi, TntWideChars,
  PathUtils,
  {$IFDEF O_DELPHI_XE2}
  Vcl.Forms,
  {$ELSE}
  Forms,
  {$ENDIF}
  StrUtils;

{$IFNDEF O_DELPHI_2007UP}
const
  FILE_ATTRIBUTE_NOT_CONTENT_INDEXED = $00002000;
{$ENDIF}

function GetShortName(sLongName: OWideString): OWideString;
var
  sShortName    : OWideString;
  nShortNameLen : integer;
begin
  SetLength(sShortName, MAX_PATH);
  nShortNameLen := GetShortPathNameW(
    PWideChar(sLongName), PWideChar(sShortName), MAX_PATH - 1
  );
  if (0 = nShortNameLen) then
  begin
    // handle errors...
  end;
  SetLength(sShortName, nShortNameLen);
  Result := sShortName;
end;

function RenameFileShell(aOldFileName, aNewFileName: OWideString): Boolean;
var
  Fo: TSHFileOpStructW;
  ppFrom, ppTo: OWideString;
begin
  FillChar(Fo, sizeof(Fo), #0);
  Fo.Wnd    := 0;
  Fo.wFunc  := FO_RENAME;

  ppFrom := NormPath(aOldFileName, False, '\') + #0#0;
  Fo.pFrom  := PWideChar(ppFrom);
  ppTo := NormPath(aNewFileName, False, '\') + #0#0;
  Fo.pTo  := PWideChar(ppTo);

  //if aShowConfirmation then  THERE ARE NEVER SHOWN CONFIRMATIONS BY WINDOWS
    Fo.fFlags := FOF_ALLOWUNDO;
  //else
    //Fo.fFlags := FOF_ALLOWUNDO + FOF_NOCONFIRMATION;

  if SHFileOperationW(Fo) <> 0 then begin
    //MessageBox(0, PAnsiChar(SysErrorMessage(GetLastError)), 'Error', MB_OK OR MB_ICONERROR);
    Result := False;
  end else
    Result := True;
end;

function DeleteFileToTrash(aFileName: OWideString; aShowConfirmation: Boolean): Boolean;
var xStrL: TOWideStringList;
begin
  xStrL := TOWideStringList.Create;
  try
    xStrL.Text := aFileName;
    Result := DeleteFilesToTrash(xStrL, aShowConfirmation);
  finally
    xStrL.Free;
  end;
end;

function DeleteFilesToTrash(aFileNames: TOWideStrings; aShowConfirmation: Boolean): Boolean;
var
  Fo: TSHFileOpStructW;
  buffer: OWideString;
  I: Integer;
begin
  FillChar(Fo, sizeof(Fo), #0);
  Fo.Wnd    := 0;
  Fo.wFunc  := FO_DELETE;
  buffer := '';
  for I := 0 to aFileNames.Count - 1 do
    buffer := buffer + NormPath(aFileNames[I], False, '\') + #0;
  buffer := buffer + #0;
  Fo.pFrom  := PWideChar(buffer);
  if aShowConfirmation then
    Fo.fFlags := FOF_ALLOWUNDO
  else
    Fo.fFlags := FOF_ALLOWUNDO + FOF_NOCONFIRMATION;

  if SHFileOperationW(Fo) <> 0 then begin
    //MessageBox(0, PAnsiChar(SysErrorMessage(GetLastError)), 'Error', MB_OK OR MB_ICONERROR);
    Result := False;
  end else
    Result := True;
end;

function WidePathCanonicalize(Path: OWideString): OWideString;
var
  buf: array[0..MAX_PATH] of WideChar;
begin
  PathCanonicalizeW(buf, PWideChar(Path));
  Result := buf;
end;

function WideRelToAbs(const RelPath, BasePath: OWideString): OWideString;
begin
  if (Length(RelPath) > 1) and (Relpath[2] = ':') then begin//absolute path!
    Result := RelPath;
  end else begin
    Result := WidePathCanonicalize(NormPath(BasePath, True, '\') + RelPath);
  end;
end;

function WideAbsToRel(const AbsPath, BasePath: OWideString): OWideString;
var
  Path: array[0..MAX_PATH-1] of WideChar;
begin
  PathRelativePathToW(@Path[0], PWideChar(BasePath), FILE_ATTRIBUTE_DIRECTORY, PWideChar(AbsPath), 0);
  if Path <> '' then//same disk
    Result := Path
  else//different disks, return absolute path
    Result := AbsPath;
end;

function OGetWindowsDirectory(lpBuffer: POWideChar; uSize: Cardinal): Cardinal; {$IFDEF O_DELPHI_2007UP}inline;{$ENDIF}
begin
  {$IFDEF UNICODE}
  Result := GetWindowsDirectory(lpBuffer, uSize);
  {$ELSE}
    {$IFDEF O_USE_TNT}
      Result := Tnt_GetWindowsDirectoryW(lpBuffer, uSize);
    {$ELSE}
      Result := GetWindowsDirectoryW(lpBuffer, uSize);
    {$ENDIF}
  {$ENDIF}
end;

function OGetSystemDirectory(lpBuffer: POWideChar; uSize: Cardinal): Cardinal; {$IFDEF O_DELPHI_2007UP}inline;{$ENDIF}
begin
  {$IFDEF UNICODE}
  Result := GetSystemDirectory(lpBuffer, uSize);
  {$ELSE}
    {$IFDEF O_USE_TNT}
      Result := Tnt_GetSystemDirectoryW(lpBuffer, uSize);
    {$ELSE}
      Result := GetSystemDirectoryW(lpBuffer, uSize);
    {$ENDIF}
  {$ENDIF}
end;

function OGetTempPath(lpBuffer: POWideChar; uSize: Cardinal): Cardinal; {$IFDEF O_DELPHI_2007UP}inline;{$ENDIF}
begin
  {$IFDEF UNICODE}
  Result := GetTempPath(uSize, lpBuffer);
  {$ELSE}
    {$IFDEF O_USE_TNT}
      Result := Tnt_GetTempPathW(uSize, lpBuffer);
    {$ELSE}
      Result := GetTempPathW(uSize, lpBuffer);
    {$ENDIF}
  {$ENDIF}
end;

function GetSystemRootDir: OWideString;
var
  lng: DWORD;
begin
  SetLength(Result, MAX_PATH);
  lng := OGetWindowsDirectory(PWideChar(Result), MAX_PATH);
  SetLength(Result, lng);
  Result := NormPath(Result, True, '\');
end;

function GetSystem32Dir: OWideString;
var
  lng: DWORD;
begin
  SetLength(Result, MAX_PATH);
  lng := OGetSystemDirectory(PWideChar(Result), MAX_PATH);
  SetLength(Result, lng);
  Result := NormPath(Result, True, '\');
end;

function GetWindowsTempDir: OWideString;
var
  lng: DWORD;
begin
  SetLength(Result, MAX_PATH);
  lng := OGetTempPath(PWideChar(Result), MAX_PATH);
  SetLength(Result, lng);
  Result := NormPath(Result, True, '\');
end;

function GetWindowsTempDirA: AnsiString;
var
  lng: DWORD;
begin
  SetLength(Result, MAX_PATH);
  lng := GetTempPathA(MAX_PATH, PAnsiChar(Result));
  SetLength(Result, lng);
  Result := AnsiString(NormPath(OWideString(Result), True, '\'));
end;

function GetUniqueFileName(Path, Prefix, Ext: OWideString): OWideString;
var
  I: Integer;
  F: THandle;
begin
  Result := '';
  Path := NormPath(Path, True, '\');
  if (Length(Ext) > 0) and (Ext[1] <> '.') then
    Ext := '.'+Ext;

  I := 0;
  repeat
    Result := Path+Prefix+IntToStr(I)+Ext;

    F := CreateFileW(PWideChar(Result), GENERIC_WRITE, 0, nil, CREATE_NEW,
        FILE_ATTRIBUTE_TEMPORARY or FILE_ATTRIBUTE_NOT_CONTENT_INDEXED, 0);
        
    Inc(I);
  until F <> INVALID_HANDLE_VALUE;

  CloseHandle(F);
end;

function OExtractFilePath(aFileName: OWideString): OWideString;
begin
  {$IFDEF O_USE_TNT}
  Result := WideExtractFilePath(aFileName);
  {$ELSE}
  Result := ExtractFilePath(aFileName);
  {$ENDIF}
end;

function OExtractFileDir(aFileName: OWideString): OWideString; 
begin
  {$IFDEF O_USE_TNT}
  Result := WideExtractFileDir(aFileName);
  {$ELSE}
  Result := ExtractFileDir(aFileName);
  {$ENDIF}
end;

function OExtractFileName(aFileName: OWideString): OWideString; 
begin
  {$IFDEF O_USE_TNT}
  Result := WideExtractFileName(aFileName);
  {$ELSE}
  Result := ExtractFileName(aFileName);
  {$ENDIF}
end;

function OExtractFileExt(aFileName: OWideString): OWideString; 
begin
  {$IFDEF O_USE_TNT}
  Result := WideExtractFileExt(aFileName);
  {$ELSE}
  Result := ExtractFileExt(aFileName);
  {$ENDIF}
end;

function OFileExists(aFileName: OWideString): Boolean; 
begin
  {$IFDEF O_USE_TNT}
  Result := WideFileExists(aFileName);
  {$ELSE}
  Result := FileExists(aFileName);
  {$ENDIF}
end;

function GetUniqueFileName2(aFileName: OWideString): OWideString;
var
  I: Integer;
  xFileName, xExt: OWideString;
begin
  Result := '';
  aFileName := NormPath(aFileName, False, '\');
  xFileName := OExtractFilePath(aFileName)+WideExtractFileNameOnly(aFileName);
  xExt := OExtractFileExt(aFileName);

  Result := aFileName;
  I := 0;
  while OFileExists(Result) do begin
    Result := xFileName+' ('+Format('%.2d', [I])+')'+xExt;
    Inc(I);
  end;
end;

function ShowFileInExplorer(const aFileName: OWideString; const aExploreDirs: Boolean; const aShowCmd: Integer): THandle;
var xParams: OWideString;
begin
  xParams := '/select, "'+aFileName+'"';
  if aExploreDirs then
    xParams := '/e, '+xParams;

  Result := ExecuteFile('EXPLORER.EXE', xParams, OExtractFileDir(aFileName), aShowCmd);
end;

function ExecuteFile(const FileName: OWideString): THandle; overload;
begin
  Result := ExecuteFile(FileName, '', OExtractFileDir(FileName), 0);
end;

function ExecuteFile(const FileName, Params, DefaultDir: OWideString;
  ShowCmd: Integer): THandle;
var
  zFileName, zParams, zDir: array[0..2000] of WideChar;//2000 is the real barrier, I tested it
begin
  {$IFDEF UNICODE}
  Result := ShellExecuteW(0, 'open',
    StrPCopy(zFileName, FileName), StrPCopy(zParams, Params),
    StrPCopy(zDir, DefaultDir), ShowCmd);
  {$ELSE}
  Result := ShellExecuteW(0, 'open',
    StrPCopyWW(zFileName, FileName), StrPCopyWW(zParams, Params),
    StrPCopyWW(zDir, DefaultDir), ShowCmd);
  {$ENDIF}
end;

function ExecuteFileLong(const FileName, Params, DefaultDir: OWideString): LongBool;
var
  xI: TShellExecuteInfoW;
  zFileName, zParams, zDir: array[0..65535] of WideChar;
begin
  FillChar(xI, SizeOf(xI), 0) ;
  xI.cbSize := SizeOf(TShellExecuteInfoW);
  {$IFDEF UNICODE}
  xI.lpFile := StrPCopy(zFileName, FileName);
  xI.lpParameters := StrPCopy(zParams, Params);
  xI.lpDirectory := StrPCopy(zDir, DefaultDir);
  {$ELSE}
  xI.lpFile := StrPCopyWW(zFileName, FileName);
  xI.lpParameters := StrPCopyWW(zParams, Params);
  xI.lpDirectory := StrPCopyWW(zDir, DefaultDir);
  {$ENDIF}
  Result := ShellExecuteExW(@xI);
end;

function ExecuteFileAndWaitLong(const FileName, Params, DefaultDir: OWideString): LongBool;
var
  xI: TShellExecuteInfoW;
  zFileName, zParams, zDir: array[0..65535] of WideChar;
  exitCode: DWord;
begin
  FillChar(xI, SizeOf(xI), 0) ;
  xI.cbSize := SizeOf(TShellExecuteInfo);
  {$IFDEF UNICODE}
  xI.lpFile := StrPCopy(zFileName, FileName);
  xI.lpParameters := StrPCopy(zParams, Params);
  xI.lpDirectory := StrPCopy(zDir, DefaultDir);
  {$ELSE}
  xI.lpFile := StrPCopyWW(zFileName, FileName);
  xI.lpParameters := StrPCopyWW(zParams, Params);
  xI.lpDirectory := StrPCopyWW(zDir, DefaultDir);
  {$ENDIF}
  Result := ShellExecuteExW(@xI);
  repeat
    exitCode := WaitForSingleObject(xI.hProcess, 500);
    Application.ProcessMessages;
  until (exitCode <> WAIT_TIMEOUT);
end;

procedure RunDosWithResults(aDosApp: OWideString; aResults: TOWideStrings);
const
  ReadBuffer = 2400;//2400;
var
  Security: TSecurityAttributes;
  ReadPipe, WritePipe: THandle;
  start: {$IFDEF O_DELPHI_7}TntWindows.{$ENDIF}TStartUpInfoW;
  ProcessInfo : Windows.TProcessInformation;
  Buffer: PWideChar;
  //BytesRead: DWord;
  Apprunning: DWord;
begin

  with Security do begin
    nlength := SizeOf(TSecurityAttributes) ;
    binherithandle := true;
    lpsecuritydescriptor := nil;
  end;

  if CreatePipe(ReadPipe, WritePipe, @Security, 0) then begin
    Buffer := AllocMem(ReadBuffer + 1) ;
    FillChar(Start,Sizeof(Start),#0) ;
    start.cb := SizeOf(start) ;
    start.hStdOutput := WritePipe;
    start.hStdInput := ReadPipe;
    start.dwFlags := STARTF_USESTDHANDLES + STARTF_USESHOWWINDOW;
    start.wShowWindow := SW_HIDE;

    if CreateProcessW(nil,
           PWideChar(aDosApp),
           @Security,
           @Security,
           true,
           NORMAL_PRIORITY_CLASS,
           nil,
           nil,
           start,
           ProcessInfo)
    then begin
      repeat
        Apprunning := WaitForSingleObject(ProcessInfo.hProcess,100) ;
        Application.ProcessMessages;
        {repeat
          BytesRead := 0;
          //ReadFile(ReadPipe, Buffer[0], ReadBuffer, BytesRead, nil) ;
          //Buffer[BytesRead]:= #0;
          //OemToAnsi(Buffer,Buffer) ;
          //if Assigned(aResults) then
          //  aResults.Text := aResults.Text + OWideString(Buffer) ;
        until (BytesRead < ReadBuffer) ;}
      until (Apprunning <> WAIT_TIMEOUT) ;
     end;
     FreeMem(Buffer) ;
     CloseHandle(ProcessInfo.hProcess) ;
     CloseHandle(ProcessInfo.hThread) ;
     CloseHandle(ReadPipe) ;
     CloseHandle(WritePipe) ;
  end;
end;

function NormFileName(FileName: OWideString; CharReplaceTo: WideChar): OWideString;
begin
  Result := FileName;
  Result := OWideStringReplace(Result, '\', CharReplaceTo, [rfReplaceAll]);
  Result := OWideStringReplace(Result, '/', CharReplaceTo, [rfReplaceAll]);
  Result := OWideStringReplace(Result, ':', CharReplaceTo, [rfReplaceAll]);
  Result := OWideStringReplace(Result, '*', CharReplaceTo, [rfReplaceAll]);
  Result := OWideStringReplace(Result, '?', CharReplaceTo, [rfReplaceAll]);
  Result := OWideStringReplace(Result, '"', CharReplaceTo, [rfReplaceAll]);
  Result := OWideStringReplace(Result, '<', CharReplaceTo, [rfReplaceAll]);
  Result := OWideStringReplace(Result, '>', CharReplaceTo, [rfReplaceAll]);
  Result := OWideStringReplace(Result, '|', CharReplaceTo, [rfReplaceAll]);
end;

function NormPath(Path: OWideString; const AddSlash: Boolean; const Slash: WideChar): OWideString ;
begin
  if (Length(Path) > 0) then begin
    if AddSlash and (not (AnsiChar(Path[length(Path)]) in ['\', '/'])) then
      Path := Path + Slash
    else if not AddSlash and (AnsiChar(Path[length(Path)]) in ['\', '/']) then
      Delete(Path, Length(Path), 1) ;
  end;
  Path := OWideStringReplace(Path, '\', Slash, [rfReplaceAll]);
  Path := OWideStringReplace(Path, '/', Slash, [rfReplaceAll]);

  result := Path ;
end;

function NormPathA(Path: AnsiString; const AddSlash: Boolean; const Slash: AnsiChar): AnsiString;
begin
  if (Length(Path) > 0) then begin
    if AddSlash and (not (AnsiChar(Path[length(Path)]) in ['\', '/'])) then
      Path := Path + Slash
    else if not AddSlash and (AnsiChar(Path[length(Path)]) in ['\', '/']) then
      Delete(Path, Length(Path), 1) ;
  end;
  Path := AnsiString(StringReplace(OWideString(Path), '\', OWideString(Slash), [rfReplaceAll]));
  Path := AnsiString(StringReplace(OWideString(Path), '/', OWideString(Slash), [rfReplaceAll]));

  result := Path ;
end;

procedure WideFindFilesMulti(Path: OWideString; const Masks: Array of OWideString; const Attr: Integer; const outStrList: TOWideStrings;
  const FullPath: Boolean = true; const MaxCount: Integer = 0); overload;
var I: Integer;
begin
  for I := 0 to High(Masks) do begin
    WideFindFiles(Path, Masks[I], Attr, outStrList, FullPath, MaxCount);
  end;
end;


procedure WideFindFiles(Path, Mask: OWideString; const Attr: Integer;
  const outStrList: TOWideStrings; const FullPath: Boolean; const MaxCount: Integer);
{$IFDEF O_USE_TNT}
var
  SearchRec: TSearchRecW;
  IsFound: Boolean;
  FileName: OWideString;
begin
  Path := NormPath(Path, True, '\');
  IsFound := WideFindFirst(Path+Mask, Attr, SearchRec) = 0;
  try
    while IsFound and ((MaxCount = 0) or (outStrList.Count < MaxCount)) do begin
      if FullPath then
        FileName := Path + SearchRec.Name
      else
        FileName := SearchRec.Name;

      outStrList.Add(FileName);
      IsFound := WideFindNext(SearchRec) = 0;
    end;
  finally
    WideFindClose(SearchRec);
  end;
{$ELSE}
var
  SearchRec: TSearchRec;
  IsFound: Boolean;
  FileName: OWideString;
begin
  Path := NormPath(Path, True, '\');
  IsFound := FindFirst(Path+Mask, Attr, SearchRec) = 0;
  try
    while IsFound and ((MaxCount = 0) or (outStrList.Count < MaxCount)) do begin
      if FullPath then
        FileName := Path + SearchRec.Name
      else
        FileName := SearchRec.Name;

      outStrList.Add(FileName);
      IsFound := FindNext(SearchRec) = 0;
    end;
  finally
    FindClose(SearchRec);
  end;
{$ENDIF}
end;

function WideFileSize(FileName: OWideString): Int64;
{$IFDEF O_USE_TNT}
var
  SearchRec: TSearchRecW;
begin
  if WideFindFirst(FileName, faAnyFile, SearchRec) = 0 then
    Result := Int64(SearchRec.FindData.nFileSizeHigh) shl Int64(32) +
      Int64(SearchRec.FindData.nFileSizeLow)
  else
    Result := 0;

  WideFindClose(SearchRec);
{$ELSE}
var
  SearchRec: TSearchRec;
begin
  if FindFirst(FileName, faAnyFile, SearchRec) = 0 then
    Result := Int64(SearchRec.FindData.nFileSizeHigh) shl Int64(32) +
      Int64(SearchRec.FindData.nFileSizeLow)
  else
    Result := 0;

  FindClose(SearchRec);
{$ENDIF}
end;

function WideFileSizeText(FileSize: Int64): OWideString;
var Size: Extended;
const Dec = 1024;
begin
  Size := FileSize;
  if Size < Dec then begin
    Result := IntToStr(Round(Size)) + ' B';
    exit;
  end;
  Size := Size / Dec;
  if Size < Dec then begin
    Result := FloatToStrF(Size, ffFixed, 3, 1) + ' KB';
    exit;
  end;
  Size := Size / Dec;
  if Size < Dec then begin
    Result := FloatToStrF(Size, ffFixed, 3, 1) + ' MB';
    exit;
  end;
  Size := Size / Dec;
  if Size < Dec then begin
    Result := FloatToStrF(Size, ffFixed, 3, 1) + ' GB';
    exit;
  end;
end;

function WideFileSizeText(FileName: OWideString): OWideString;
begin
  Result := WideFileSizeText(WideFileSize(FileName));
end;

function OLastDelimiter(const Delimiters, S: OWideString): Integer;
begin
  {$IFDEF O_USE_TNT}
  Result := WideLastDelimiter(Delimiters, S);
  {$ELSE}
  Result := LastDelimiter(Delimiters, S);
  {$ENDIF}
end;

function WideExtractFileNameOnly(FileName: OWideString): OWideString;
var
  I: Integer;
begin
  Result := OExtractFileName(FileName);

  I := OLastDelimiter('.\:', Result);
  if (I > 0) and (Result[I] = '.') then
    Delete(Result, I, MaxInt);
end;

function WideExtractFileExtOnly(FileName: OWideString): OWideString;
begin
  Result := OExtractFileExt(FileName);
  if Length(Result) > 0 then
    Delete(Result, 1, 1);
end;

{ TFilesToDelete }

function ODeleteFile(const Filename: OWideString): Boolean;
begin
  {$IFDEF O_USE_TNT}
  Result := WideDeleteFile(Filename);
  {$ELSE}
  Result := DeleteFile(Filename);
  {$ENDIF}
end;

destructor TFilesToDelete.Destroy;
var I: Integer;
begin
  for I := 0 to Count-1 do
  try
    ODeleteFile(Strings[I]);
  except
  end;

  inherited;
end;

{ TOHMemoryStream }

procedure TOHMemoryStream.LoadFromFileAllowShare(const FileName: OWideString);
var
  Stream: TStream;
begin
  Stream := TOFileStream.Create(FileName, fmOpenRead or fmShareDenyNone);
  try
    Stream.Position := 0;
    LoadFromStream(Stream);
  finally
    Stream.Free;
  end;
end;

procedure TOHMemoryStream.LoadFromFileW(const FileName: OWideString);
begin
  LoadFromFileAllowShare(FileName);
end;

procedure TOHMemoryStream.SaveToFileW(const FileName: OWideString);
var
  Stream: TStream;
begin
  Stream := TOFileStream.Create(FileName, fmCreate);
  try
    SaveToStream(Stream);
  finally
    Stream.Free;
  end;
end;

end.
